/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef COMPGEOMETRYCPP
#define COMPGEOMETRYCPP

#include "CompGeometry.h"



double areaOfPoly(const array<Point3>& poly)
{
	int i, j = poly.size() - 1;
	double area = 0.0;

	for (i = 0; i < poly.size(); i++)
	{
		area += poly[j].x * poly[i].y -
				  poly[j].y * poly[i].x;
		j = i;
	}

	return area;
}

bool isPolyClockwise(const array<Point3>& poly)
{
	return areaOfPoly(poly) < 0.0;
}

void flipPoly(array<Point3>& poly)
{
	Point3 swap;
	int count = poly.size() >> 1;

	int j = poly.size() - 1;
	
	for (int i = 0; i < count; i++)
	{
		swap = poly[i];
		poly[i] = poly[j];
		poly[j] = swap;
		j--;		
	}
}

double areaOfTriangle(const Point3& a, const Point3& b, const Point3& c)
{
	return (a.x * b.y - a.y * b.x) +
			 (b.x * c.y - b.y * c.x) +
			 (c.x * a.y - c.y * a.x);
}

bool isTriangleClockwise(const Point3& a, const Point3& b, const Point3& c)
{
	return areaOfTriangle(a, b, c) < 0.0;
}

bool triangleContainsPoint(const Point3& a, const Point3& b, const Point3& c,
									const Point3& p)
{
	double i = areaOfTriangle(a, b, p);
	
	if (i > 0.0)
	{
		double j = areaOfTriangle(b, c, p);
		if (j > 0.0)
		{
			double k = areaOfTriangle(c, a, p);
			return k > 0.0;
		}
	}
	if (i < 0.0)
	{
		double j = areaOfTriangle(b, c, p);
		if (j < 0.0)
		{
			double k = areaOfTriangle(c, a, p);
			return k < 0.0;
		}
	}
	return false;
}

void generateEdgePlanes(const array<Point3>& poly, array<Plane>& edgePlanes)
{
	int j = poly.size() - 1;

	for (int i = 0; i < poly.size(); i++)
	{
		edgePlanes.push_back(
			Plane( poly[j], poly[i], poly[j] + Vector3(0.0, 0.0, 1.0) ) );
		j = i;
	}
}

bool isPolySelfIntersecting(const array<Point3>& poly)
{
	array<Plane> edgePlanes;
	double t;		//required by Plane::intersect()

	if (poly.size() < 4)
	{
		return false;
	}

	generateEdgePlanes(poly, edgePlanes);

	//int eps = edgePlanes.size();


	int i = edgePlanes.size() - 1;
	for (int j = 0; j < edgePlanes.size(); j++)
	{
		
		int a = poly.size() - 1;
		for (int b = 0; b < poly.size(); b++)
		{
			if (b != i && a != i && a != j)
			{
				if ( edgePlanes[j].intersect( poly[a], poly[b], t ) &&
					  edgePlanes[b].intersect( poly[i], poly[j], t ) )
				{
					return true;
				}
			}
			a = b;
		}

		i = j;
	}

	return false;
}

bool polyHasDegenerateEdges(const array<Point3>& poly)
{
	int j = poly.size() - 1;

	for (int i = 0; i < poly.size(); i++)
	{
		if ( poly[i].distanceSquared( poly[j] ) < EPSILON_SQUARED )
		{
			return true;
		}
		
		j = i;
	}
	
	return false;
}

void flattenPoly(array<Point3>& poly)
{
	for (int i = 0; i < poly.size(); i++)
	{
		poly[i].z = 0.0;
	}
}




#endif
