/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLCREATETORUS_CPP
#define __TOOLCREATETORUS_CPP

#include <strstream>
#include <math.h>

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>


#include "ToolCreateTorus.h"

#include "JournalMesh.h"
#include "CMesh.h"

#include "Point3.h"
#include "Patch.h"





void* __create_ProcCreateTorus()
{
	return (void*)( new ProcCreateTorus() );
}

ObjectType type_ProcCreateTorus( "ProcCreateTorus", "ProcAxis",
											__create_ProcCreateTorus );




ProcCreateTorus::ProcCreateTorus() : ProcAxis()
{
	majSegs = minSegs = 20;
	majRad = 100.0;
	minRad = 40.0;
}

ProcCreateTorus::ProcCreateTorus(Axis a, int majS, int minS,
											double majR, double minR)
											: ProcAxis(a)
{
	majSegs = majS;
	minSegs = minS;
	majRad = majR;
	minRad = minR;
}

ProcCreateTorus::~ProcCreateTorus()
{
}



void ProcCreateTorus::read(FILE *f)
{
	ProcAxis::read(f);

	fread(&majSegs, sizeof(int), 1, f);
	fread(&minSegs, sizeof(int), 1, f);
	fread(&majRad, sizeof(double), 1, f);
	fread(&minRad, sizeof(double), 1, f);
}

void ProcCreateTorus::write(FILE *f)
{
	ProcAxis::write(f);

	fwrite(&majSegs, sizeof(int), 1, f);
	fwrite(&minSegs, sizeof(int), 1, f);
	fwrite(&majRad, sizeof(double), 1, f);
	fwrite(&minRad, sizeof(double), 1, f);
}





string ProcCreateTorus::getDescription() const
{
	return "Torus";
}

const ObjectType* ProcCreateTorus::getType() const
{
	return &type_ProcCreateTorus;
}

const ObjectType* ProcCreateTorus::getTypeStatic()
{
	return &type_ProcCreateTorus;
}



const ObjectType* ProcCreateTorus::getInputType() const
{
	return JournalObject::getTypeStatic();
}



JournalProcess* ProcCreateTorus::clone() const
{
	return new ProcCreateTorus(axis, majSegs, minSegs, majRad, minRad);
}





int ProcCreateTorus::getMajorSegments() const
{
	return majSegs;
}

int ProcCreateTorus::getMinorSegments() const
{
	return minSegs;
}

double ProcCreateTorus::getMajorRadius() const
{
	return majRad;
}

double ProcCreateTorus::getMinorRadius() const
{
	return minRad;
}

	
void ProcCreateTorus::setMajorSegments(int n)
{
	majSegs = n;
	modified();
}

void ProcCreateTorus::setMinorSegments(int n)
{
	minSegs = n;
	modified();
}

void ProcCreateTorus::setMajorRadius(double r)
{
	majRad = r;
	modified();
}

void ProcCreateTorus::setMinorRadius(double r)
{
	minRad = r;
	modified();
}




const ObjectType* ProcCreateTorus::_getOutputType(
								const ObjectType* inputType) const
{
	return JournalMesh::getTypeStatic();
}

JournalObject* ProcCreateTorus::_process(JournalObject *o,
													  JournalTaskProgress *progress)
{
	delete o;

	JournalMesh *hMesh = new JournalMesh();
	CMesh *mesh = hMesh->getMesh();

	vector<CVertex*> vertexPtrs;



	double majAng, majInc = 2.0 * M_PI / (double)majSegs;
	double minAng, minInc = -2.0 * M_PI / (double)minSegs;


	mesh->reserveVertices( majSegs * minSegs );
	mesh->reserveFaces( majSegs * minSegs );


	Axes3 axes = getAxes();


	majAng = 0.0;
	for (int i = 0; i < majSegs; i++)
	{
		double x, y, z, r;
		double cosMaj = cos(majAng);
		double sinMaj = sin(majAng);

		minAng = 0.0;
		for (int j = 0; j < minSegs; j++)
		{
			z = minRad * sin(minAng);
			r = minRad * cos(minAng);

			x = (r + majRad) * cosMaj;
			y = (r + majRad) * sinMaj;

			mesh->addVertex(
				Point3() + (axes.i * x) + (axes.j * y) + (axes.k * z), true );


			minAng += minInc;
		}

		majAng += majInc;
	}


	double u = 0.0, uInc = 1.0 / (double)minSegs;
	double v = 0.0, vInc = 1.0 / (double)majSegs;
	vector<Point3> tex;
	tex.reserve( ( majSegs + 1 ) * ( minSegs + 1 ) );

	for (int i = 0; i <= majSegs; i++)
	{
		u = 0.0;

		for (int j = 0; j <= minSegs; j++)
		{
			tex.push_back( Point3( u, v, 0.0 ) );
			u += uInc;
		}

		v += vInc;
	}

	mesh->addPatch();

	vector<int> faceQuads;
	vector<int> texQuads;

	patchGenerateQuadList( faceQuads, minSegs, majSegs, true, true );
	patchGenerateQuadList( texQuads, minSegs, majSegs, false, false );

	for (int i = 0; i < (int)faceQuads.size(); i += 4)
	{
		mesh->addFace( faceQuads[i], tex[ texQuads[i] ],
							faceQuads[i+1], tex[ texQuads[i+1] ],
							faceQuads[i+2], tex[ texQuads[i+2] ],
							faceQuads[i+3], tex[ texQuads[i+3] ], 0 );
	}


	return (JournalObject*)hMesh;
}







//**************************************************


ToolCreateTorus::ToolCreateTorus(CadView *v)
										   : ToolAxis(v, "Torus", "---Create Torus---",
										   		  	  2, CadToolAxisChooser::PLANE)
{
	majSegsLabel = gtk_label_new("Major Segs:");
	gtk_widget_ref(majSegsLabel);
	gtk_widget_show(majSegsLabel);
	
	majSegsEntry = new NumericEntry( 20.0, 1.0, 1048576.0, 1.0, 32.0, 0.5, 0,
												GTK_SIGNAL_FUNC(majSegsChanged), this );
	gtk_widget_show( majSegsEntry->getWidget() );




	minSegsLabel = gtk_label_new("Minor Segs:");
	gtk_widget_ref(minSegsLabel);
	gtk_widget_show(minSegsLabel);
	minSegsEntry = new NumericEntry( 20.0, 1.0, 1048576.0, 1.0, 32.0, 0.5, 0,
												GTK_SIGNAL_FUNC(minSegsChanged), this );
	gtk_widget_show( minSegsEntry->getWidget() );




	majRadLabel = gtk_label_new("Major Radius:");
	gtk_widget_ref(majRadLabel);
	gtk_widget_show(majRadLabel);
	majRadEntry = new NumericEntry( 100.0, 0.0, 1000000000.0, 0.5, 50.0, 0.5, 4,
											  GTK_SIGNAL_FUNC(majRadChanged), this );
	gtk_widget_show( majRadEntry->getWidget() );




	minRadLabel = gtk_label_new("Minor Radius:");
	gtk_widget_ref(minRadLabel);
	gtk_widget_show(minRadLabel);
	minRadEntry = new NumericEntry( 40.0, 0.0, 1000000000.0, 0.5, 50.0, 0.5, 4,
											  GTK_SIGNAL_FUNC(minRadChanged), this );
	gtk_widget_show( minRadEntry->getWidget() );




	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);

	gtk_signal_connect( GTK_OBJECT(doneButton), "clicked",
							  GTK_SIGNAL_FUNC(doneClicked), this );


	layout.add(majSegsLabel);
	layout.add( majSegsEntry->getWidget() );
	layout.add(minSegsLabel);
	layout.add( minSegsEntry->getWidget() );
	layout.add(majRadLabel);
	layout.add( majRadEntry->getWidget() );
	layout.add(minRadLabel);
	layout.add( minRadEntry->getWidget() );
	layout.lineSpace();
	layout.addAsLine(doneButton);
}

ToolCreateTorus::~ToolCreateTorus()
{
	delete majSegsEntry;
	delete minSegsEntry;
	delete majRadEntry;
	delete minRadEntry;
	
	gtk_widget_unref(majSegsLabel);
	gtk_widget_unref(minSegsLabel);
	gtk_widget_unref(majRadLabel);
	gtk_widget_unref(minRadLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolCreateTorus::getProcessType() const
{
	return ProcCreateTorus::getTypeStatic();
}



void ToolCreateTorus::initialise()
{
	ToolAxis::initialise();

	if ( hasProcess() )
	{
		ProcCreateTorus *proc = (ProcCreateTorus*)journalProcess;

		majSegsEntry->setValue( proc->getMajorSegments() );
		minSegsEntry->setValue( proc->getMinorSegments() );
		majRadEntry->setValue( proc->getMajorRadius() );
		minRadEntry->setValue( proc->getMinorRadius() );
	}
}



void ToolCreateTorus::majSegsChanged(NumericEntry *widget,
												 ToolCreateTorus *klass)
{
	if ( klass->hasProcess() )
	{
		ProcCreateTorus *proc = (ProcCreateTorus*)klass->journalProcess;
		proc->setMajorSegments( klass->majSegsEntry->getValueAsInt() );
		klass->modified();
	}
}

void ToolCreateTorus::minSegsChanged(NumericEntry *widget,
												 ToolCreateTorus *klass)
{
	if ( klass->hasProcess() )
	{
		ProcCreateTorus *proc = (ProcCreateTorus*)klass->journalProcess;
		proc->setMinorSegments( klass->minSegsEntry->getValueAsInt() );
		klass->modified();
	}
}

void ToolCreateTorus::majRadChanged(NumericEntry *widget,
												ToolCreateTorus *klass)
{
	if ( klass->hasProcess() )
	{
		ProcCreateTorus *proc = (ProcCreateTorus*)klass->journalProcess;
		proc->setMajorRadius( klass->majRadEntry->getValue() );
		klass->modified();
	}
}

void ToolCreateTorus::minRadChanged(NumericEntry *widget,
												ToolCreateTorus *klass)
{
	if ( klass->hasProcess() )
	{
		ProcCreateTorus *proc = (ProcCreateTorus*)klass->journalProcess;
		proc->setMinorRadius( klass->minRadEntry->getValue() );
		klass->modified();
	}
}

void ToolCreateTorus::doneClicked(GtkWidget *widget, ToolCreateTorus *klass)
{
	klass->done();
}



#endif

