/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
/************************
 Licensed under the terms
 and conditions described
 in the file 'COPYING'
 ************************/
#ifndef __HERMITESURFACECONTROLPOINT_CPP
#define __HERMITESURFACECONTROLPOINT_CPP

#include "HermiteSurfaceControlPoint.h"





HermiteSurfaceControlPoint::HermiteSurfaceControlPoint()
{
	markFlag = false;
}

HermiteSurfaceControlPoint::HermiteSurfaceControlPoint(const Point3& pos,
																		 const Vector3& sTan,
																		 const Vector3& tTan,
																		 const Vector3& twistVec)
{
	screenPosition = position = pos;
	sTangent = sTan;
	tTangent = tTan;
	twistVector = twistVec;
	markFlag = false;

	screenSTan = position + sTangent;
	screenTTan = position + tTangent;
	screenTwist = position + twistVector;
}



void HermiteSurfaceControlPoint::read(FILE *f)
{
	position.read(f);
	sTangent.read(f);
	tTangent.read(f);
	twistVector.read(f);
	fread( &markFlag, sizeof(bool), 1, f );

	screenPosition = position;
	screenSTan = position + sTangent;
	screenTTan = position + tTangent;
	screenTwist = position + twistVector;
}

void HermiteSurfaceControlPoint::write(FILE *f) const
{
	position.write(f);
	sTangent.write(f);
	tTangent.write(f);
	twistVector.write(f);
	fwrite( &markFlag, sizeof(bool), 1, f );
}




void HermiteSurfaceControlPoint::mark()
{
	markFlag = true;
}

void HermiteSurfaceControlPoint::unmark()
{
	markFlag = false;
}

bool HermiteSurfaceControlPoint::isMarked() const
{
	return markFlag;
}



bool HermiteSurfaceControlPoint::checkVisibilityBit(int b) const
{
	return visData & (1 << b);
}

long HermiteSurfaceControlPoint::getVisibilityData() const
{
	return visData;
}

bool HermiteSurfaceControlPoint::isVisible() const
{
	return visData == 0;
}



void HermiteSurfaceControlPoint::setGeometry(const Point3& pos,
															const Vector3& sTan,
															const Vector3& tTan,
															const Vector3& twistVec)
{
	screenPosition = position = pos;
	sTangent = sTan;
	tTangent = tTan;
	twistVector = twistVec;

	screenSTan = position + sTangent;
	screenTTan = position + tTangent;
	screenTwist = position + twistVector;
}

void HermiteSurfaceControlPoint::setPosition(const Point3& pos)
{
	screenPosition = position = pos;
	fr = 0;
}

void HermiteSurfaceControlPoint::setSTangent(const Vector3& sTan)
{
	sTangent = sTan;
	screenSTan = position + sTangent;
	fr = 0;
}

void HermiteSurfaceControlPoint::setTTangent(const Vector3& tTan)
{
	tTangent = tTan;
	screenTTan = position + tTangent;
	fr = 0;
}

void HermiteSurfaceControlPoint::setTwistVector(const Vector3& twistVec)
{
	twistVector = twistVec;
	screenTwist = position + twistVector;
	fr = 0;
}

const Point3& HermiteSurfaceControlPoint::getPosition() const
{
	return position;
}

const Vector3& HermiteSurfaceControlPoint::getSTangent() const
{
	return sTangent;
}

const Vector3& HermiteSurfaceControlPoint::getTTangent() const
{
	return tTangent;
}

const Vector3& HermiteSurfaceControlPoint::getTwistVector() const
{
	return twistVector;
}

const Point3& HermiteSurfaceControlPoint::getScreenPosition() const
{
	return screenPosition;
}

const Point3& HermiteSurfaceControlPoint::getScreenSTangent() const
{
	return screenSTan;
}

const Point3& HermiteSurfaceControlPoint::getScreenTTangent() const
{
	return screenTTan;
}

const Point3& HermiteSurfaceControlPoint::getScreenTwist() const
{
	return screenTwist;
}





void HermiteSurfaceControlPoint::copyFrom(const HermiteSurfaceControlPoint *v)
{
	screenPosition = position = v->position;
	sTangent = v->sTangent;
	tTangent = v->tTangent;
	twistVector = v->twistVector;
	markFlag = v->markFlag;
	fr = 0;

	screenSTan = position + sTangent;
	screenTTan = position + tTangent;
	screenTwist = position + twistVector;
}


void HermiteSurfaceControlPoint::transform(const Matrix4& m, TransformMask mask)
{
	if (mask & POSITION_BIT)
	{
		screenPosition = position = position * m;
		fr = 0;
	}
	if (mask & STANGENT_BIT)
	{
		sTangent = sTangent * m;
		screenSTan = position + sTangent;
		fr = 0;
	}
	if (mask & TTANGENT_BIT)
	{
		tTangent = tTangent * m;
		screenTTan = position + tTangent;
		fr = 0;
	}
	if (mask & TWISTVEC_BIT)
	{
		twistVector = twistVector * m;
		screenTwist = position + twistVector;
		fr = 0;
	}
}

void HermiteSurfaceControlPoint::transform(const Transformation *x,
														 TransformMask mask)
{
	if (mask & POSITION_BIT)
	{
		screenPosition = position = x->transformPoint(position);
		fr = 0;
	}
	if (mask & STANGENT_BIT)
	{
		sTangent = x->transformVector(sTangent, position);
		screenSTan = position + sTangent;
		fr = 0;
	}
	if (mask & TTANGENT_BIT)
	{
		tTangent = x->transformVector(tTangent, position);
		screenTTan = position + tTangent;
		fr = 0;
	}
	if (mask & TWISTVEC_BIT)
	{
		twistVector = x->transformVector(twistVector, position);
		screenTwist = position + twistVector;
		fr = 0;
	}
}


void HermiteSurfaceControlPoint::select(SelectPredicate predicate,
													 const SelectRegion *region)
{
	bool alreadySelected = markFlag;
	bool inRegion;

	if ( region != NULL )
	{
		inRegion = region->isSelected(position);
	}
	else
	{
		inRegion = true;
	}
	
	bool newSelectState =
		selectPredicateIsSelected( predicate, alreadySelected, inRegion );

	if ( newSelectState )
	{
		mark();
	}
	else
	{
		unmark();
	}
}



void HermiteSurfaceControlPoint::update(const Transformation *w2s,
													 const ViewFrustrum& frus,
													 const Frame& currentFrame)
{
	if (fr != currentFrame)
	{
		screenPosition = w2s->transformPoint(position);
		visData = frus.pointBitList(position);

		screenSTan = w2s->transformPoint( position + sTangent );
		screenTTan = w2s->transformPoint( position + tTangent );
		screenTwist = w2s->transformPoint( position + twistVector );

		fr = currentFrame;
	}
}



HermiteSurfaceControlPoint HermiteSurfaceControlPoint::backup() const
{
	HermiteSurfaceControlPoint v;
	v.position = position;
	v.sTangent = sTangent;
	v.tTangent = tTangent;
	v.twistVector = twistVector;
	v.markFlag = markFlag;
	
	return v;
}

void HermiteSurfaceControlPoint::restore(const HermiteSurfaceControlPoint& bak)
{
	setPosition( bak.position );
	setSTangent( bak.sTangent );
	setTTangent( bak.tTangent );
	setTwistVector( bak.twistVector );
	
	if ( bak.markFlag )
	{
		mark();
	}
	else
	{
		unmark();
	}	
}

	


#endif
