/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __JOURNAL_CPP
#define __JOURNAL_CPP

#include "Journal.h"

#include "fileio.h"
#include "MultiThreading.h"






JournalTask::JournalTask()
{
	journal = NULL;
	progress = NULL;
}

void JournalTask::run()
{
	journal->taskResult = journal->getObjectUnthreaded( objectIndex, progress );

	progress->lock();
	progress->setComplete(true);
	progress->unlock();
	
	journal->busyMutex->unlock();
}




void JournalListener::journalModified(Journal *journ)
{
}





Journal::Cache::Cache(int cacheSize)
{
	size = cacheSize;
}

Journal::Cache::~Cache()
{
	for (list<const JournalObject*>::iterator iter = objects.begin();
		  iter != objects.end();
		  ++iter)
	{
		delete *iter;
	}
}


void Journal::Cache::setCacheSize(int cacheSize)
{
	size = cacheSize;
}


bool Journal::Cache::hasEntryFor(const JournalProcess *p) const
{
	return processToObject.count(p) > 0;
}

const JournalObject* Journal::Cache::operator[](const JournalProcess *p)
{
	if ( hasEntryFor(p) )
	{
		const JournalObject *obj = processToObject[p];
		//objects is now most recently used; move to back of list
		objects.remove(obj);
		objects.push_back(obj);
		
		return processToObject[p];
	}
	else
	{
		return NULL;
	}
}


void Journal::Cache::add(const JournalProcess *p,
								 const JournalObject *o)
{
	objects.push_back(o);
	processToObject[p] = o;
	objectToProcess[o] = p;

	removeOldEntries();	
}

void Journal::Cache::remove(const JournalProcess *p)
{
	if ( !hasEntryFor(p) )
	{
		return;
	}
	
	const JournalObject *o = processToObject[p];

	objects.remove(o);
	objectToProcess.erase(o);
	processToObject.erase(p);

	delete o;
}


void Journal::Cache::removeOldEntries()
{
	while ((int)objects.size() > size)
	{
		processToObject.erase( objectToProcess[objects.front()] );
		objectToProcess.erase( objects.front() );
		delete objects.front();
		objects.pop_front();
	}
}









Journal::Journal(int cacheSize)
{
	objectCache = Cache(cacheSize);
	firstObject = JournalObject();
	busyMutex = new GSCMutex();
	owner = NULL;
}

Journal::Journal(JournalCollection *journOwner, int cacheSize)
{
	objectCache = Cache(cacheSize);
	firstObject = JournalObject();
	busyMutex = new GSCMutex();
	owner = journOwner;
}

Journal::~Journal()
{
	delete busyMutex;
}



void Journal::read(FILE *f)
{
	int historySize;

	fread(&historySize, sizeof(int), 1, f);


	processes.reserve(historySize);

	for (int i = 0; i < historySize; i++)
	{
		string typeName = readString(f);

		JournalProcess *process =
			(JournalProcess*)getTypeManager().create(typeName);

		process->read(f);

		add(process);

		delete process;
	}
}

void Journal::write(FILE *f)
{
	int historySize = size();

	fwrite(&historySize, sizeof(int), 1, f);


	for (int i = 0; i < size(); i++)
	{
		string typeName = processes[i]->getTypeName();

		writeString(f, typeName);

		processes[i]->write(f);
	}
}




void Journal::addListener(JournalListener *listener)
{
	listeners.push_back(listener);
}

void Journal::removeListener(JournalListener *listener)
{
	listeners.remove(listener);
}





void Journal::setOwner(JournalCollection *journOwner)
{
	owner = journOwner;
}

JournalCollection* Journal::getOwner()
{
	return owner;
}

const JournalCollection* Journal::getOwner() const
{
	return owner;
}





JournalObject* Journal::getResultUnthreaded(JournalTaskProgress *progress)
{
	return getObjectUnthreaded( size() - 1, progress );
}

JournalObject* Journal::getObjectUnthreaded(int proc,
														  JournalTaskProgress *progress)
{
	return _getObject( proc, true, progress );
}



JournalTaskProgress* Journal::computeResultThreaded()
{
	return computeObjectThreaded( size() - 1 );
}

JournalTaskProgress* Journal::computeObjectThreaded(int proc)
{
	busyMutex->lock();
	
	taskProgress.lock();
	taskProgress.reset();
	taskProgress.unlock();


	journTask.journal = this;
	journTask.objectIndex = proc;	
	journTask.progress = &taskProgress;

	addJournalTask( &journTask );

	return &taskProgress;
}


JournalObject* Journal::getThreadedComputationResult()
{
	//if still busy, wait until computation is finished by waiting
	busyMutex->lock();
	//unlock
	busyMutex->unlock();
	//return result
	return taskResult;
}





int Journal::size() const
{
	return processes.size();
}

JournalProcess* Journal::getProcess(int proc)
{
	return processes[proc];
}

const ObjectType* Journal::getObjectType(int proc) const
{
	const ObjectType* t = firstObject.getType();
	
	for (int i = 0; i <= proc; i++)
	{
		t = processes[i]->getOutputType(t);
		if (t == NULL)
		{
			return NULL;
		}
	}

	return t;
}

const ObjectType* Journal::getResultType() const
{
	return getObjectType( size() - 1 );
}

int Journal::findIntegrityBreak() const
{
	const ObjectType* t = firstObject.getType();
	
	for (int i = 0; i < size(); i++)
	{
		t = processes[i]->getOutputType(t);
		if (t == NULL)
		{
			return i;
		}
	}

	return -1;
}

bool Journal::checkIntegrity() const
{
	return findIntegrityBreak() == -1;
}

JournalProcess* Journal::add(JournalProcess *p)
{
	JournalProcess *proc = p->clone();
	processes.push_back(proc);
	proc->setOwner(this);
	modified();
	return proc;
}

JournalProcess* Journal::insert(int index, JournalProcess *p)
{
	if ( index == size() )
	{
		//no need to call modified(); called in add()
		return add(p);
	}
	else if (index >= 0 && index < size())
	{
		JournalProcess *proc = p->clone();
		invalidateCache( index, size() );
		processes.insert(index, 1);
		processes[index] = proc;
		proc->setOwner(this);
		modified();
		return proc;
	}
	else
	{
		return NULL;
	}
}

void Journal::remove(int index, int n)
{
	if ( index < 0 || index >= size() )
	{
		return;
	}
	
	for (int i = 0; i < n; i++)
	{
		processes[ index + i ]->setOwner(NULL);
	}
	
	invalidateCache( index, size() );
	processes.erase( index, n );
	modified();
}

JournalProcess* Journal::replace(int index, JournalProcess *p)
{
	int s = size();

	if ( index < 0 || index >= s )
	{
		return NULL;
	}
	invalidateCache( index, size() );
	processes[index]->setOwner(NULL);
	processes.erase( index, 1 );

	if (index == s - 1)
	{
		//no need to call modified(); called in add()
		return add(p);
	}
	else
	{
		JournalProcess *proc = p->clone();
		processes.insert( index, 1 );
		processes[index] = proc;
		proc->setOwner(this);
		modified();
		return proc;
	}
}

void Journal::swap(int index1, int index2)
{
	int s = size();
	
	if ( index1 < 0 || index1 >= s || index2 < 0 || index2 >= s )
	{
		return;
	}

	if (index1 < index2)
	{
		invalidateCache( index1, size() );
	}
	else
	{
		invalidateCache( index2, size() );
	}

	JournalProcess *swapProc = processes[index1];
	processes[index1] = processes[index2];
	processes[index2] = swapProc;

	modified();
}




JournalObject* Journal::_getObject(int proc, bool cacheIt,
											  JournalTaskProgress *progress)
{
	int best = findBestCacheEntryFor(proc);

	JournalObject *obj;

	if ( best == -1 )
	{
		obj = firstObject.clone();
	}
	else
	{
		obj = objectCache[ processes[best] ]->clone();
	}


	if ( progress != NULL )
	{
		progress->lock();
		progress->setNumSubtasks( proc - best );
		progress->unlock();
	}


	for (int i = best + 1; i <= proc; i++)
	{
		if ( progress != NULL )
		{
			progress->lock();
			progress->setSubtasksComplete( i - (best + 1) );
			progress->setSubtaskDescription( processes[i]->getDescription() );
			progress->unlock();
		}
		obj = processes[i]->process( obj, progress );
	}

	if ( progress != NULL )
	{
		progress->lock();
		progress->setSubtasksComplete( proc - best );
		progress->unlock();
	}

	if (cacheIt)
	{
		objectCache.add( processes[proc], obj->clone() );
	}

	return obj;	
}

/*JournalObject* Journal::_getObject(int proc, bool cacheIt)
{
	JournalObject *obj;

	if (proc < 0)
	{
		return firstObject.clone();
	}
	
	if ( isCached(proc) )
	{
		return objectCache[ processes[proc] ]->clone();
	}
	else
	{
		obj = processes[proc]->process( _getObject(proc - 1, false) );

		if (cacheIt)
		{
			objectCache.add( processes[proc], obj->clone() );
		}

		return obj;
	}
}*/

int Journal::findBestCacheEntryFor(int proc)
{
	for (int i = proc; i >= 0; i--)
	{
		if ( isCached(i) )
		{
			return i;
		}
	}

	return -1;
}

int Journal::isCached(int proc)
{
	return objectCache.hasEntryFor( processes[proc] );
}

void Journal::invalidateCache(int begin, int end)
{
	for (int i = begin; i < end; i++)
	{
		if ( isCached(i) )
		{
			objectCache.remove( processes[i] );
		}
	}
}


void Journal::processModified(JournalProcess *p)
{
	for (int i = 0; i < size(); i++)
	{
		if (processes[i] == p)
		{
			invalidateCache( i, size() );
			return;
		}
	}
}


void Journal::modified()
{
	for ( list<JournalListener*>::iterator iter = listeners.begin();
			iter != listeners.end();
			++iter )
	{
		(*iter)->journalModified(this);
	}
}








#endif
