/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __JOURNAL_H
#define __JOURNAL_H

#include <list>
#include <map>

#include "array.h"

#include "MultiThreading.h"

#include "JournalObject.h"
#include "JournalProcess.h"
#include "JournalTaskProgress.h"
#include "JournalThreads.h"
#include "TypeManager.h"
#include "UniqueId.h"



//getObject(), etc return a pointer to an object. This object MUST BE DELETED
//BY THE CALLER




class Journal;



class JournalTask
{
public:
	Journal *journal;
	int objectIndex;
	JournalTaskProgress *progress;


	JournalTask();

	void run();
};




class JournalListener
{
protected:
	virtual void journalModified(Journal *journ);

	friend class Journal;
};




class JournalCollection
{
public:
	virtual Journal* getObject(UniqueId i) = 0;
	virtual const Journal* getObject(UniqueId i) const = 0;
	virtual UniqueId getUniqueId(const Journal *journal) const = 0;

	virtual void setDependancy(UniqueId objectId, UniqueId dependancyId) = 0;
	virtual void removeDependancy(UniqueId objectId) = 0;
};




class Journal : public JournalProcessOwner
{
private:
	class Cache
	{
	private:
		int size;
		
		list<const JournalObject*> objects;
		map<const JournalProcess*, const JournalObject*> processToObject;
		map<const JournalObject*, const JournalProcess*> objectToProcess;

	public:
		Cache(int cacheSize = 8);
		~Cache();

		void setCacheSize(int cacheSize);

		bool hasEntryFor(const JournalProcess *p) const;
		const JournalObject* operator[](const JournalProcess *p);

		void add(const JournalProcess *p, const JournalObject *o);
		void remove(const JournalProcess *p);

	private:
		void removeOldEntries();
	};



	Cache objectCache;
	array<JournalProcess*> processes;
	list<JournalListener*> listeners;
	JournalObject firstObject;


	GSCMutex *busyMutex;
	JournalTask journTask;
	JournalTaskProgress taskProgress;
	JournalObject *taskResult;
	JournalCollection *owner;
	

public:
	Journal(int cacheSize = 8);
	Journal(JournalCollection *journOwner, int cacheSize = 8);
	virtual ~Journal();


	void read(FILE *f);
	void write(FILE *f);



	void addListener(JournalListener *listener);
	void removeListener(JournalListener *listener);



	void setOwner(JournalCollection *journOwner);
	JournalCollection* getOwner();
	const JournalCollection* getOwner() const;


	
	JournalObject* getResultUnthreaded(JournalTaskProgress *progress = NULL);
	JournalObject* getObjectUnthreaded(int proc,
												  JournalTaskProgress *progress = NULL);

	JournalTaskProgress* computeResultThreaded();
	JournalTaskProgress* computeObjectThreaded(int proc);

	JournalObject* getThreadedComputationResult();

	static void threadStaticFunction(Journal *klass);
	
	
	int size() const;
	JournalProcess* getProcess(int proc);
	const ObjectType* getObjectType(int proc) const;
	const ObjectType* getResultType() const;
	int findIntegrityBreak() const;
	bool checkIntegrity() const;
	JournalProcess* add(JournalProcess *p);
	JournalProcess* insert(int index, JournalProcess *p);
	void remove(int index, int n = 1);
	JournalProcess* replace(int index, JournalProcess *p);
	void swap(int index1, int index2);


private:
	JournalObject* _getObject(int proc, bool cacheIt,
									  JournalTaskProgress *progress);
	int findBestCacheEntryFor(int proc);
	int isCached(int proc);
	void invalidateCache(int begin, int end);

	virtual void processModified(JournalProcess *p);

	void modified();



	friend class JournalTask;
};






#endif
