/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLLOFTEXTRUDE_CPP
#define __TOOLLOFTEXTRUDE_CPP

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkcheckbutton.h>


#include "Patch.h"

#include "ToolLoftPolyExtrude.h"

#include "JournalPolyline.h"
#include "JournalMesh.h"
#include "CMesh.h"

#include "Point3.h"

#include "CompGeometry.h"
#include "Triangulate.h"





void* __create_ProcLoftPolyExtrude()
{
	return (void*)( new ProcLoftPolyExtrude() );
}

ObjectType type_ProcLoftPolyExtrude( "ProcLoftPolyExtrude", "JournalProcess",
												 __create_ProcLoftPolyExtrude );




ProcLoftPolyExtrude::ProcLoftPolyExtrude() : JournalProcess()
{
	segments = 16;
	extrudeLength = 100.0;
	flattenFlag = false;
	capStartFlag = false;
	capEndFlag = false;
}

ProcLoftPolyExtrude::ProcLoftPolyExtrude(int extrudeSegs, double length,
													  bool flatten, bool capStart,
													  bool capEnd)
													  : JournalProcess()
{
	segments = extrudeSegs;
	extrudeLength = length;
	flattenFlag = flatten;
	capStartFlag = capStart;
	capEndFlag = capEnd;
}

ProcLoftPolyExtrude::~ProcLoftPolyExtrude()
{
}



void ProcLoftPolyExtrude::read(FILE *f)
{
	JournalProcess::read(f);

	fread(&segments, sizeof(int), 1, f);
	fread(&extrudeLength, sizeof(double), 1, f);
	fread(&flattenFlag, sizeof(bool), 1, f);
	fread(&capStartFlag, sizeof(bool), 1, f);
	fread(&capEndFlag, sizeof(bool), 1, f);
}

void ProcLoftPolyExtrude::write(FILE *f)
{
	JournalProcess::write(f);

	fwrite(&segments, sizeof(int), 1, f);
	fwrite(&extrudeLength, sizeof(double), 1, f);
	fwrite(&flattenFlag, sizeof(bool), 1, f);
	fwrite(&capStartFlag, sizeof(bool), 1, f);
	fwrite(&capEndFlag, sizeof(bool), 1, f);
}





string ProcLoftPolyExtrude::getDescription() const
{
	return "Extrude Polyline";
}

const ObjectType* ProcLoftPolyExtrude::getType() const
{
	return &type_ProcLoftPolyExtrude;
}

const ObjectType* ProcLoftPolyExtrude::getTypeStatic()
{
	return &type_ProcLoftPolyExtrude;
}



const ObjectType* ProcLoftPolyExtrude::getInputType() const
{
	return JournalPolyline::getTypeStatic();
}



JournalProcess* ProcLoftPolyExtrude::clone() const
{
	return new ProcLoftPolyExtrude( segments, extrudeLength,
											  flattenFlag, capStartFlag, capEndFlag );
}


int ProcLoftPolyExtrude::getExtrudeSegments() const
{
	return segments;
}

double ProcLoftPolyExtrude::getExtrudeLength() const
{
	return extrudeLength;
}

bool ProcLoftPolyExtrude::getFlatten() const
{
	return flattenFlag;
}

bool ProcLoftPolyExtrude::getCapStart() const
{
	return capStartFlag;
}

bool ProcLoftPolyExtrude::getCapEnd() const
{
	return capEndFlag;
}

void ProcLoftPolyExtrude::setExtrudeSegments(int n)
{
	segments = n;
	modified();
}

void ProcLoftPolyExtrude::setExtrudeLength(double l)
{
	extrudeLength = l;
	modified();
}

void ProcLoftPolyExtrude::setFlatten(bool flatten)
{
	flattenFlag = flatten;
	modified();
}

void ProcLoftPolyExtrude::setCapStart(bool capStart)
{
	capStartFlag = capStart;
	modified();
}

void ProcLoftPolyExtrude::setCapEnd(bool capEnd)
{
	capEndFlag = capEnd;
	modified();
}


const ObjectType* ProcLoftPolyExtrude::_getOutputType(
									const ObjectType* inputType) const
{
	return JournalMesh::getTypeStatic();
}

JournalObject* ProcLoftPolyExtrude::_process(JournalObject *o,
															JournalTaskProgress *progress)
{
	JournalPolyline *jPolyline = (JournalPolyline*)o;
	CPolyline *poly = jPolyline->getPoly();


	JournalMesh *jMesh = new JournalMesh();
	CMesh *mesh = jMesh->getMesh();


	if ( poly->vertices().size() < 2 )
	{
		delete o;

		return (JournalObject*)jMesh;
	}


	mesh->reserveVertices( poly->vertices().size() * (segments + 1) );
	mesh->reserveFaces( poly->vertices().size() * segments );



	//capping is only permitted if the polyline can be triangulated
	//so temporary variables are used
	bool capStart = capStartFlag;
	bool capEnd = capEndFlag;

	array<Point3> polyVerts;
	array<CGTriangle> cappingTriangles;

	for (CVertexContainer::const_iterator iter = poly->vertices().begin();
		  iter != poly->vertices().end();
		  ++iter)
	{
		polyVerts.push_back( (*iter).getWorld() );
	}

	if ( flattenFlag || capStart || capEnd )
	{
		flattenPoly(polyVerts);
	}


	//capping?
	if (capStart || capEnd)
	{
		//perform checks:
		
		if ( polyVerts.size() < 3 )	//3 or more vertices
		{
			capStart = capEnd = false;
			//printf("ProcLoftPolyExtrude::_process(): < 3 vertices\n");
		}
		else if ( areaOfPoly(polyVerts) == 0.0 )	//area != 0.0
		{
			capStart = capEnd = false;
			//printf("ProcLoftPolyExtrude::_process(): area == 0.0\n");
		}
		else if ( isPolySelfIntersecting(polyVerts) )	//not self intersecting
		{
			capStart = capEnd = false;
			//printf("ProcLoftPolyExtrude::_process(): self-intersecting\n");
		}
		else if ( polyHasDegenerateEdges(polyVerts) )	//no degenerate edges
		{
			capStart = capEnd = false;
			//printf("ProcLoftPolyExtrude::_process(): degenerate edges\n");
		}
		else if ( !triangulate(polyVerts, cappingTriangles) )	//triangulate
		{
			capStart = capEnd = false;
			//printf("ProcLoftPolyExtrude::_process(): could not triangulate\n");
		}
	}

	
	int numEdges, polySize;
	numEdges = polySize = poly->vertices().size();

	if ( !poly->isClosed() )
	{
		numEdges--;
	}


	vector<Point3> tex;
	vector<Point3> capTex;
	tex.reserve( ( segments + 1 ) * polySize );


	Vector3 delta = Vector3(0.0, 0.0, extrudeLength) *
						( 1.0 / (double)segments );
	Vector3 translation;

	double u = 0.0, uInc = 1.0 / (double)numEdges;
	double v = 0.0, vInc = 1.0 / (double)segments;

	for (int i = 0; i <= segments; i++)
	{
		u = 0.0;
		for (int j = 0; j < polySize; j++)
		{
			mesh->addVertex( polyVerts[j] + translation );
		}

		for (int j = 0; j <= numEdges; j++)
		{
			tex.push_back( Point3( u, v, 0.0 ) );
			u += uInc;
		}

		translation += delta;
		v += vInc;
	}


	if ( capStart || capEnd )
	{
		BBox box = BBox(polyVerts);

		capTex.reserve( polyVerts.size() );

		for (int i = 0; i < polyVerts.size(); i++)
		{
			const Point3 &vtx = polyVerts[i];

			double u = ( vtx.x - box.getLower().x ) / box.getWidth();
			double v = ( vtx.y - box.getLower().y ) / box.getHeight();

			capTex.push_back( Point3( u, v, 0.0 ) );
		}
	}

	mesh->addPatch();


	if (capStart)
	{
		mesh->addPatch();
		for (int i = 0; i < cappingTriangles.size(); i++)
		{
			int a = cappingTriangles[i].a;
			int b = cappingTriangles[i].b;
			int c = cappingTriangles[i].c;
			mesh->addFace( a, capTex[a], b, capTex[b], c, capTex[c], 1 );
		}
	}

	vector<int> faceQuads, texQuads;

	patchGenerateQuadList( faceQuads, numEdges, segments, poly->isClosed(), false );
	patchGenerateQuadList( texQuads, numEdges, segments, false, false );

	for (int i = 0; i < (int)faceQuads.size(); i += 4)
	{
		mesh->addFace( faceQuads[ i + 0 ], tex[ texQuads[ i + 0 ] ],
							faceQuads[ i + 1 ], tex[ texQuads[ i + 1 ] ],
							faceQuads[ i + 2 ], tex[ texQuads[ i + 2 ] ],
							faceQuads[ i + 3 ], tex[ texQuads[ i + 3 ] ], 0 );
	}
	
	if (capEnd)
	{
		int offset = (segments * polySize);

		mesh->addPatch();

		int patch = capStart ? 2 : 1;

		for (int i = 0; i < cappingTriangles.size(); i++)
		{
			int a = cappingTriangles[i].a;
			int b = cappingTriangles[i].c;
			int c = cappingTriangles[i].b;
			mesh->addFace( offset + a, capTex[a], offset + b, capTex[b],
								offset + c, capTex[c], patch );
		}
	}



	delete o;

	return (JournalObject*)jMesh;
}







//**************************************************


ToolLoftPolyExtrude::ToolLoftPolyExtrude(CadView *v)
				: ProcessTool( v, "Extrude", "---Extrude Polyline---", 2 )
{
	flattenCheck = gtk_check_button_new_with_label("Flatten Polyline");
	gtk_widget_ref(flattenCheck);
	gtk_widget_show(flattenCheck);
	gtk_signal_connect( GTK_OBJECT(flattenCheck), "clicked",
							  GTK_SIGNAL_FUNC(flattenClicked), this );
	


	segmentsLabel = gtk_label_new("Segments:");
	gtk_widget_ref(segmentsLabel);
	gtk_widget_show(segmentsLabel);

	segmentsEntry = new NumericEntry( 16.0, 1.0, 1048576.0, 1.0, 32.0,
													  0.3, 0,
													  GTK_SIGNAL_FUNC(segmentsChanged),
													  this );
	gtk_widget_show( segmentsEntry->getWidget() );
	

	lengthLabel = gtk_label_new("Length:");
	gtk_widget_ref(lengthLabel);
	gtk_widget_show(lengthLabel);

	lengthEntry = new NumericEntry( 100.0, 0.0, 1.0e8, 0.1, 1.0, 0.3, 4,
										     GTK_SIGNAL_FUNC(lengthChanged), this );
	gtk_widget_show( lengthEntry->getWidget() );
	


	capStartCheck = gtk_check_button_new_with_label("Cap start");
	gtk_widget_ref(capStartCheck);
	gtk_widget_show(capStartCheck);
	gtk_signal_connect( GTK_OBJECT(capStartCheck), "clicked",
							  GTK_SIGNAL_FUNC(capStartClicked), this );

	capEndCheck = gtk_check_button_new_with_label("Cap end");
	gtk_widget_ref(capEndCheck);
	gtk_widget_show(capEndCheck);
	gtk_signal_connect( GTK_OBJECT(capEndCheck), "clicked",
							  GTK_SIGNAL_FUNC(capEndClicked), this );


	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);

	gtk_signal_connect( GTK_OBJECT(doneButton), "clicked",
							  GTK_SIGNAL_FUNC(doneClicked), this );




	layout.addAsLine(flattenCheck);
	layout.lineSpace(1);
	layout.add(segmentsLabel);
	layout.add( segmentsEntry->getWidget() );
	layout.newLine();
	layout.add(lengthLabel);
	layout.add( lengthEntry->getWidget() );
	layout.lineSpace(1);
	layout.add(capStartCheck);
	layout.add(capEndCheck);
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolLoftPolyExtrude::~ToolLoftPolyExtrude()
{
	delete segmentsEntry;

	gtk_widget_unref(flattenCheck);
	gtk_widget_unref(segmentsLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolLoftPolyExtrude::getProcessType() const
{
	return ProcLoftPolyExtrude::getTypeStatic();
}



void ToolLoftPolyExtrude::initialise()
{
	ProcessTool::initialise();

	if ( hasProcess() )
	{
		ProcLoftPolyExtrude *proc = (ProcLoftPolyExtrude*)journalProcess;

		bool flatten = proc->getFlatten();
		bool capStart = proc->getCapStart();
		bool capEnd = proc->getCapEnd();
	
		segmentsEntry->setValue( proc->getExtrudeSegments() );
		lengthEntry->setValue( proc->getExtrudeLength() );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(flattenCheck), flatten );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(capStartCheck), capStart );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(capEndCheck), capEnd );
	}
}



void ToolLoftPolyExtrude::segmentsChanged(NumericEntry *widget,
														ToolLoftPolyExtrude *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolyExtrude *proc = (ProcLoftPolyExtrude*)klass->journalProcess;

		proc->setExtrudeSegments( klass->segmentsEntry->getValueAsInt() );
		klass->modified();
	}
}

void ToolLoftPolyExtrude::lengthChanged(NumericEntry *widget,
													 ToolLoftPolyExtrude *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolyExtrude *proc = (ProcLoftPolyExtrude*)klass->journalProcess;

		proc->setExtrudeLength( klass->lengthEntry->getValue() );
		klass->modified();
	}
}

void ToolLoftPolyExtrude::flattenClicked(GtkWidget *widget,
													  ToolLoftPolyExtrude *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolyExtrude *proc = (ProcLoftPolyExtrude*)klass->journalProcess;

		bool flatten = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(klass->flattenCheck) ) != 0;

		proc->setFlatten(flatten);
		klass->modified();
	}
}

void ToolLoftPolyExtrude::capStartClicked(GtkWidget *widget,
														ToolLoftPolyExtrude *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolyExtrude *proc = (ProcLoftPolyExtrude*)klass->journalProcess;

		bool capStart = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(klass->capStartCheck) ) != 0;

		proc->setCapStart(capStart);
		klass->modified();
	}
}

void ToolLoftPolyExtrude::capEndClicked(GtkWidget *widget,
													 ToolLoftPolyExtrude *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolyExtrude *proc = (ProcLoftPolyExtrude*)klass->journalProcess;

		bool capEnd = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(klass->capEndCheck) ) != 0;

		proc->setCapEnd(capEnd);
		klass->modified();
	}
}

void ToolLoftPolyExtrude::doneClicked(GtkWidget *widget,
												  ToolLoftPolyExtrude *klass)
{
	klass->done();
}



#endif

