/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLLOFTSPIN_CPP
#define __TOOLLOFTSPIN_CPP

#include <math.h>

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkcheckbutton.h>


#include "Patch.h"

#include "ToolLoftPolySpin.h"

#include "JournalPolyline.h"
#include "JournalMesh.h"
#include "CMesh.h"

#include "Point3.h"





void* __create_ProcLoftPolySpin()
{
	return (void*)( new ProcLoftPolySpin() );
}

ObjectType type_ProcLoftPolySpin( "ProcLoftPolySpin", "JournalProcess",
											 __create_ProcLoftPolySpin );




ProcLoftPolySpin::ProcLoftPolySpin() : JournalProcess()
{
	segments = 16;
	flattenFlag = false;
}

ProcLoftPolySpin::ProcLoftPolySpin(int spinSegs, bool flatten)
											  : JournalProcess()
{
	segments = spinSegs;
	flattenFlag = flatten;
}

ProcLoftPolySpin::~ProcLoftPolySpin()
{
}



void ProcLoftPolySpin::read(FILE *f)
{
	JournalProcess::read(f);

	fread(&segments, sizeof(int), 1, f);
	fread(&flattenFlag, sizeof(bool), 1, f);
}

void ProcLoftPolySpin::write(FILE *f)
{
	JournalProcess::write(f);

	fwrite(&segments, sizeof(int), 1, f);
	fwrite(&flattenFlag, sizeof(bool), 1, f);
}





string ProcLoftPolySpin::getDescription() const
{
	return "Spin Polyline";
}

const ObjectType* ProcLoftPolySpin::getType() const
{
	return &type_ProcLoftPolySpin;
}

const ObjectType* ProcLoftPolySpin::getTypeStatic()
{
	return &type_ProcLoftPolySpin;
}



const ObjectType* ProcLoftPolySpin::getInputType() const
{
	return JournalPolyline::getTypeStatic();
}



JournalProcess* ProcLoftPolySpin::clone() const
{
	return new ProcLoftPolySpin(segments, flattenFlag);
}


int ProcLoftPolySpin::getSpinSegments() const
{
	return segments;
}

bool ProcLoftPolySpin::getFlatten() const
{
	return flattenFlag;
}

void ProcLoftPolySpin::setSpinSegments(int n)
{
	segments = n;
	modified();
}

void ProcLoftPolySpin::setFlatten(bool flatten)
{
	flattenFlag = flatten;
	modified();
}


const ObjectType* ProcLoftPolySpin::_getOutputType(
									const ObjectType* inputType) const
{
	return JournalMesh::getTypeStatic();
}

JournalObject* ProcLoftPolySpin::_process(JournalObject *o,
														JournalTaskProgress *progress)
{
	JournalPolyline *jPolyline = (JournalPolyline*)o;
	CPolyline *poly = jPolyline->getPoly();


	JournalMesh *jMesh = new JournalMesh();
	CMesh *mesh = jMesh->getMesh();


	if ( poly->vertices().size() < 2 )
	{
		delete o;

		return (JournalObject*)jMesh;
	}


	mesh->reserveVertices( poly->vertices().size() * segments );
	mesh->reserveFaces( poly->vertices().size() * segments );


	vector<CVertex*> vertexPtrs;


	
	double delta = (M_PI * 2.0) / (double)segments;
	double rotation = 0.0;
	Matrix4 rotMatrix;


	int numEdges, nVerts;
	numEdges = nVerts = poly->vertices().size();

	if ( !poly->isClosed() )
	{
		numEdges--;
	}


	double u = 0.0, uInc = 1.0 / (double)segments;
	double v = 0.0, vInc = 1.0 / (double)numEdges;

	vector<Point3> tex;


	array<int> vertexIndices, textureIndices;
	int vertexCount = 0, texVertexCount = 0;

	for (CVertexContainer::const_iterator iter = poly->vertices().begin();
		  iter != poly->vertices().end();
		  ++iter)
	{
		rotation = 0.0;
		rotMatrix = Matrix4();
		Point3 position = (*iter).getWorld();

		if (flattenFlag)
		{
			position.z = 0.0;
		}
		
		if ( ( position.x == 0.0 ) && ( position.z == 0.0 ) )
		{
			mesh->addVertex( (*iter).getWorld() );
			tex.push_back( Point3( 0.5, v, 0.0 ) );
			
			for (int i = 0; i < segments; i++)
			{
				vertexIndices.push_back( vertexCount );
			}

			for (int i = 0; i <= segments; i++)
			{
				textureIndices.push_back( texVertexCount );
			}
		
			vertexCount++;
			texVertexCount++;
		}
		else
		{
			u = 0.0;
			for (int i = 0; i < segments; i++)
			{
				mesh->addVertex( position * rotMatrix );
				vertexIndices.push_back( vertexCount );
				vertexCount++;
				rotation += delta;
				rotMatrix = Matrix4_rotateY( rotation );
			}

			for (int i = 0; i <= segments; i++)
			{
				tex.push_back( Point3( u, v, 0.0 ) );
				textureIndices.push_back( texVertexCount );
				texVertexCount++;
				u += uInc;
			}
		}

		v += vInc;
	}

	if ( poly->isClosed() )
	{
		Point3 position = poly->vertices().front().getWorld();

		if (flattenFlag)
		{
			position.z = 0.0;
		}
		
		if ( ( position.x == 0.0 ) && ( position.z == 0.0 ) )
		{
			tex.push_back( Point3( 0.5, v, 0.0 ) );
			
			for (int i = 0; i <= segments; i++)
			{
				textureIndices.push_back( texVertexCount );
			}
		
			texVertexCount++;
		}
		else
		{
			u = 0.0;
			for (int i = 0; i <= segments; i++)
			{
				tex.push_back( Point3( u, v, 0.0 ) );
				textureIndices.push_back( texVertexCount );
				texVertexCount++;
				u += uInc;
			}
		}
	}

	mesh->addPatch();

	vector<int> faceQuads, texQuads;

	patchGenerateQuadList( faceQuads, segments, numEdges, true, poly->isClosed() );
	patchGenerateQuadList( texQuads, segments, numEdges, false, false );


	int vtxIndex[4], texVtxIndex[4];
	for (int i = 0; i < (int)faceQuads.size(); i += 4)
	{
		vtxIndex[0] = vertexIndices[ faceQuads[ i ] ];
		vtxIndex[1] = vertexIndices[ faceQuads[ i + 1 ] ];
		vtxIndex[2] = vertexIndices[ faceQuads[ i + 2 ] ];
		vtxIndex[3] = vertexIndices[ faceQuads[ i + 3 ] ];

		texVtxIndex[0] = textureIndices[ texQuads[ i ] ];
		texVtxIndex[1] = textureIndices[ texQuads[ i + 1 ] ];
		texVtxIndex[2] = textureIndices[ texQuads[ i + 2 ] ];
		texVtxIndex[3] = textureIndices[ texQuads[ i + 3 ] ];

		bool topEdgeDegenerate = vtxIndex[0] == vtxIndex[1];
		bool bottomEdgeDegenerate = vtxIndex[2] == vtxIndex[3];

		if (topEdgeDegenerate)
		{
			if (!bottomEdgeDegenerate)
			{
				mesh->addFace( vtxIndex[0], tex[ texVtxIndex[0] ],
									vtxIndex[2], tex[ texVtxIndex[2] ],
									vtxIndex[3], tex[ texVtxIndex[3] ], 0 );
			}
		}
		else
		{
			if (bottomEdgeDegenerate)
			{
				mesh->addFace( vtxIndex[0], tex[ texVtxIndex[0] ],
									vtxIndex[1], tex[ texVtxIndex[1] ],
									vtxIndex[2], tex[ texVtxIndex[2] ], 0 );
			}
			else
			{
				mesh->addFace( vtxIndex[0], tex[ texVtxIndex[0] ],
									vtxIndex[1], tex[ texVtxIndex[1] ],
									vtxIndex[2], tex[ texVtxIndex[2] ],
									vtxIndex[3], tex[ texVtxIndex[3] ], 0 );
			}
		}
	}

	delete o;

	return (JournalObject*)jMesh;
}







//**************************************************


ToolLoftPolySpin::ToolLoftPolySpin(CadView *v)
			: ProcessTool( v, "Spin", "---Spin Polyline---", 2 )
{
	flattenCheck = gtk_check_button_new_with_label("Flatten Polyline");
	gtk_widget_ref(flattenCheck);
	gtk_widget_show(flattenCheck);
	gtk_signal_connect( GTK_OBJECT(flattenCheck), "clicked",
							  GTK_SIGNAL_FUNC(flattenClicked), this );
	
	segmentsLabel = gtk_label_new("Segments:");
	gtk_widget_ref(segmentsLabel);
	gtk_widget_show(segmentsLabel);


	segmentsEntry = new NumericEntry( 16.0, 1.0, 1048576.0, 1.0, 32.0,
													  0.3, 0,
													  GTK_SIGNAL_FUNC(segmentsChanged),
													  this );
	gtk_widget_show( segmentsEntry->getWidget() );
	
	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);

	gtk_signal_connect( GTK_OBJECT(doneButton), "clicked",
							  GTK_SIGNAL_FUNC(doneClicked), this );


	layout.addAsLine(flattenCheck);
	layout.add(segmentsLabel);
	layout.add( segmentsEntry->getWidget() );
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolLoftPolySpin::~ToolLoftPolySpin()
{
	delete segmentsEntry;

	gtk_widget_unref(flattenCheck);
	gtk_widget_unref(segmentsLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolLoftPolySpin::getProcessType() const
{
	return ProcLoftPolySpin::getTypeStatic();
}



void ToolLoftPolySpin::initialise()
{
	ProcessTool::initialise();

	if ( hasProcess() )
	{
		ProcLoftPolySpin *proc = (ProcLoftPolySpin*)journalProcess;

		bool flatten = proc->getFlatten();
	
		segmentsEntry->setValue( proc->getSpinSegments() );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(flattenCheck), flatten );
	}
}



void ToolLoftPolySpin::segmentsChanged(NumericEntry *widget,
														 ToolLoftPolySpin *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolySpin *proc = (ProcLoftPolySpin*)klass->journalProcess;
		proc->setSpinSegments( klass->segmentsEntry->getValueAsInt() );
		klass->modified();
	}
}

void ToolLoftPolySpin::flattenClicked(GtkWidget *widget, ToolLoftPolySpin *klass)
{
	if ( klass->hasProcess() )
	{
		ProcLoftPolySpin *proc = (ProcLoftPolySpin*)klass->journalProcess;
		
		bool flatten = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(klass->flattenCheck) ) != 0;

		proc->setFlatten(flatten);
		klass->modified();
	}
}

void ToolLoftPolySpin::doneClicked(GtkWidget *widget, ToolLoftPolySpin *klass)
{
	klass->done();
}



#endif

