/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __BBOX_CPP
#define __BBOX_CPP

#include "BBox.h"



BBox::BBox()
{
}

BBox::BBox(const Point3& p)
{
	l = u = p;
}

BBox::BBox(const Point3& lower, const Point3& upper)
{
	if (lower.x < upper.x)
	{
		l.x = lower.x;
		u.x = upper.x;
	}
	else
	{
		u.x = lower.x;
		l.x = upper.x;
	}
	if (lower.y < upper.y)
	{
		l.y = lower.y;
		u.y = upper.y;
	}
	else
	{
		u.y = lower.y;
		l.y = upper.y;
	}
	if (lower.z < upper.z)
	{
		l.z = lower.z;
		u.z = upper.z;
	}
	else
	{
		u.z = lower.z;
		l.z = upper.z;
	}
}

BBox::BBox(const array<Point3>& pts)
{
	if (pts.size() == 0)
	{
		return;
	}

	l = u = pts[0];

	for (int i = 1; i < pts.size(); i++)
	{
		addPoint(pts[i]);
	}
}

BBox::BBox(const vector<Point3>& pts)
{
	if (pts.size() == 0)
	{
		return;
	}

	l = u = pts[0];

	for (int i = 1; i < (int)pts.size(); i++)
	{
		addPoint(pts[i]);
	}
}


void BBox::addPoint(const Point3& p)
{
	if (p.x < l.x)
	{
		l.x = p.x;
	}
	if (p.x > u.x)
	{
		u.x = p.x;
	}

	if (p.y < l.y)
	{
		l.y = p.y;
	}
	if (p.y > u.y)
	{
		u.y = p.y;
	}

	if (p.z < l.z)
	{
		l.z = p.z;
	}
	if (p.z > u.z)
	{
		u.z = p.z;
	}
}

bool BBox::check(const Point3& p) const
{
	return (p.x >= l.x) && (p.x <= u.x) &&
			 (p.y >= l.y) && (p.y <= u.y) &&
			 (p.z >= l.z) && (p.z <= u.z);
}

int BBox::check(const Plane& p) const
{
	bool first, cur;
	int i;

	first = getPoint(0) > p;

	for (i = 1; i < 8; i++)
	{
		cur = getPoint(i) > p;
		if (cur != first)
		{
			return 0;
		}
	}

	if (first)
	{
		return 1;
	}
	else
	{
		return -1;
	}
}

bool BBox::intersect(const Line3& ln) const
{
	Point3 ip;
	double frac;

	//try intersecting with planes perpendicular to the x-axis first
	double one_over_dx = 1.0 / (ln.e.x - ln.s.x);
	//see if the ln.s and ln.e are on opposite sides of the plane
	if ((ln.s.x > l.x) != (ln.e.x > l.x))
	{
		//get the intersection point in ip
		frac = (l.x - ln.s.x) * one_over_dx;
		ip = ln.s + (ln.e - ln.s) * frac;
		//check to see if it is within the bounds of the bounding box
		if (ip.y >= l.y  &&  ip.y <= u.y  &&  ip.z >= l.z  &&  ip.z <= u.z)
		{
			return true;
		}
	}
	if ((ln.s.x > u.x) != (ln.e.x > u.x))
	{
		frac = (u.x - ln.s.x) * one_over_dx;
		ip = ln.s + (ln.e - ln.s) * frac;
		if (ip.y >= l.y  &&  ip.y <= u.y  &&  ip.z >= l.z  &&  ip.z <= u.z)
		{
			return true;
		}
	}

	//planes perpendicular to y-axis
	double one_over_dy = 1.0 / (ln.e.y - ln.s.y);
	if ((ln.s.y > l.y) != (ln.e.y > l.y))
	{
		frac = (l.y - ln.s.y) * one_over_dy;
		ip = ln.s + (ln.e - ln.s) * frac;
		if (ip.x >= l.x  &&  ip.x <= u.x  &&  ip.z >= l.z  &&  ip.z <= u.z)
		{
			return true;
		}
	}
	if ((ln.s.y > u.y) != (ln.e.y > u.y))
	{
		frac = (u.y - ln.s.y) * one_over_dy;
		ip = ln.s + (ln.e - ln.s) * frac;
		if (ip.x >= l.x  &&  ip.x <= u.x  &&  ip.z >= l.z  &&  ip.z <= u.z)
		{
			return true;
		}
	}

	//planes perpendicular to z-axis
	double one_over_dz = 1.0 / (ln.e.z - ln.s.z);
	if ((ln.s.z > l.z) != (ln.e.z > l.z))
	{
		frac = (l.z - ln.s.z) * one_over_dz;
		ip = ln.s + (ln.e - ln.s) * frac;
		if (ip.x >= l.x  &&  ip.x <= u.x  &&  ip.y >= l.y  &&  ip.y <= u.y)
		{
			return true;
		}
	}
	if ((ln.s.z > u.z) != (ln.e.z > u.z))
	{
		frac = (u.z - ln.s.z) * one_over_dz;
		ip = ln.s + (ln.e - ln.s) * frac;
		if (ip.x >= l.x  &&  ip.x <= u.x  &&  ip.y >= l.y  &&  ip.y <= u.y)
		{
			return true;
		}
	}

	return false;
}

Point3 BBox::getLower() const
{
	return l;
}

Point3 BBox::getUpper() const
{
	return u;
}

Point3 BBox::getCentre() const
{
	return Point3((l.x + u.x) * 0.5,
					  (l.y + u.y) * 0.5,
					  (l.z + u.z) * 0.5);
}

Point3 BBox::getPoint(int i) const
{
	switch(i)
	{
	case 0:
		return l;
	case 1:
		return Point3(u.x, l.y, l.z);
	case 2:
		return Point3(l.x, u.y, l.z);
	case 3:
		return Point3(u.x, u.y, l.z);
	case 4:
		return Point3(l.x, l.y, u.z);
	case 5:
		return Point3(u.x, l.y, u.z);
	case 6:
		return Point3(l.x, u.y, u.z);
	case 7:
		return u;
	default:
		return l;
	}
}


double BBox::getWidth() const
{
	return u.x - l.x;
}

double BBox::getHeight() const
{
	return u.y - l.y;
}

double BBox::getDepth() const
{
	return u.z - l.z;
}



#endif
