/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#include "Matrix4.h"


//default constructor : initialise to identity matrix
Matrix4::Matrix4()
{
	iM00 = 1.0;	iM01 = 0.0;	iM02 = 0.0;	iM03 = 0.0;
	iM10 = 0.0;	iM11 = 1.0;	iM12 = 0.0;	iM13 = 0.0;
	iM20 = 0.0;	iM21 = 0.0;	iM22 = 1.0;	iM23 = 0.0;
	iM30 = 00.;	iM31 = 0.0;	iM32 = 0.0;	iM33 = 1.0;
}

//constructor which takes the matrix contents as parameters
Matrix4::Matrix4(double m00, double m01, double m02, double m03,
					  double m10, double m11, double m12, double m13,
					  double m20, double m21, double m22, double m23,
					  double m30, double m31, double m32, double m33)
{
	iM00 = m00;	iM01 = m01;	iM02 = m02;	iM03 = m03;
	iM10 = m10;	iM11 = m11;	iM12 = m12;	iM13 = m13;
	iM20 = m20;	iM21 = m21;	iM22 = m22;	iM23 = m23;
	iM30 = m30;	iM31 = m31;	iM32 = m32;	iM33 = m33;
}

//constructors which take four vectors for use as base vectors
Matrix4::Matrix4(const Vector3& i, const Vector3& j, const Vector3& k,
					  const Vector3& xlat)
{
/*
i.x		i.y		i.z		0
j.x		j.y		j.z		0
k.x		k.y		k.z		0
xlat.x	xlat.y	xlat.z	1
*/
	iM00 = i.x;		iM01 = i.y;		iM02 = i.z;		iM03 = 0.0;
	iM10 = j.x;		iM11 = j.y;		iM12 = j.z;		iM13 = 0.0;
	iM20 = k.x;		iM21 = k.y;		iM22 = k.z;		iM23 = 0.0;
	iM30 = xlat.x;	iM31 = xlat.y;	iM32 = xlat.z;	iM33 = 1.0;
}

//constructors which take four vectors for use as base vectors
Matrix4::Matrix4(const Point3& o, const Point3& oi, const Point3& oj,
					  const Point3& ok)
{
	Vector3 i = oi - o;
	Vector3 j = oj - o;
	Vector3 k = ok - o;
	Vector3 xlat = o.toVector3();
/*
i.x		i.y		i.z		0
j.x		j.y		j.z		0
k.x		k.y		k.z		0
xlat.x	xlat.y	xlat.z	1
*/
	iM00 = i.x;		iM01 = i.y;		iM02 = i.z;		iM03 = 0.0;
	iM10 = j.x;		iM11 = j.y;		iM12 = j.z;		iM13 = 0.0;
	iM20 = k.x;		iM21 = k.y;		iM22 = k.z;		iM23 = 0.0;
	iM30 = xlat.x;	iM31 = xlat.y;	iM32 = xlat.z;	iM33 = 1.0;
}

//copies the contents of the matrix over
Matrix4& Matrix4::operator=(const Matrix4& m)
{
	iM00 = m.iM00;	iM01 = m.iM01;	iM02 = m.iM02;	iM03 = m.iM03;
	iM10 = m.iM10;	iM11 = m.iM11;	iM12 = m.iM12;	iM13 = m.iM13;
	iM20 = m.iM20;	iM21 = m.iM21;	iM22 = m.iM22;	iM23 = m.iM23;
	iM30 = m.iM30;	iM31 = m.iM31;	iM32 = m.iM32;	iM33 = m.iM33;
	return *this;
}

//matrix addition : add each element
Matrix4 Matrix4::operator+(const Matrix4& m) const
{
	return Matrix4(
		iM00+m.iM00,	iM01+m.iM01,	iM02+m.iM02,	iM03+m.iM03,
		iM10+m.iM10,	iM11+m.iM11,	iM12+m.iM12,	iM13+m.iM13,
		iM20+m.iM20,	iM21+m.iM21,	iM22+m.iM22,	iM23+m.iM23,
		iM30+m.iM30,	iM31+m.iM31,	iM32+m.iM32,	iM33+m.iM33);
}

//matrix subtraction : subtract each element
Matrix4 Matrix4::operator-(const Matrix4& m) const
{
	return Matrix4(
		iM00-m.iM00,	iM01-m.iM01,	iM02-m.iM02,	iM03-m.iM03,
		iM10-m.iM10,	iM11-m.iM11,	iM12-m.iM12,	iM13-m.iM13,
		iM20-m.iM20,	iM21-m.iM21,	iM22-m.iM22,	iM23-m.iM23,
		iM30-m.iM30,	iM31-m.iM31,	iM32-m.iM32,	iM33-m.iM33);
}

//matrix multiplication : standard multiplication of two 4x4 matrices
Matrix4 Matrix4::operator*(const Matrix4& m) const
{
	return Matrix4(
		iM00*m.iM00 +	iM01*m.iM10 +	iM02*m.iM20 +	iM03*m.iM30,
		iM00*m.iM01 +	iM01*m.iM11 +	iM02*m.iM21 +	iM03*m.iM31,
		iM00*m.iM02 +	iM01*m.iM12 +	iM02*m.iM22 +	iM03*m.iM32,
		iM00*m.iM03 +	iM01*m.iM13 +	iM02*m.iM23 +	iM03*m.iM33,

		iM10*m.iM00 +	iM11*m.iM10 +	iM12*m.iM20 +	iM13*m.iM30,
		iM10*m.iM01 +	iM11*m.iM11 +	iM12*m.iM21 +	iM13*m.iM31,
		iM10*m.iM02 +	iM11*m.iM12 +	iM12*m.iM22 +	iM13*m.iM32,
		iM10*m.iM03 +	iM11*m.iM13 +	iM12*m.iM23 +	iM13*m.iM33,

		iM20*m.iM00 +	iM21*m.iM10 +	iM22*m.iM20 +	iM23*m.iM30,
		iM20*m.iM01 +	iM21*m.iM11 +	iM22*m.iM21 +	iM23*m.iM31,
		iM20*m.iM02 +	iM21*m.iM12 +	iM22*m.iM22 +	iM23*m.iM32,
		iM20*m.iM03 +	iM21*m.iM13 +	iM22*m.iM23 +	iM23*m.iM33,

		iM30*m.iM00 +	iM31*m.iM10 +	iM32*m.iM20 +	iM33*m.iM30,
		iM30*m.iM01 +	iM31*m.iM11 +	iM32*m.iM21 +	iM33*m.iM31,
		iM30*m.iM02 +	iM31*m.iM12 +	iM32*m.iM22 +	iM33*m.iM32,
		iM30*m.iM03 +	iM31*m.iM13 +	iM32*m.iM23 +	iM33*m.iM33);
}

//negate the matrix
Matrix4 Matrix4::operator-() const
{
	return Matrix4(-iM00,	-iM01,	-iM02,	-iM03,
						 -iM10,	-iM11,	-iM12,	-iM13,
						 -iM20,	-iM21,	-iM22,	-iM23,
						 -iM30,	-iM31,	-iM32,	-iM33);
}

//inverse of the matrix
Matrix4 Matrix4::operator~() const
{
	//determinants of 2x2 submatrices
	double kplo = iMK*iMP-iML*iMO,	jpln = iMJ*iMP-iML*iMN,	jokn = iMJ*iMO-iMK*iMN;
	double iplm = iMI*iMP-iML*iMM,	iokm = iMI*iMO-iMK*iMM,	injm = iMI*iMN-iMJ*iMM;
	double gpho = iMG*iMP-iMH*iMO,	ifhn = iMF*iMP-iMH*iMN,	fogn = iMF*iMO-iMG*iMN;
	double ephm = iME*iMP-iMH*iMM,	eogm = iME*iMO-iMG*iMM,	enfm = iME*iMN-iMF*iMM;
	double glhk = iMG*iML-iMH*iMK,	flhj = iMF*iML-iMH*iMJ,	fkgj = iMF*iMK-iMG*iMJ;
	double elhi = iME*iML-iMH*iMI,	ekgi = iME*iMK-iMG*iMI,	ejfi = iME*iMJ-iMF*iMI;
	//determinants of 3x3 submatrices
	double d00 = (iMF*kplo - iMG*jpln + iMH*jokn);
	double d01 = (iME*kplo - iMG*iplm + iMH*iokm);
	double d02 = (iME*jpln - iMF*iplm + iMH*injm);
	double d03 = (iME*jokn - iMF*iokm + iMG*injm);

	double d10 = (iMB*kplo - iMC*jpln + iMD*jokn);
	double d11 = (iMA*kplo - iMC*iplm + iMD*iokm);
	double d12 = (iMA*jpln - iMB*iplm + iMD*injm);
	double d13 = (iMA*jokn - iMB*iokm + iMC*injm);

	double d20 = (iMB*gpho - iMC*ifhn + iMD*fogn);
	double d21 = (iMA*gpho - iMC*ephm + iMD*eogm);
	double d22 = (iMA*ifhn - iMB*ephm + iMD*enfm);
	double d23 = (iMA*fogn - iMB*eogm + iMC*enfm);

	double d30 = (iMB*glhk - iMC*flhj + iMD*fkgj);
	double d31 = (iMA*glhk - iMC*elhi + iMD*ekgi);
	double d32 = (iMA*flhj - iMB*elhi + iMD*ejfi);
	double d33 = (iMA*fkgj - iMB*ekgi + iMC*ejfi);
	//reciprocal of the determinant of the 4x4
	double dr = 1.0 / (iMA*d00 - iMB*d01 + iMC*d02 - iMD*d03);

	return Matrix4( d00*dr,		-d10*dr,		 d20*dr,	-d30*dr,
						 -d01*dr,	 d11*dr,		-d21*dr,	 d31*dr,
						  d02*dr,	-d12*dr,		 d22*dr,	-d32*dr,
						 -d03*dr,	 d13*dr,		-d23*dr,	 d33*dr);
}

//matrix determinant
double Matrix4::determinant() const
{
	//determinants of 2x2 submatrices
	double kplo = iMK*iMP-iML*iMO,	jpln = iMJ*iMP-iML*iMN,	jokn = iMJ*iMO-iMK*iMN;
	double iplm = iMI*iMP-iML*iMM,	iokm = iMI*iMO-iMK*iMM,	injm = iMI*iMN-iMJ*iMM;
	//determinants of 3x3 submatrices
	double d00 = iMA * (iMF*kplo - iMG*jpln + iMH*jokn);
	double d01 = iMB * (iME*kplo - iMG*iplm + iMH*iokm);
	double d02 = iMC * (iME*jpln - iMF*iplm + iMH*injm);
	double d03 = iMD * (iME*jokn - iMF*iokm + iMG*injm);
	//determinant of the 4x4
	return (iMA*d00 - iMB*d01 + iMC*d02 - iMD*d03);
}

Vector3 Matrix4::transformUnit(const Vector3& v, double minLenSquared,
										 double maxLenSquared)
{
	Vector3 r = v * (*this);
	double l = r.lengthSquared();
	if ((l < minLenSquared) || (l > maxLenSquared))
	{
		r.normalise();
	}
	
	return r;
}





Point3 operator*(const Point3& p, const Matrix4& m)
{
	return Point3(
		p.x * m.iM00 +	p.y * m.iM10 +	p.z * m.iM20 +	m.iM30,
		p.x * m.iM01 +	p.y * m.iM11 +	p.z * m.iM21 +	m.iM31,
		p.x * m.iM02 +	p.y * m.iM12 +	p.z * m.iM22 +	m.iM32 );
}

//multiply the matrix by a vector (transform the vector)
Vector3 operator*(const Vector3& v, const Matrix4& m)
{
	return Vector3( 
		v.x * m.iM00 +	v.y * m.iM10 +	v.z * m.iM20,
		v.x * m.iM01 + v.y * m.iM11 +	v.z * m.iM21,
		v.x * m.iM02 +	v.y * m.iM12 +	v.z * m.iM22 );
}

Plane operator*(const Plane& p, const Matrix4& m)
{
	Point3 a = Point3(p.n * p.d) * m;
	Vector3 n = (p.n * m).getNormalised();
	return Plane(n, a);
}


Point3& operator*=(Point3& p, const Matrix4& m)
{
	double x =	p.x * m.iM00 +	p.y * m.iM10 +	p.z * m.iM20 +	m.iM30;
	double y =	p.x * m.iM01 +	p.y * m.iM11 +	p.z * m.iM21 +	m.iM31;
	p.z =			p.x * m.iM02 +	p.y * m.iM12 +	p.z * m.iM22 +	m.iM32;
	p.x = x;
	p.y = y;
	return p;
}

Vector3& operator*=(Vector3& v, const Matrix4& m)
{
	double x =	v.x * m.iM00 +	v.y * m.iM10 +	v.z * m.iM20;
	double y =	v.x * m.iM01 +	v.y * m.iM11 +	v.z * m.iM21;
	v.z =			v.x * m.iM02 +	v.y * m.iM12 +	v.z * m.iM22;
	v.x = x;
	v.y = y;
	return v;
}

Plane& operator*=(Plane& p, const Matrix4& m)
{
	Point3 pointOnPlane = Point3( p.n * p.d );
	pointOnPlane *= m;
	p.n *= m;
	p.n.normalise();
	p.d = pointOnPlane.dotProduct( p.n );
	return p;
}





Matrix4_translate::Matrix4_translate(const Vector3& t)
{
/*
1		0		0		0
0		1		0		0
0		0		1		0
t.x	t.y	t.z	1
*/
	*((Matrix4*)this) = Matrix4(
		1.0,	0.0,	0.0,	0.0,
		0.0,	1.0,	0.0,	0.0,
		0.0,	0.0,	1.0,	0.0,
		t.x,	t.y,	t.z,	1.0);
}


Matrix4_scale::Matrix4_scale(const Vector3& s)
{	
/*
s.x	0		0		0
0		s.y	0		0
0		0		s.z	0
0		0		0		1
*/
	*((Matrix4*)this) = Matrix4(
		s.x,	0.0,	0.0,	0.0,
		0.0,	s.y,	0.0,	0.0,
		0.0,	0.0,	s.z,	0.0,
		0.0,	0.0,	0.0,	1.0);
}


Matrix4_rotateX::Matrix4_rotateX(double a)
{
/*
let s = sin(a)
	 c = cos(a)
  
1		0		0		0
0		c		s		0
0		-s		c		0
0		0		0		1
*/
	double s = (double)sin(a);
	double c = (double)cos(a);
	*((Matrix4*)this) = Matrix4(
		1.0,	0.0,	0.0,	0.0,
		0.0,	c,		s,		0.0,
		0.0,	-s,	c,		0.0,
		0.0,	0.0,	0.0,	1.0);
}

Matrix4_rotateY::Matrix4_rotateY(double a)
{
/*
let s = sin(a)
	 c = cos(a)
  
c		0		-s		0
0		1		0		0
s		0		c		0
0		0		0		1
*/
	double s = (double)sin(a);
	double c = (double)cos(a);
	*((Matrix4*)this) = Matrix4(
		c,		0.0,	-s,	0.0,
		0.0,	1.0,	0.0,	0.0,
		s,		0.0,	c,		0.0,
		0.0,	0.0,	0.0,	1.0);
}

Matrix4_rotateZ::Matrix4_rotateZ(double a)
{
/*
let s = sin(a)
	 c = cos(a)
  
c		s		0		0
-s		c		0		0
0		0		1		0
0		0		0		1
*/
	double s = (double)sin(a);
	double c = (double)cos(a);
	*((Matrix4*)this) = Matrix4(
		c,		s,		0.0,	0.0,
		-s,	c,		0.0,	0.0,
		0.0,	0.0,	1.0,	0.0,
		0.0,	0.0,	0.0,	1.0);
}

Matrix4_rotate::Matrix4_rotate(const Vector3& axis, double a)
{
 	double w, x, y, z;
   double xx, yy, zz, xy, xz, yz, wx, wy, wz;
   double s;

	a *= 0.5;
	s = (double)sin(a);
	w = (double)cos(a);
	x = axis.x * s;
	y = axis.y * s;
	z = axis.z * s;

   xx = 2.0*x*x;	yy = 2.0*y*y;	zz = 2.0*z*z;
   xy = 2.0*x*y;	xz = 2.0*x*z;	yz = 2.0*y*z;
   wx = 2.0*w*x;	wy = 2.0*w*y;	wz = 2.0*w*z;

/*
the quaternion q:
						w=sin(a/2)
						x=axis.x*cos(a/2)
						y=axis.y*cos(a/2)
						z=axis.z*cos(a/2) represents the rotation
						
matrix:
1 - 2yy - 2zz		2xy + 2wz			2xz - 2wy			0
2xy - 2wz			1 - 2xx - 2zz		2yz + 2wx			0
2xz + 2wy			2yz - 2wx			1 - 2xx - 2yy		0
0						0						0						1
*/
	*((Matrix4*)this) = Matrix4(
		1.0 - yy - zz,		xy + wz,				xz - wy,				0.0,
		xy - wz,				1.0 - xx - zz,		yz + wx,				0.0,
		xz + wy,				yz - wx,				1.0 - xx - yy,		0.0,
		0.0,					0.0,					0.0,					1.0);
}

/*Matrix4_rotate::Matrix4_rotate(const MQuaternion& q)
{
   double xx, yy, zz, xy, xz, yz, wx, wy, wz;

   xx = 2*q.x*q.x;	yy = 2*q.y*q.y;	zz = 2*q.z*q.z;
	xy = 2*q.x*q.y;	xz = 2*q.x*q.z;	yz = 2*q.y*q.z;
   wx = 2*q.w*q.x;	wy = 2*q.w*q.y;	wz = 2*q.w*q.z;*/

/*
matrix:
1 - 2yy - 2zz		2xy + 2wz			2xz - 2wy			0
2xy - 2wz			1 - 2xx - 2zz		2yz + 2wx			0
2xz + 2wy			2yz - 2wx			1 - 2xx - 2yy		0
0						0						0						1
*/
/*	*((Matrix4*)this) = Matrix4(
		1.0 - yy - zz,		xy + wz,			xz - wy,			0.0,
		xy - wz,				1 - xx - zz,	yz + wx,			0.0,
		xz + wy,				yz - wx,			1 - xx - yy,	0.0,
		0.0,					0.0,				0.0,				1.0);
}*/


Matrix4_shearX::Matrix4_shearX(double yShear, double zShear)
{/*
1			sy/sx		sz/sx		0
0			1			0			0
0			0			1			0
0			0			0			1
*/
	*((Matrix4*)this) = Matrix4(
			1.0,			yShear,		zShear,		0.0,
			0.0,			1.0,			0.0,			0.0,
			0.0,			0.0,			1.0,			0.0,
			0.0,			0.0,			0.0,			1.0);
}

Matrix4_shearY::Matrix4_shearY(double xShear, double zShear)
{
/*
1			0			0			0
sx/sy		1			sz/sy		0
0			0			1			0
0			0			0			1
*/
	*((Matrix4*)this) = Matrix4(
			1.0,			0.0,			0.0,			0.0,
			xShear,		1.0,			zShear,		0.0,
			0.0,			0.0,			1.0,			0.0,
			0.0,			0.0,			0.0,			1.0);
}

Matrix4_shearZ::Matrix4_shearZ(double xShear, double yShear)
{
/*
1			0			0			0
0			1			0			0
sx/sz		sy/sz		1			0
0			0			0			1
*/
	*((Matrix4*)this) = Matrix4(
			1.0,			0.0,			0.0,			0.0,
			0.0,			1.0,			0.0,			0.0,
			xShear,		yShear,		1.0,			0.0,
			0.0,			0.0,			0.0,			1.0);
}


Matrix4_reflect::Matrix4_reflect(const Vector3& n)
{
	double nx = -2.0 * n.x, ny = -2.0 * n.y, nz = -2.0 * n.z;

/*
1 - 2xx		-2xy			-2xz			0
-2yx			1 - 2yy		-2yz			0
-2zx			-2zy			1 - 2zz		0
0				0				0				1
*/
	*((Matrix4*)this) = Matrix4(
		1 + nx * n.x,		nx * n.y, 			nx * n.z,			0.0,
		ny * n.x,			1 + ny * n.y,		ny * n.z,			0.0,
		nz * n.x,			nz * n.y,			1 + nz * n.z,		0.0,
		0.0,					0.0,					0.0,					1.0);
}

Matrix4_reflect::Matrix4_reflect(const Plane& p)
{
	double nx = -2 * p.n.x, ny = -2 * p.n.y, nz = -2 * p.n.z;
/*
1 - 2xx		-2xy			-2xz			0
-2yx			1 - 2yy		-2yz			0
-2zx			-2zy			1 - 2zz		0
2dx			2dy			2dz			1
*/
	*((Matrix4*)this) = Matrix4(
		1 + nx * p.n.x,	nx * p.n.y, 		nx * p.n.z,			0.0,
		ny * p.n.x,			1 + ny * p.n.y,	ny * p.n.z,			0.0,
		nz * p.n.x,			nz * p.n.y,			1 + nz * p.n.z,	0.0,
		nx * p.d,			ny * p.d,			nz * p.d,			1.0);
}

Matrix4_spaceToSpace::Matrix4_spaceToSpace(const Vector3& i, const Vector3& j,
															const Vector3& k, const Vector3& u,
															const Vector3& v, const Vector3& w)
{
	//inverse of matric ijk multiplied by matrix uvw
	Matrix4 temp = ~Matrix4(i,j,k) * Matrix4(u,v,w);
	*this = *((Matrix4_spaceToSpace*)&temp);
}

Matrix4_axisToAxis::Matrix4_axisToAxis(const Vector3& src, const Vector3& dst)
{
	//find the angle between the vectors
	//and generate a rotation matrix, rotating that amount, where the axis
	//of rotation is the vector that is perpendicular to src and dst
	double cosAngle = src.dotProduct(dst);
	
	//check if src and dst are the same vector
	if ( cosAngle == 1.0 )
	{
		//if so, identity
		*((Matrix4*)this) = Matrix4();
	}
	else
	{
		//calculate rotation axis
		Vector3 axis = src.crossProduct(dst);
		//calculate angle
		double angle = acos(cosAngle);
		//generate the matrix
		Matrix4 temp = Matrix4_rotate(axis, angle);
		//convert to Matrix4_axisToAxis type
		*this = *((Matrix4_axisToAxis*)&temp);
	}
}


