/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
/*
	by:					Geoffrey French
	application:		gSculpt
	last revision:		11/Mar/2001
	title:				4x4 transformation matrix class with mathematical operations.
	puspose:				Transforming vectors for the purpose of 3d engines. Designed
							for speed

  Thanks to Dr. Gavin Cawley for the idea of deriving matrix classes to allow
  constructors with different names. See below.

  Implemented as follows:

  use the macros iMA, iMB,.... iMP to access the members by letter
  use the macros iMyx where x and y are numbers in the range 0..3 to access the
   members where x and y are the coordinates of the matrix member

  the matrices are implemented so that vectors are represented as ROW vectors,
   rather than columns
*/
#ifndef MATRIX4H
#define MATRIX4H

#include <math.h>

#include "radians.h"

#include "Point3.h"
#include "Vector3.h"
#include "Plane.h"



//a selection of macros to make acessing the members of the matrix easy
#define iMA d[0]
#define iMB d[1]
#define iMC d[2]
#define iMD d[3]
#define iME d[4]
#define iMF d[5]
#define iMG d[6]
#define iMH d[7]
#define iMI d[8]
#define iMJ d[9]
#define iMK d[10]
#define iML d[11]
#define iMM d[12]
#define iMN d[13]
#define iMO d[14]
#define iMP d[15]

#define iM00 d[0]
#define iM01 d[1]
#define iM02 d[2]
#define iM03 d[3]
#define iM10 d[4]
#define iM11 d[5]
#define iM12 d[6]
#define iM13 d[7]
#define iM20 d[8]
#define iM21 d[9]
#define iM22 d[10]
#define iM23 d[11]
#define iM30 d[12]
#define iM31 d[13]
#define iM32 d[14]
#define iM33 d[15]



class Matrix4
{
public:
	double d[16];					//16 (4x4) floats

	//default constructor
	Matrix4();
	//constructor taking the data directly
	Matrix4(double m00, double m01, double m02, double m03, double m10, double m11, double m12, double m13,
				double m20, double m21, double m22, double m23, double m30, double m31, double m32, double m33);
	//construct using base vectors (2 methods)
	Matrix4(const Vector3& i, const Vector3& j, const Vector3& k,
           const Vector3& xlat = Vector3());
	Matrix4(const Point3& o, const Point3& oi, const Point3& oj,
			  const Point3& ok);

	//mathematical operators
	Matrix4& operator=(const Matrix4& m);
	Matrix4 operator+(const Matrix4& m) const;
	Matrix4 operator-(const Matrix4& m) const;
	Matrix4 operator*(const Matrix4& m) const;
	Matrix4 operator-() const;
	Matrix4 operator~() const;		//inverse

	//determinant
	double determinant() const;

	Vector3 transformUnit(const Vector3& v, double minLenSquared,
								 double maxLenSquared);
};


//multiply matrix by a Point3; transform the point
Point3 operator*(const Point3& p, const Matrix4& m);
//multiply matrix by a Vector3; transform the vector
Vector3 operator*(const Vector3& v, const Matrix4& m);
//multiply matrix by a plane
Plane operator*(const Plane& p, const Matrix4& m);

Point3& operator*=(Point3& p, const Matrix4& m);
Vector3& operator*=(Vector3& v, const Matrix4& m);
Plane& operator*=(Plane& p, const Matrix4& m);


/*
To cope with the different transformations that can be represented by a matrix, constructors
could be used, however, the parameters would be the same for different constructors;
ambiguit would arise. To allow these constructors to be named, classes are derived from
the Matrix4 class.
*/

class Matrix4_translate : public Matrix4		//translation
{
public:
	Matrix4_translate(const Vector3& t);
};


class Matrix4_scale : public Matrix4			//scale
{
public:
	Matrix4_scale(const Vector3& s);
};


class Matrix4_rotateX : public Matrix4			//rotate around x-axis
{
public:
	Matrix4_rotateX(double a);
};

class Matrix4_rotateY : public Matrix4			//rotate around y-axis
{
public:
	Matrix4_rotateY(double a);
};

class Matrix4_rotateZ : public Matrix4			//rotate around z-axis
{
public:
	Matrix4_rotateZ(double a);
};

class Matrix4_rotate : public Matrix4			//rotate around arbitary axis
{
public:
	Matrix4_rotate(const Vector3& axis, double a);
	//Matrix4_rotate(const MQuaternion& q);
};


class Matrix4_shearX : public Matrix4			//shear along x-axis
{
public:
	Matrix4_shearX(double yShear, double zShear);
};

class Matrix4_shearY : public Matrix4			//shear along y-axis
{
public:
	Matrix4_shearY(double xShear, double zShear);
};

class Matrix4_shearZ : public Matrix4			//shear along z-axis
{
public:
	Matrix4_shearZ(double xShear, double yShear);
};


class Matrix4_reflect : public Matrix4			//reflect in plane
{
public:
	Matrix4_reflect(const Vector3& normal);
	Matrix4_reflect(const Plane& plane);
};

class Matrix4_spaceToSpace : public Matrix4	//space to space
{
public:
	Matrix4_spaceToSpace(const Vector3& i, const Vector3& j,
                         const Vector3& k, const Vector3& u,
                         const Vector3& v, const Vector3& w);
};

class Matrix4_axisToAxis : public Matrix4		//axis to axis
{
public:
	Matrix4_axisToAxis(const Vector3& src, const Vector3& dst);
};


#endif


