/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef VECTOR3CPP
#define VECTOR3CPP

#include <math.h>

#include "Vector3.h"



Vector3::Vector3()
{
	x = y = z = 0.0;
}

Vector3::Vector3(double ix, double iy, double iz)
{
	x = ix;
	y = iy;
	z = iz;
}

Vector3::Vector3(Axis a)
{
	switch(a)
	{
	case AXIS_X:
		x = 1.0;
		y = 0.0;
		z = 0.0;
		break;
	case AXIS_Y:
		x = 0.0;
		y = 1.0;
		z = 0.0;
		break;
	case AXIS_Z:
		x = 0.0;
		y = 0.0;
		z = 1.0;
		break;
	default:
		x = 0.0;
		y = 0.0;
		z = 0.0;
		break;
	}
}


void Vector3::read(FILE *f)
{
	fread( &x, sizeof(double), 1, f );
	fread( &y, sizeof(double), 1, f );
	fread( &z, sizeof(double), 1, f );
}

void Vector3::write(FILE *f) const
{
	fwrite( &x, sizeof(double), 1, f );
	fwrite( &y, sizeof(double), 1, f );
	fwrite( &z, sizeof(double), 1, f );
}





bool Vector3::operator==(const Vector3& v) const
{
	return (x == v.x) && (y == v.y) && (z == v.z);
}

bool Vector3::operator!=(const Vector3& v) const
{
	return (x != v.x) || (y != v.y) || (z != v.z);
}


Vector3 Vector3::operator+(const Vector3& v) const
{
	return Vector3(x + v.x, y + v.y, z + v.z);
}

Vector3& Vector3::operator+=(const Vector3& v)
{
	x += v.x;
	y += v.y;
	z += v.z;
	return *this;
}

Vector3 Vector3::operator-(const Vector3& v) const
{
	return Vector3(x - v.x, y - v.y, z - v.z);
}

Vector3& Vector3::operator-=(const Vector3& v)
{
	x -= v.x;
	y -= v.y;
	z -= v.z;
	return *this;
}

Vector3 Vector3::operator*(double f) const
{
	return Vector3(x * f, y * f, z * f);
}

Vector3& Vector3::operator*=(double f)
{
	x *= f;
	y *= f;
	z *= f;
	return *this;
}

Vector3 Vector3::operator-() const
{
	return Vector3(-x, -y, -z);
}


double Vector3::dotProduct(const Vector3& v) const
{
	return (x * v.x) + (y * v.y) + (z * v.z);
}

Vector3 Vector3::crossProduct(const Vector3& v) const
{
	return Vector3(y * v.z - z * v.y,
						z * v.x - x * v.z,
						x * v.y - y * v.x);
}

double Vector3::lengthSquared() const
{
	return dotProduct(*this);
}

double Vector3::length() const
{
	return sqrt( lengthSquared() );
}

double Vector3::manhattanLength() const
{
	double tx = x, ty = y, tz = z;
	if (tx < 0.0)
	{
		tx = -tx;
	}
	if (ty < 0.0)
	{
		ty = -ty;
	}
	if (tz < 0.0)
	{
		tz = -tz;
	}
	return tx + ty + tz;
}

void Vector3::normalise()
{
	double ool = 1.0 / length();
	*this *= ool;
}

void Vector3::normalise(double l)
{
	double lol = l / length();
	*this *= lol;
}

Vector3 Vector3::getNormalised() const
{
	double ool = 1.0 / length();
	return Vector3(x * ool, y * ool, z * ool);
}

Vector3 Vector3::getNormalised(double l) const
{
	double lol = l / length();
	return *this * lol;
}



#endif
