/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __BACKGROUNDOBJMANAGER_CPP
#define __BACKGROUNDOBJMANAGER_CPP

#include "BackgroundObjManager.h"



BackgroundObjManager::Entry::Entry(Journal *journ,
											  ObjPainterManager *painterManager)
{
	journal = journ;

	activeFlag = false;

	//check this history's integrity before proceeding
	if ( journal->checkIntegrity() )
	{
		journalObject = journal->getResultUnthreaded();
		painter = painterManager->createPainter( journalObject->getType() );
		if ( painter != NULL )
		{
			painter->setObject( journalObject );
		}
	}
	else
	{
		journalObject = NULL;
		painter = NULL;
	}
}

BackgroundObjManager::Entry::~Entry()
{
	if ( painter != NULL )
	{
		delete painter;
	}
	delete journalObject;
}


void BackgroundObjManager::Entry::activate(CadView *view)
{
	if (!activeFlag)
	{
		if ( painter != NULL )
		{
			view->addEventHandler(painter);
			painter->inBackground();
			view->redraw();
		}
		activeFlag = true;
	}	
}

void BackgroundObjManager::Entry::deactivate(CadView *view)
{
	if (activeFlag)
	{
		if ( painter != NULL )
		{
			view->removeEventHandler( painter );
			view->redraw();
		}
		activeFlag = false;
	}
}

void BackgroundObjManager::Entry::update(ObjPainterManager *painterManager,
													  CadView *view, bool redrawFlag)
{
	if ( activeFlag  &&  painter != NULL )
	{
		view->removeEventHandler( painter );
	}

	delete journalObject;
	if ( painter != NULL )
	{
		delete painter;
	}


	//check the journal's integrity
	if ( journal->checkIntegrity() )
	{
		journalObject = journal->getResultUnthreaded();
		painter = painterManager->createPainter( journalObject->getType() );
		
		if ( activeFlag  &&  painter != NULL )
		{
			view->addEventHandler( painter );
			painter->setObject( journalObject );
			painter->inBackground();
			if (redrawFlag)
			{
				view->redraw();
			}
		}
	}
	else
	{
		journalObject = NULL;
		painter = NULL;
	}
}




BackgroundObjManager::BackgroundObjManager(CadView *vw,
														 ObjPainterManager *pManager)
{
	view = vw;
	painterManager = pManager;
}


bool BackgroundObjManager::exists(const string& objName)
{
	return entries.count(objName) > 0;
}

void BackgroundObjManager::add(const string& objName, Journal *journ,
										 bool active)
{
	if ( !exists(objName) )
	{
		Entry *e = new Entry(journ, painterManager);
		if (active)
		{
			e->activate(view);
		}
		entries[objName] = e;
	}
}

void BackgroundObjManager::remove(const string& objName)
{
	if ( exists(objName) )
	{
		Entry *e = entries[objName];
		e->deactivate(view);
		delete e;
		entries.erase(objName);
	}
}

void BackgroundObjManager::activate(const string& objName)
{
	if ( exists(objName) )
	{
		entries[objName]->activate(view);
	}	
}
	
void BackgroundObjManager::deactivate(const string& objName)
{
	if ( exists(objName) )
	{
		entries[objName]->deactivate(view);
	}
}

void BackgroundObjManager::update(const string& objName)
{
	if ( exists(objName) )
	{
		entries[objName]->update(painterManager, view, true);
	}
}

void BackgroundObjManager::updateAll()
{
	map<string,Entry*>::iterator iter;

	for (iter = entries.begin(); iter != entries.end(); ++iter)
	{
		(*iter).second->update(painterManager, view, false);
	}

	view->redraw();
}

#endif
