/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __OBJECTCOLLECTION_CPP
#define __OBJECTCOLLECTION_CPP

#include "JournalModel.h"


#include "ObjectCollection.h"
#include "CModel.h"




ObjectCollection::ObjectCollection()
{
	uniqueIdCounter = 0;
}

ObjectCollection::~ObjectCollection()
{
	EntryTable::iterator iter;

	for (iter = objects.begin(); iter != objects.end(); ++iter)
	{
		for (LsManList::iterator lsManIter = listManagers.begin();
			  lsManIter != listManagers.end();
			  ++lsManIter)
		{
			(*lsManIter)->remove( (*iter).second.name );
			delete (*iter).second.journal;
		}
	}
}


void ObjectCollection::read(FILE *f)
{
	UniqueId maximumId = 0;
	
	objects.clear();
	nameToId.clear();
	
	for (LsManList::iterator lsManIter = listManagers.begin();
		  lsManIter != listManagers.end();
		  ++lsManIter)
	{
		(*lsManIter)->clear();
	}
	
	int numObjects;
	fread(&numObjects, sizeof(int), 1, f);

	for (int i = 0; i < numObjects; i++)
	{
		UniqueId id;
		string name;

		//read ID and name
		fread(&id, sizeof(UniqueId), 1, f);
		name = readString(f);

		//maintain maximum ID
		if (maximumId < id)
		{
			maximumId = id;
		}

		//add to nameToId and the list managers
		nameToId[name] = id;
		for (LsManList::iterator lsManIter = listManagers.begin();
			  lsManIter != listManagers.end();
			  ++lsManIter)
		{
			(*lsManIter)->add(name);
		}

		//create new entry and get ref to it
		Entry &entry = objects[id];

		//set name entry
		entry.name = name;

		//read number of dependants
		int numDependants;
		fread(&numDependants, sizeof(int), 1, f);

		//read dependants
		for (int j = 0; j < numDependants; j++)
		{
			UniqueId dependantId;
			fread(&dependantId, sizeof(UniqueId), 1, f);
			entry.dependants.insert(dependantId);
		}

		//read dependancy
		fread( &( entry.dependancy ), sizeof(UniqueId), 1, f );

		//read history
		entry.journal = new Journal(this);
		entry.journal->read(f);
	}

	uniqueIdCounter = maximumId + 1;
}

void ObjectCollection::write(FILE *f)
{
	int numObjects = objects.size();
	fwrite(&numObjects, sizeof(int), 1, f);

	for (	EntryTable::iterator iter = objects.begin();
			iter != objects.end();
			++iter )
	{
		//get a ref to the entry
		Entry &entry = (*iter).second;
		
		UniqueId id = (*iter).first;


		//write ID and name
		fwrite(&id, sizeof(UniqueId), 1, f);
		writeString(f, entry.name);


		//write number of dependants
		int numDependants = entry.dependants.size();
		fwrite(&numDependants, sizeof(int), 1, f);
		
		//write dependants
		for ( set<UniqueId>::const_iterator iter2 = entry.dependants.begin();
				iter2 != entry.dependants.end();
				++iter2 )
		{
			UniqueId dependantId = *iter2;

			fwrite(&dependantId, sizeof(UniqueId), 1, f);
		}

		//write dependancy
		fwrite( &( entry.dependancy ), sizeof(UniqueId), 1, f );

		//write the journal
		entry.journal->write(f);
	}
}

void ObjectCollection::exportExplicit(FILE *f)
{
	int numObjects = 0;

	//note the position of the number of objects, and skip past it
	int positionOfNumberOfObjects = ftell(f);
	fseek( f, sizeof(int), SEEK_CUR );

	for (	EntryTable::iterator iter = objects.begin();
			iter != objects.end();
			++iter )
	{
		//get a ref to the entry
		Entry &entry = (*iter).second;
		
		//check the integrity of the journal
		if ( entry.journal->checkIntegrity() )
		{
			JournalObject *result = entry.journal->getResultUnthreaded();

			//only write it if it is a subclass of JournalModel
			if ( result->getType()->isSubclassOf( JournalModel::getTypeStatic() ) )
			{
				//write the object name
				writeString( f, entry.name );

				//get the type name and write it
				string typeName = result->getType()->getName();
				writeString( f, typeName );

				//write the object
				JournalModel *jModel = (JournalModel*)result;
				CModel *model = jModel->getModel();
				model->write(f);

				//increment the number of objects stored in the file
				numObjects++;
			}

			delete result;
		}
	}

	//note the position of the end
	int endPosition = ftell(f);

	//seek back to the position of the number of objects and write it
	fseek( f, positionOfNumberOfObjects, SEEK_SET );
	fwrite( &numObjects, sizeof(int), 1, f );

	//go to end
	fseek( f, endPosition, SEEK_SET );	
}



bool ObjectCollection::exists(const string& name) const
{
	return nameToId.count(name) > 0;
}

bool ObjectCollection::exists(UniqueId i) const
{
	return objects.count(i) > 0;
}

void ObjectCollection::add(const string& name)
{
	if ( !exists(name) )
	{
		Entry e;
		e.journal = new Journal(this);
		e.name = name;
		e.dependancy = -1;
				
		UniqueId id = createUniqueId();
		
		nameToId[name] = id;
		objects[id] = e;

		for (LsManList::iterator lsManIter = listManagers.begin();
			  lsManIter != listManagers.end();
			  ++lsManIter)
		{
			(*lsManIter)->add(name);
		}
	}
}

void ObjectCollection::remove(const string& name)
{
	if ( exists(name) )
	{
		for (LsManList::iterator lsManIter = listManagers.begin();
			  lsManIter != listManagers.end();
			  ++lsManIter)
		{
			(*lsManIter)->remove(name);
		}

		UniqueId id = nameToId[name];
		nameToId.erase(name);
		
		delete objects[id].journal;
		objects.erase(id);
	}
}

string ObjectCollection::getSelectionName(ListManager *lsMan) const
{
	return lsMan->getSelection();
}

Journal* ObjectCollection::getObject(const string& name)
{
	if ( exists(name) )
	{
		return objects[ nameToId[name] ].journal;
	}
	else
	{
		return NULL;
	}
}

const Journal* ObjectCollection::getObject(const string& name) const
{
	if ( exists(name) )
	{
		IdTable::const_iterator nameIterator = nameToId.find(name);
		UniqueId id = (*nameIterator).second;
		EntryTable::const_iterator idIterator = objects.find(id);
		return (*idIterator).second.journal;
	}
	else
	{
		return NULL;
	}
}

Journal* ObjectCollection::getObject(UniqueId i)
{
	EntryTable::iterator iter = objects.find(i);
	if ( iter != objects.end() )
	{
		return (*iter).second.journal;
	}
	else
	{
		return NULL;
	}
}

const Journal* ObjectCollection::getObject(UniqueId i) const
{
	EntryTable::const_iterator iter = objects.find(i);
	if ( iter != objects.end() )
	{
		return (*iter).second.journal;
	}
	else
	{
		return NULL;
	}
}

Journal* ObjectCollection::getSelection(ListManager *lsMan)
{
	string name = getSelectionName(lsMan);

	return getObject(name);
}



UniqueId ObjectCollection::getUniqueId(const string& name) const
{
	IdTable::const_iterator iter = nameToId.find(name);
	if ( iter != objects.end() )
	{
		return (*iter).second;
	}
	else
	{
		return -1;
	}
}

UniqueId ObjectCollection::getUniqueId(const Journal *journal) const
{
	//linear search for it
	for (EntryTable::const_iterator iter = objects.begin();
		  iter != objects.end();
		  ++iter)
	{
		if ( (*iter).second.journal == journal )
		{
			return (*iter).first;
		}
	}

	return -1;
}

string ObjectCollection::getName(UniqueId i) const
{
	EntryTable::const_iterator iter = objects.find(i);
	if ( iter != objects.end() )
	{
		return (*iter).second.name;
	}
	else
	{
		return "";
	}
}



ListManager* ObjectCollection::getListManager()
{
	ListManager *lsMan = new ListManager();

	for (EntryTable::const_iterator iter = objects.begin();
		  iter != objects.end();
		  ++iter)
	{
		lsMan->add( (*iter).second.name );
	}


	listManagers.push_back(lsMan);
	
	return lsMan;
}

void ObjectCollection::putbackListManager(ListManager *lsMan)
{
	listManagers.remove(lsMan);
	delete lsMan;
}



bool ObjectCollection::hasDependants(UniqueId objectId) const
{
	EntryTable::const_iterator iter = objects.find(objectId);
	if ( iter != objects.end() )
	{
		return (*iter).second.dependants.size() > 0;
	}
	else
	{
		return false;
	}
}

bool ObjectCollection::hasDependant(UniqueId objectId,
												UniqueId dependantId) const
{
	EntryTable::const_iterator iter = objects.find(objectId);
	if (  ( iter != objects.end() ) &&
			( objects.find(dependantId) != objects.end() )  )
	{
		return (*iter).second.dependants.count(dependantId) > 0;
	}
	else
	{
		return false;
	}
}

bool ObjectCollection::hasDependancyPath(UniqueId start, UniqueId end) const
{
	if ( start == end )
	{
		return true;
	}
	
	EntryTable::const_iterator iter = objects.find(end);

	//check if the starting and ending points exist
	if (  ( iter != objects.end() ) &&
			( objects.find(start) != objects.end() )  )
	{
		//loop forever, the loop exits via a return or break statement
		//
		//we start at the end, and work backwards toward the start; from the
		//bottom of the tree to the top (assuming the root node is at the top)
		while(1)
		{
			//advance up one branch
			end = (*iter).second.dependancy;

			if ( end == -1 )
			{
				//reached NULL: exit and return false
				break;
			}
			else if ( end == start )
			{
				//reached the starting position, exit and return true
				return true;
			}
			else
			{
				//loop over
				iter = objects.find(end);
			}
		}
	}
	//didn't find a path or objects did not exist
	return false;
}


void ObjectCollection::setDependancy(UniqueId objectId, UniqueId dependancyId)
{
	EntryTable::iterator iter = objects.find(objectId);
	EntryTable::iterator depIter = objects.find(dependancyId);
	if (  ( iter != objects.end() ) &&
			( depIter != objects.end() )  )
	{
		(*iter).second.dependancy = dependancyId;
		(*depIter).second.dependants.insert(objectId);
	}
}

void ObjectCollection::removeDependancy(UniqueId objectId)
{
	EntryTable::iterator iter = objects.find(objectId);
	if ( iter != objects.end() )
	{
		UniqueId dependancyId = (*iter).second.dependancy;
		EntryTable::iterator depIter = objects.find(dependancyId);

		if ( depIter != objects.end() )
		{
			(*depIter).second.dependants.erase(objectId);
		}
	}
}

bool ObjectCollection::getDependantNames(UniqueId objectId,
													  set<string>& dependantNames) const
{
	EntryTable::const_iterator iter = objects.find(objectId);
	if ( iter != objects.end() )
	{
		const set<UniqueId> &listOfDependantIds = (*iter).second.dependants;

		for ( set<UniqueId>::const_iterator iter = listOfDependantIds.begin();
				iter != listOfDependantIds.end();
				++iter )
		{
			EntryTable::const_iterator nameIter = objects.find( *iter );
			if ( nameIter != objects.end() )		//should always be true
			{
				dependantNames.insert( (*nameIter).second.name );
			}
		}
		
		return true;
	}
	else
	{
		return false;
	}
}



UniqueId ObjectCollection::createUniqueId()
{
	return uniqueIdCounter++;
}




#endif
