/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLPOLYREORDER_CPP
#define __TOOLPOLYREORDER_CPP

#include <strstream>
#include <math.h>

#include <vector>

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkradiobutton.h>


#include "ToolPolyReorder.h"

#include "JournalPolyline.h"

#include "Point3.h"





void* __create_ProcPolyReorder()
{
	return (void*)( new ProcPolyReorder() );
}

ObjectType type_ProcPolyReorder( "ProcPolyReorder", "JournalProcess",
										  __create_ProcPolyReorder );




ProcPolyReorder::ProcPolyReorder() : JournalProcess()
{
	operation = SWAP_CONSECUTIVE;
}

ProcPolyReorder::ProcPolyReorder(ReorderOperation op)
											: JournalProcess()
{
	operation = op;
}

ProcPolyReorder::~ProcPolyReorder()
{
}



void ProcPolyReorder::read(FILE *f)
{
	JournalProcess::read(f);

	fread(&operation, sizeof(ReorderOperation), 1, f);
}

void ProcPolyReorder::write(FILE *f)
{
	JournalProcess::write(f);

	fwrite(&operation, sizeof(ReorderOperation), 1, f);
}





string ProcPolyReorder::getDescription() const
{
	return "Polyline Reorder";
}

const ObjectType* ProcPolyReorder::getType() const
{
	return &type_ProcPolyReorder;
}

const ObjectType* ProcPolyReorder::getTypeStatic()
{
	return &type_ProcPolyReorder;
}



const ObjectType* ProcPolyReorder::getInputType() const
{
	return JournalPolyline::getTypeStatic();
}



JournalProcess* ProcPolyReorder::clone() const
{
	return new ProcPolyReorder();
}



ProcPolyReorder::ReorderOperation ProcPolyReorder::getOperation() const
{
	return operation;
}

void ProcPolyReorder::setOperation(ReorderOperation op)
{
	operation = op;
	modified();
}



const ObjectType* ProcPolyReorder::_getOutputType(
									const ObjectType* inputType) const
{
	return JournalPolyline::getTypeStatic();
}

JournalObject* ProcPolyReorder::_process(JournalObject *o,
													  JournalTaskProgress *progress)
{
	JournalPolyline *hPolyline = (JournalPolyline*)o;
	CPolyline *poly = hPolyline->getPoly();

	CVertexContainer &vertices = poly->vertices();


	vector<CPolyline::VertexType> selectedVertices;

	if ( operation != LINK_AB  &&  operation != LINK_BA )
	{
		for (int i = 0; i < vertices.size(); i++)
		{
			if ( vertices[i].isMarked() )
			{
				selectedVertices.push_back( vertices[i] );
			}
		}
	}


	switch(operation)
	{
	case SWAP_CONSECUTIVE:
		swapConsecutive(selectedVertices);
		break;
	case SWAP_HALVES:
		swapHalves(selectedVertices);
		break;
	case REVERSE:
		reverse(selectedVertices);
		break;
	case LINK_AB:
		linkAB(vertices);
		break;
	case LINK_BA:
		linkBA(vertices);
		break;
	default:
		break;
	}


	if ( operation != LINK_AB  &&  operation != LINK_BA )
	{
		int j = 0;
		for (int i = 0; i < vertices.size(); i++)
		{
			if ( vertices[i].isMarked() )
			{
				vertices[i] = selectedVertices[j];
				j++;
			}
		}
	}

	
	return o;
}


void ProcPolyReorder::swapConsecutive(
			vector<CPolyline::VertexType>& vertices) const
{
	for (int i = 0; i < (int)vertices.size(); i += 2)
	{
		CPolyline::VertexType swap;

		swap = vertices[i];
		vertices[i] = vertices[i+1];
		vertices[i+1] = swap;
	}
}

void ProcPolyReorder::swapHalves(vector<CPolyline::VertexType>& vertices) const
{
	int halfSize = (int)vertices.size() >> 1;
	
	for (int i = 0; i < halfSize; i++)
	{
		CPolyline::VertexType swap;

		swap = vertices[i];
		vertices[i] = vertices[i+halfSize];
		vertices[i+halfSize] = swap;
	}
}

void ProcPolyReorder::reverse(vector<CPolyline::VertexType>& vertices) const
{
	int i = 0;
	int j = vertices.size() - 1;
	while (i < j)
	{
		CPolyline::VertexType swap;

		swap = vertices[i];
		vertices[i] = vertices[j];
		vertices[j] = swap;

		i++;
		j--;
	}
}

void ProcPolyReorder::linkAB(CVertexContainer& vertices) const
{
	int a, b;
	if ( findIndices(vertices, a, b) )
	{
		link(vertices, a, b);
	}
}

void ProcPolyReorder::linkBA(CVertexContainer& vertices) const
{
	int a, b;
	if ( findIndices(vertices, a, b) )
	{
		link(vertices, b, a);
	}
}

bool ProcPolyReorder::findIndices(const CVertexContainer& vertices,
											 int &a, int &b) const
{
	int indices[2];

	int j = 0;

	for (int i = 0; i < (int)vertices.size(); i++)
	{
		if ( vertices[i].isMarked() )
		{
			indices[j++] = i;
			if (j >= 2)
			{
				break;
			}
		}
	}

	if (j >= 2)
	{
		a = indices[0];
		b = indices[1];
		return true;
	}
	else
	{
		return false;
	}
}

void ProcPolyReorder::link(CVertexContainer& vertices, int a, int b) const
{
	int afterA = (a + 1) % vertices.size();

	CVertexContainer::VertexType swap;

	swap = vertices[afterA];
	vertices[afterA] = vertices[b];
	vertices[b] = swap;
}







//**************************************************


ToolPolyReorder::ToolPolyReorder(CadView *v)
			: ProcessTool( v, "Delete", "---Polyline Delete---", 1 )
{
	ignoreRadioEvents = false;

	
	operationLabel = gtk_label_new("Operation:");
	gtk_widget_ref(operationLabel);
	gtk_widget_show(operationLabel);
	

	GSList *group = (GSList*)NULL;
	swapConsecRadio = gtk_radio_button_new_with_label(group, "Swap Consective");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(swapConsecRadio) );
	swapHalvesRadio = gtk_radio_button_new_with_label(group, "Swap Halves");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(swapHalvesRadio) );
	reverseRadio = gtk_radio_button_new_with_label(group, "Reverse");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(reverseRadio) );
	linkABRadio = gtk_radio_button_new_with_label(group, "Link A-B");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(linkABRadio) );
	linkBARadio = gtk_radio_button_new_with_label(group, "Link B-A");

	gtk_widget_ref(swapConsecRadio);
	gtk_widget_ref(swapHalvesRadio);
	gtk_widget_ref(reverseRadio);
	gtk_widget_ref(linkABRadio);
	gtk_widget_ref(linkBARadio);
	gtk_widget_show(swapConsecRadio);
	gtk_widget_show(swapHalvesRadio);
	gtk_widget_show(reverseRadio);
	gtk_widget_show(linkABRadio);
	gtk_widget_show(linkBARadio);

	gtk_signal_connect( GTK_OBJECT(swapConsecRadio), "clicked",
							  GTK_SIGNAL_FUNC(swapConsecClicked), this );
	gtk_signal_connect( GTK_OBJECT(swapHalvesRadio), "clicked",
							  GTK_SIGNAL_FUNC(swapHalvesClicked), this );
	gtk_signal_connect( GTK_OBJECT(reverseRadio), "clicked",
							  GTK_SIGNAL_FUNC(reverseClicked), this );
	gtk_signal_connect( GTK_OBJECT(linkABRadio), "clicked",
							  GTK_SIGNAL_FUNC(linkABClicked), this );
	gtk_signal_connect( GTK_OBJECT(linkBARadio), "clicked",
							  GTK_SIGNAL_FUNC(linkBAClicked), this );




	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);

	gtk_signal_connect( GTK_OBJECT(doneButton), "clicked",
							  GTK_SIGNAL_FUNC(doneClicked), this );


	layout.addAsLine(operationLabel);
	layout.add(swapConsecRadio);
	layout.add(swapHalvesRadio);
	layout.add(reverseRadio);
	layout.add(linkABRadio);
	layout.add(linkBARadio);
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolPolyReorder::~ToolPolyReorder()
{
	gtk_widget_unref(operationLabel);
	gtk_widget_unref(swapConsecRadio);
	gtk_widget_unref(swapHalvesRadio);
	gtk_widget_unref(reverseRadio);
	gtk_widget_unref(linkABRadio);
	gtk_widget_unref(linkBARadio);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolPolyReorder::getProcessType() const
{
	return ProcPolyReorder::getTypeStatic();
}



void ToolPolyReorder::initialise()
{
	ProcessTool::initialise();

	if ( hasProcess() )
	{
		ProcPolyReorder *proc = (ProcPolyReorder*)journalProcess;

		ProcPolyReorder::ReorderOperation operation = proc->getOperation();

		ignoreRadioEvents = true;

		switch(operation)
		{
		case ProcPolyReorder::SWAP_CONSECUTIVE:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(swapConsecRadio),
													true );
			break;
		case ProcPolyReorder::SWAP_HALVES:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(swapHalvesRadio),
													true );
			break;
		case ProcPolyReorder::REVERSE:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(reverseRadio),
													true );
			break;
		default:
			break;
		}

		ignoreRadioEvents = false;
	}
}




void ToolPolyReorder::swapConsecClicked(GtkWidget *widget,
													 ToolPolyReorder *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolyReorder *proc = (ProcPolyReorder*)klass->journalProcess;
		
		proc->setOperation(ProcPolyReorder::SWAP_CONSECUTIVE);
		klass->modified();
	}
}

void ToolPolyReorder::swapHalvesClicked(GtkWidget *widget,
													 ToolPolyReorder *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolyReorder *proc = (ProcPolyReorder*)klass->journalProcess;
		
		proc->setOperation(ProcPolyReorder::SWAP_HALVES);
		klass->modified();
	}
}

void ToolPolyReorder::reverseClicked(GtkWidget *widget,
												 ToolPolyReorder *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolyReorder *proc = (ProcPolyReorder*)klass->journalProcess;
		
		proc->setOperation(ProcPolyReorder::REVERSE);
		klass->modified();
	}
}

void ToolPolyReorder::linkABClicked(GtkWidget *widget,
												ToolPolyReorder *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolyReorder *proc = (ProcPolyReorder*)klass->journalProcess;
		
		proc->setOperation(ProcPolyReorder::LINK_AB);
		klass->modified();
	}
}

void ToolPolyReorder::linkBAClicked(GtkWidget *widget,
												ToolPolyReorder *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolyReorder *proc = (ProcPolyReorder*)klass->journalProcess;
		
		proc->setOperation(ProcPolyReorder::LINK_BA);
		klass->modified();
	}
}

void ToolPolyReorder::doneClicked(GtkWidget *widget, ToolPolyReorder *klass)
{
	klass->done();
}



#endif

