/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLPOLYSUBDIVIDE_CPP
#define __TOOLPOLYSUBDIVIDE_CPP

#include <strstream>
#include <math.h>

#include <vector>

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkradiobutton.h>


#include "JournalPolyline.h"

#include "Point3.h"

#include "ToolPolySubdivide.h"





void* __create_ProcPolySubdivide()
{
	return (void*)( new ProcPolySubdivide() );
}

ObjectType type_ProcPolySubdivide( "ProcPolySubdivide", "JournalProcess",
											  __create_ProcPolySubdivide );




ProcPolySubdivide::ProcPolySubdivide() : JournalProcess()
{
	numberToInsert = 1;
	selectStatus = SELECTSTATUS_SELECTED;
	curveType = CURVETYPE_CATMULROM;
}

ProcPolySubdivide::ProcPolySubdivide(int numberOfVerticesToInsert,
												 SelectStatus selStatus,
												 CurveType curveTp)
										 : JournalProcess()
{
	numberToInsert = numberOfVerticesToInsert;
	selectStatus = selStatus;
	curveType = curveTp;
}

ProcPolySubdivide::~ProcPolySubdivide()
{
}



void ProcPolySubdivide::read(FILE *f)
{
	JournalProcess::read(f);

	fread(&numberToInsert, sizeof(int), 1, f);
	fread(&selectStatus, sizeof(SelectStatus), 1, f);
	fread(&curveType, sizeof(CurveType), 1, f);
}

void ProcPolySubdivide::write(FILE *f)
{
	JournalProcess::write(f);

	fwrite(&numberToInsert, sizeof(int), 1, f);
	fwrite(&selectStatus, sizeof(SelectStatus), 1, f);
	fwrite(&curveType, sizeof(CurveType), 1, f);
}





string ProcPolySubdivide::getDescription() const
{
	return "Polyline Subdivide";
}

const ObjectType* ProcPolySubdivide::getType() const
{
	return &type_ProcPolySubdivide;
}

const ObjectType* ProcPolySubdivide::getTypeStatic()
{
	return &type_ProcPolySubdivide;
}



const ObjectType* ProcPolySubdivide::getInputType() const
{
	return JournalPolyline::getTypeStatic();
}



JournalProcess* ProcPolySubdivide::clone() const
{
	return new ProcPolySubdivide();
}



int ProcPolySubdivide::getNumberOfVerticesToInsert() const
{
	return numberToInsert;
}

ProcPolySubdivide::SelectStatus ProcPolySubdivide::getSelectStatus() const
{
	return selectStatus;
}

ProcPolySubdivide::CurveType ProcPolySubdivide::getCurveType() const
{
	return curveType;
}

void ProcPolySubdivide::setNumberOfVerticesToInsert(int n)
{
	numberToInsert = n;
	modified();
}

void ProcPolySubdivide::setSelectStatus(SelectStatus selStatus)
{
	selectStatus = selStatus;
	modified();
}

void ProcPolySubdivide::setCurveType(CurveType curveTp)
{
	curveType = curveTp;
	modified();
}



const ObjectType* ProcPolySubdivide::_getOutputType(
									const ObjectType* inputType) const
{
	return JournalPolyline::getTypeStatic();
}

JournalObject* ProcPolySubdivide::_process(JournalObject *o,
														 JournalTaskProgress *progress)
{
	JournalPolyline *hPolyline = (JournalPolyline*)o;
	CPolyline *poly = hPolyline->getPoly();

	CVertexContainer &vertices = poly->vertices();

	vector<int> verticesAdded;
	vector<Point3> vertexPositions;



	vertexPositions.reserve( vertices.size() );
	for (int i = 0; i < (int)vertices.size(); i++)
	{
		vertexPositions.push_back( vertices[i].getWorld() );
	}


	int i, j, n;


	j = 1;
	n = 0;
	while (j < vertices.size())
	{
		i = j - 1;

		if ( vertices[i].isMarked() && vertices[j].isMarked() )
		{
			insert( poly, i, verticesAdded, vertexPositions, n );

			j += numberToInsert;
		}
		j++;
		n++;
	}

	if ( poly->isClosed() )
	{
		i = vertices.size() - 1;
		n = vertexPositions.size() - 1;
		j = 0;
		
		if ( vertices[i].isMarked() && vertices[j].isMarked() )
		{
			insert( poly, i, verticesAdded, vertexPositions, n );
		}
	}


	if ( selectStatus == SELECTSTATUS_ONLYSELECTED )
	{
		for (int i = 0; i < vertices.size(); i++)
		{
			vertices[i].unmark();
		}

		for (int i = 0; i < (int)verticesAdded.size(); i++)
		{
			vertices[ verticesAdded[i] ].mark();
		}
	}



	return o;
}

void ProcPolySubdivide::insert(CPolyline *poly, int position,
										 vector<int>& verticesAdded,
										 const vector<Point3>& vertexPositions,
										 int n) const
{
	if ( curveType == CURVETYPE_LINEAR )
	{
		int nPlus1 = ( n + 1 ) % vertexPositions.size();

		Point3 pointAtN = vertexPositions[n];
		Point3 pointAtNPlus1 = vertexPositions[nPlus1];

		Vector3 delta = ( pointAtNPlus1 - pointAtN ) *
									( 1.0 / (double)( numberToInsert + 1 ) );
		Point3 location = pointAtN + delta;

		position++;

		for (int i = 0; i < numberToInsert; i++)
		{
			poly->addVertexAt( position + i, location,
									 selectStatus == SELECTSTATUS_SELECTED );
								 
			if ( selectStatus == SELECTSTATUS_ONLYSELECTED )
			{
				verticesAdded.push_back( position + i );
			}
		
			location += delta;
		}
	}
	else if ( curveType == CURVETYPE_CATMULROM )
	{
		int nMinus1, nPlus1, nPlus2;

		if ( poly->isClosed() )
		{
			nMinus1 = n - 1;
			if ( nMinus1 < 0 )
			{
				nMinus1 += vertexPositions.size();
			}
			nPlus1 = ( n + 1 ) % vertexPositions.size();
			nPlus2 = ( n + 2 ) % vertexPositions.size();
		}
		else
		{
			nMinus1 = MAX( 0, n - 1 );
			nPlus1 = MIN( (int)vertexPositions.size() - 1, n + 1 );
			nPlus2 = MIN( (int)vertexPositions.size() - 1, n + 2 );
		}


		Vector3 a = vertexPositions[nMinus1].toVector3();
		Vector3 b = vertexPositions[n].toVector3();
		Vector3 c = vertexPositions[nPlus1].toVector3();
		Vector3 d = vertexPositions[nPlus2].toVector3();

		Vector3 basisGeoRow1 = ( -a + b*3.0 - c*3.0 + d ) * 0.5;
		Vector3 basisGeoRow2 = ( a*2.0 - b*5.0 + c*4.0 - d ) * 0.5;
		Vector3 basisGeoRow3 = ( -a + c ) * 0.5;
		Vector3 basisGeoRow4 = b;

		double delta = 1.0 / (double)( numberToInsert + 1 );
		double t = delta;

		position++;

		for (int i = 0; i < numberToInsert; i++)
		{
			double t2 = t * t;
			double t3 = t2 * t;

			Point3 location = basisGeoRow1 * t3 +
									basisGeoRow2 * t2 +
									basisGeoRow3 * t +
									basisGeoRow4;
			
			poly->addVertexAt( position + i, location,
									 selectStatus == SELECTSTATUS_SELECTED );
								 
			if ( selectStatus == SELECTSTATUS_ONLYSELECTED )
			{
				verticesAdded.push_back( position + i );
			}
		
			t += delta;
		}
	}
}







//**************************************************


ToolPolySubdivide::ToolPolySubdivide(CadView *v)
			: ProcessTool( v, "Subdivide", "---Subdivide Polyline---", 1 )
{
	selectStatusLabel = gtk_label_new("Select:");
	gtk_widget_ref(selectStatusLabel);
	gtk_widget_show(selectStatusLabel);

	
	GSList *group = (GSList*)NULL;
	selectedRadio = gtk_radio_button_new_with_label(group, "Selected");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(selectedRadio) );
	unselectedRadio = gtk_radio_button_new_with_label(group, "Unselected");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(unselectedRadio) );
	onlySelectedRadio = gtk_radio_button_new_with_label(group, "Only selected");

	gtk_widget_ref(selectedRadio);
	gtk_widget_ref(unselectedRadio);
	gtk_widget_ref(onlySelectedRadio);
	
	gtk_widget_show(selectedRadio);
	gtk_widget_show(unselectedRadio);
	gtk_widget_show(onlySelectedRadio);


	gtk_signal_connect( GTK_OBJECT(selectedRadio), "clicked",
							  GTK_SIGNAL_FUNC(selectedClicked), this );
	gtk_signal_connect( GTK_OBJECT(unselectedRadio), "clicked",
							  GTK_SIGNAL_FUNC(unselectedClicked), this );
	gtk_signal_connect( GTK_OBJECT(onlySelectedRadio), "clicked",
							  GTK_SIGNAL_FUNC(onlySelectedClicked), this );
	




	curveTypeLabel = gtk_label_new("Curve Type:");
	gtk_widget_ref( curveTypeLabel );
	gtk_widget_show( curveTypeLabel );


	group = (GSList*)NULL;
	linearRadio = gtk_radio_button_new_with_label( group, "Linear" );
	group = gtk_radio_button_group( GTK_RADIO_BUTTON( linearRadio ) );
	catmulromRadio = gtk_radio_button_new_with_label( group, "Catmul-Rom" );

	gtk_widget_ref( linearRadio );
	gtk_widget_ref( catmulromRadio );
	
	gtk_widget_show( linearRadio );
	gtk_widget_show( catmulromRadio );


	gtk_signal_connect( GTK_OBJECT(linearRadio), "clicked",
							  GTK_SIGNAL_FUNC(linearClicked), this );
	gtk_signal_connect( GTK_OBJECT(catmulromRadio), "clicked",
							  GTK_SIGNAL_FUNC(catmumromClicked), this );




	numLabel = gtk_label_new("Number to insert:");
	gtk_widget_ref(numLabel);
	gtk_widget_show(numLabel);

	numEntry = new NumericEntry( 1.0, 1.0, 1048576.0, 1.0, 32.0, 0.5, 0,
										  GTK_SIGNAL_FUNC(numChanged), this );
	gtk_widget_show( numEntry->getWidget() );



	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);
	
	gtk_signal_connect( GTK_OBJECT(doneButton), "clicked",
							  GTK_SIGNAL_FUNC(doneClicked), this );


	layout.addAsLine(selectStatusLabel);
	layout.add(selectedRadio);
	layout.add(unselectedRadio);
	layout.add(onlySelectedRadio);
	layout.lineSpace(1);
	layout.add(linearRadio);
	layout.add(catmulromRadio);
	layout.lineSpace(1);
	layout.add(numLabel);
	layout.add( numEntry->getWidget() );
	layout.lineSpace(1);	
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolPolySubdivide::~ToolPolySubdivide()
{
	delete numEntry;
	
	gtk_widget_unref(selectStatusLabel);
	gtk_widget_unref(selectedRadio);
	gtk_widget_unref(unselectedRadio);
	gtk_widget_unref(onlySelectedRadio);
	gtk_widget_unref(curveTypeLabel);
	gtk_widget_unref(linearRadio);
	gtk_widget_unref(catmulromRadio);
	gtk_widget_unref(numLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolPolySubdivide::getProcessType() const
{
	return ProcPolySubdivide::getTypeStatic();
}




void ToolPolySubdivide::initialise()
{
	ProcessTool::initialise();

	if ( hasProcess() )
	{	
		ProcPolySubdivide *proc = (ProcPolySubdivide*)journalProcess;

		int num = proc->getNumberOfVerticesToInsert();
		ProcPolySubdivide::SelectStatus selStat = proc->getSelectStatus();
		ProcPolySubdivide::CurveType curveType = proc->getCurveType();


		numEntry->setValue(num);


		ignoreRadioEvents = true;

		switch(selStat)
		{
		case ProcPolySubdivide::SELECTSTATUS_SELECTED:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(selectedRadio), true );
			break;
		case ProcPolySubdivide::SELECTSTATUS_UNSELECTED:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(unselectedRadio), true );
			break;
		case ProcPolySubdivide::SELECTSTATUS_ONLYSELECTED:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(onlySelectedRadio), true );
			break;
		default:
			break;
		}

		switch(curveType)
		{
		case ProcPolySubdivide::CURVETYPE_LINEAR:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(linearRadio), true );
			break;
		case ProcPolySubdivide::CURVETYPE_CATMULROM:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(catmulromRadio), true );
			break;
		default:
			break;
		}

		ignoreRadioEvents = false;
	}
}




void ToolPolySubdivide::selectedClicked(GtkWidget *widget,
													 ToolPolySubdivide *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolySubdivide *proc = (ProcPolySubdivide*)klass->journalProcess;
		
		proc->setSelectStatus( ProcPolySubdivide::SELECTSTATUS_SELECTED );
		klass->modified();
	}
}

void ToolPolySubdivide::unselectedClicked(GtkWidget *widget,
														ToolPolySubdivide *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolySubdivide *proc = (ProcPolySubdivide*)klass->journalProcess;
		
		proc->setSelectStatus( ProcPolySubdivide::SELECTSTATUS_UNSELECTED );
		klass->modified();
	}
}

void ToolPolySubdivide::onlySelectedClicked(GtkWidget *widget,
														  ToolPolySubdivide *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolySubdivide *proc = (ProcPolySubdivide*)klass->journalProcess;
		
		proc->setSelectStatus( ProcPolySubdivide::SELECTSTATUS_ONLYSELECTED );
		klass->modified();
	}
}



void ToolPolySubdivide::linearClicked(GtkWidget *widget,
												  ToolPolySubdivide *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolySubdivide *proc = (ProcPolySubdivide*)klass->journalProcess;
		
		proc->setCurveType( ProcPolySubdivide::CURVETYPE_LINEAR );
		klass->modified();
	}
}

void ToolPolySubdivide::catmumromClicked(GtkWidget *widget,
													  ToolPolySubdivide *klass)
{
	if ( klass->hasProcess() && !klass->ignoreRadioEvents )
	{
		ProcPolySubdivide *proc = (ProcPolySubdivide*)klass->journalProcess;
		
		proc->setCurveType( ProcPolySubdivide::CURVETYPE_CATMULROM );
		klass->modified();
	}
}



void ToolPolySubdivide::numChanged(NumericEntry *widget,
											  ToolPolySubdivide *klass)
{
	if ( klass->hasProcess() )
	{
		ProcPolySubdivide *proc = (ProcPolySubdivide*)klass->journalProcess;

		proc->setNumberOfVerticesToInsert( klass->numEntry->getValueAsInt() );
		klass->modified();
	}
}


void ToolPolySubdivide::doneClicked(GtkWidget *widget, ToolPolySubdivide *klass)
{
	klass->done();
}



#endif

