/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLSELECTRECTANGULAR_CPP
#define __TOOLSELECTRECTANGULAR_CPP

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkradiobutton.h>

#include <strstream>

#include "ToolSelectRectangular.h"

#include "JournalModel.h"
#include "Matrix4.h"
#include "ModelDraw.h"





RectangularSelectRegion::RectangularSelectRegion() : SelectRegion()
{
	//make sure the rectangle contains NOTHING
	lower = Point3(1.0, 1.0, 1.0);
	upper = Point3(-1.0, -1.0, -1.0);
}

RectangularSelectRegion::RectangularSelectRegion(
										const Matrix4& worldToScreen,
										const Point3& lowerCorner,
										const Point3& upperCorner )
										: SelectRegion(worldToScreen)
{
	lower = lowerCorner;
	upper = upperCorner;
}


bool RectangularSelectRegion::isSelected2D(const Point3& point) const
{
	return point.x >= lower.x && point.x <= upper.x &&
			 point.y >= lower.y && point.y <= upper.y;
}






void* __create_ProcSelectRectangular()
{
	return (void*)( new ProcSelectRectangular() );
}

ObjectType type_ProcSelectRectangular( "ProcSelectRectangular",
													"ProcSelectBase",
													__create_ProcSelectRectangular );








ProcSelectRectangular::ProcSelectRectangular() : ProcSelectBase()
{
}

ProcSelectRectangular::ProcSelectRectangular(SelectPredicate pred,
							  const RectangularSelectRegion& region)
							  : ProcSelectBase(pred)
{
	selectRegion = region;
}
	

ProcSelectRectangular::~ProcSelectRectangular()
{
}



void ProcSelectRectangular::read(FILE *f)
{
	ProcSelectBase::read(f);

	fread(&selectRegion, sizeof(RectangularSelectRegion), 1, f);
}

void ProcSelectRectangular::write(FILE *f)
{
	ProcSelectBase::write(f);

	fwrite(&selectRegion, sizeof(RectangularSelectRegion), 1, f);
}





string ProcSelectRectangular::getDescription() const
{
	return "Rectangular Select";
}

const ObjectType* ProcSelectRectangular::getType() const
{
	return &type_ProcSelectRectangular;
}

const ObjectType* ProcSelectRectangular::getTypeStatic()
{
	return &type_ProcSelectRectangular;
}



JournalProcess* ProcSelectRectangular::clone() const
{
	return new ProcSelectRectangular(selectPredicate, selectRegion);
}



void ProcSelectRectangular::setSelectRegion(const RectangularSelectRegion& region)
{
	selectRegion = region;
	modified();
}

const RectangularSelectRegion& ProcSelectRectangular::getSelectRegion() const
{
	return selectRegion;
}



void ProcSelectRectangular::selectInModel(JournalModel *model) const
{
	model->getModel()->select( selectPredicate, &selectRegion );
}







ToolSelectRectangular::ToolSelectRectangular( CadView *v,
															 ObjPainterManager *pman )
									: ToolSelectBase( v, pman, "Rect",
															"Rectangular Select" )
{
	drawRegionFlag = false;
	
	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);

	gtk_widget_show(doneButton);


	gtk_signal_connect( GTK_OBJECT(doneButton), "clicked",
							  GTK_SIGNAL_FUNC(doneClicked), this );
	

	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolSelectRectangular::~ToolSelectRectangular()
{
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolSelectRectangular::getProcessType() const
{
	return ProcSelectRectangular::getTypeStatic();
}



void ToolSelectRectangular::start()
{
	ToolSelectBase::start();

	if ( hasProcess() )
	{
		ProcSelectRectangular *proc = (ProcSelectRectangular*)journalProcess;

		initialSelectRegion = proc->getSelectRegion();
	}
	p1 = startCursorPosition.getScreenInt();
	drawRegionFlag = true;
}

void ToolSelectRectangular::finish(bool terminated)
{
	ToolSelectBase::finish(terminated);

	if ( hasProcess() )
	{
		ProcSelectRectangular *proc = (ProcSelectRectangular*)journalProcess;

		drawRegionFlag = false;
		if (terminated)
		{
			proc->setSelectRegion( initialSelectRegion );
			modified();
		}
		else
		{
			updateView();
		}
	}
}

void ToolSelectRectangular::function()
{
	ToolSelectBase::function();

	if ( hasProcess() )
	{
		ProcSelectRectangular *proc = (ProcSelectRectangular*)journalProcess;

		p2 = currentCursorPosition.getScreenInt();

		Point3 startPoint = startCursorPosition.getScreenFloat();
		Point3 currentPoint = currentCursorPosition.getScreenFloat();

		BBox boundingBox = BBox(startPoint);
		boundingBox.addPoint(currentPoint);

		RectangularSelectRegion region =
			RectangularSelectRegion( view->getCadArea()->getWorldScreenMatrix(),
							 boundingBox.getLower(), boundingBox.getUpper() );

		proc->setSelectRegion( region );

		modified();
	}
}


void ToolSelectRectangular::paint()
{
	if (drawRegionFlag)
	{
		int lowerX = p1.x < p2.x ? p1.x : p2.x;
		int lowerY = p1.y < p2.y ? p1.y : p2.y;
		int upperX = p1.x > p2.x ? p1.x : p2.x;
		int upperY = p1.y > p2.y ? p1.y : p2.y;

		IRect2 rectangle = IRect2( IPoint2( lowerX, lowerY ),
											IPoint2( upperX, upperY ) );	

		modelDrawSelectBox( view->getCadArea(), rectangle );
	}
}






void ToolSelectRectangular::doneClicked( GtkWidget *widget,
													  ToolSelectRectangular *klass )
{
	klass->done();
}


#endif

