/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
/*	by :					Geoffrey French
	application :		gSculpt
	last revision :	16/Apr/2000
	title :				Array template class
	purpose :			Provides a class for the puspose of manipulating C arrays
							which does most of the paperwork, and provides functions
							for the manipulation of the contents

  This array class has the ability to resize the array as required; three
  pieces of information are stored;:
  sz		the size or number of elements,
  cap		the number of elements the array can hold until a resize is required,
  dlt		the delta, when the array's capacity must be increased, this is the
  			amount the capacity is increased by.
*/


#ifndef ARRAYH
#define ARRAYH

#include <stdio.h>


#include "dualiterator.h"



#ifndef __array_defaultDelta			//this is the default value for the delta
#define __array_defaultDelta 8
#endif

/*
*******************************************************************************
											array
*******************************************************************************
*/

template <class T>
	class array
{
private:
	//the data, size, capacity, delta, and end pointer
	T *data;
	int sz, cap, dlt;
	T *en;

	//recalculate the end pointer
	void calcEnd()
	{
		en = data + sz;
	}
	
	//increase the capacity by dlt
	void incrementCapacity()
	{
		setCapacity(cap + dlt);
	}

	//typedef the self type for convenience
	typedef array<T> self;

public:
	//typedef iterator types
	typedef T* iterator;
	typedef const T* const_iterator;
	typedef dualiterator<iterator, iterator, const_iterator> dual_iterator;
	typedef dualiterator<const_iterator, iterator, const_iterator> const_dual_iterator;

	//default constructor: 0 elements, initial capacity and delta of default value
	array()
	{
		sz = 0;
		cap = dlt = __array_defaultDelta;
		data = new T[cap];
		calcEnd();
	}
	
	//constructor (given initial capacity)
	array(int initialCap)
	{
		sz = 0;
		cap = initialCap;
		dlt = __array_defaultDelta;
		data = new T[cap];
		calcEnd();
	}

	//constructor (given initial capacity and delta)
	array(int initialCap, int delta)
	{
		sz = 0;
		cap = initialCap;
		dlt = delta;
		data = new T[cap];
		calcEnd();
	}

	//construct from C array
	array(T *inData, int aSize, int initialCap = aSize)
	{
		sz = aSize;
		cap = initialCap;
		dlt = __array_defaultDelta;
		data = new T[cap];
		for (int i = 0; i < sz; i++)
		{
			data[i] = inData[i];
		}
		calcEnd();
	}
	
	//construct from C array (2)
	array(T *inData, int aSize, int initialCap, int delta)
	{
		sz = aSize;
		cap = initialCap;
		dlt = delta;
		data = new T[cap];
		for (int i = 0; i < sz; i++)
		{
			data[i] = inData[i];
		}
		calcEnd();
	}

	//copy constructor
	array(const self& a)
	{
		sz = a.sz;
		cap = a.cap;
		dlt = a.dlt;
		data = new T[cap];
		for (int i = 0; i < sz; i++)
		{
			data[i] = a.data[i];
		}
		//completely different array; end ptr must be recalculated, not copied
		calcEnd();
	}

	//destructor
	~array()
	{
		if (data != NULL)
		{
			delete[] data;
		}
		en = NULL;
		sz = cap = dlt = 0;
	}
	

	//assignment operator: copy
	self& operator=(const self& a)
	{
		//delete old contents
		if (data != NULL)
		{
			delete[] data;
		}

		//get new contents
		sz = a.sz;
		cap = a.cap;
		dlt = a.dlt;
		data = new T[cap];
		for (int i = 0; i < sz; i++)
		{
			data[i] = a.data[i];
		}
		calcEnd();

		//return a reference to this
		return *this;
	}


	//== comparison operator
	bool operator==(const self& a) const
	{
		//compare sizes followed by each element
		if (sz != a.sz)
		{
			return false;
		}

		for (int i = 0; i < sz; i++)
		{
			if (data[i] != a.data[i])
			{
				return false;
			}
		}

		return true;
	}


	bool operator!=(const self& a) const
	{
		return !(*this == a);
	}


	//set the capacity
	void setCapacity(int c)
	{
		//assign the new capacity
		cap = c;
		//shrink the size if necessary
		if (cap < sz)
		{
			sz = cap;
		}
		//allocate new array, and copy data
		T *tmp = new T[cap];
		for (int i = 0; i < sz; i++)
		{
			tmp[i] = data[i];
		}
		//remove the old array
		delete[] data;
		//make the new array take it's place
		data = tmp;
		//new array; must recalc end ptr
		calcEnd();
	}

	//reserve r elements; make sure capacity is at least r
	void reserve(int r)
	{
		if (r > cap)
		{
			setCapacity(r);
		}
	}

	//resize the array
	void resize(int s)
	{
		//increase the capacity if necessary
		if (s > cap)
		{
			setCapacity(s);
		}
		//need to set sz AFTER increasing capacity, or setCapacity() will segfault
		sz = s;
		//size has changed; end will be at a different location now
		calcEnd();
	}

	//increase size by n elements
	void increaseSize(int n = 1)
	{
		//calculate new size
		int newSize = sz + n;

		//see if the current capacity is sufficient
		if (newSize > cap)
		{
			//its insufficient
			//if increasing the capacity by delta is sufficient, do it, else
			//let resize() increase the capacity to the required amount
			if (newSize <= (cap + dlt))
			{
				incrementCapacity();
			}
			resize(sz + n);
		}
		else
		{
			//current capacity is sufficient
			//increase size and recalc end ptr
			sz += n;
			calcEnd();
		}
	}

	//clear the array
	void clear()
	{
		sz = 0;
		calcEnd();
	}
	
	//insert n spaces at position p
	void insert(int p, int n = 1)
	{
		//if the starting position > size
		if (p >= sz)
		{
			//elements dont need to be moved forward; juts resize
			resize(p + n);
		}
		else
		{
			//make copy of size
			int oldSz = sz;
			//increase size to accomodate extra spaces
			increaseSize(n);
			//move elements forward
			for (int i = oldSz - 1; i >= p; i--)
			{
				data[i + n] = data[i];
			}
		}
	}

	//remove n spaces at p
	void erase(int p, int n = 1)
	{
		//if starting position > sz, don't bother; nothing to do
		if (p < sz)
		{
			if ((p + n) >= sz)
			{
				//if (p + n) >= sz, then only elements at the end are removed, so
				//elements don't need to be moved back, just remove those affected
				resize(p);
			}
			else
			{
				//move subsequent elements back
				for (int i = p; i <= (sz - n); i++)
				{
					data[i] = data[i + n];
				}

				//reduce size
				resize(sz - n);
			}
		}
	}

	//erase an element pointed to by an iterator
	void erase(iterator pos)
	{
		//search and delete
		for (int i = 0; i < sz; i++)
		{
			if ( (data + i) == pos )
			{
				erase(i);
				return;
			}
		}
	}

	//add an element
	int push_back(const T& element)
	{
		//make the necessary space
		if (sz >= cap)
		{
			incrementCapacity();
		}
		//add it
		data[sz++] = element;
		calcEnd();
		return sz - 1;
	}


	//first element
	T& front()
	{
		return data[0];
	}

	const T& front() const
	{
		return data[0];
	}


	//last element
	T& back()
	{
		return data[sz - 1];
	}

	const T& back() const
	{
		return data[sz - 1];
	}


	//get size
	int size() const
	{
		return sz;
	}

	//get capacity
	int capacity() const
	{
		return cap;
	}


	//get element at specific index
	T& operator[](int i)
	{
		return data[i];
	}

	const T& operator[](int i) const
	{
		return data[i];
	}


	//iterators pointing to beginning, last element, and end
	iterator begin()
	{
		return data;
	}

	iterator last()
	{
		return en - 1;
	}

	iterator end()
	{
		return en;
	}

	const_iterator begin() const
	{
		return data;
	}

	const_iterator last() const
	{
		return en - 1;
	}

	const_iterator end() const
	{
		return en;
	}

	dual_iterator beginDual()
	{
		return dual_iterator(data);
	}

	dual_iterator lastDual()
	{
		return dual_iterator(en - 1);
	}

	dual_iterator endDual()
	{
		return dual_iterator(en);
	}

	dual_iterator circularDual()
	{
		return dual_iterator(en - 1, data);
	}

	const_dual_iterator beginDual() const
	{
		return const_dual_iterator(data);
	}

	const_dual_iterator lastDual() const
	{
		return const_dual_iterator(en - 1);
	}

	const_dual_iterator endDual() const
	{
		return const_dual_iterator(en);
	}

	const_dual_iterator circularDual() const
	{
		return const_dual_iterator(en - 1, data);
	}
};


#endif
