/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __CVERTEXCONTAINER_CPP
#define __CVERTEXCONTAINER_CPP

#include "CVertexContainer.h"


/*
*******************************************************************************
										CadVertexContainer
*******************************************************************************
*/

CVertexContainer::CVertexContainer() : array<CVertex>()
{
}


void CVertexContainer::read(FILE *f)
{
	clear();

	int numVertices;
	fread( &numVertices, sizeof(int), 1, f );

	for (int i = 0; i < numVertices; i++)
	{
		CVertex vertex;

		vertex.read(f);
		push_back(vertex);
	}
}

void CVertexContainer::write(FILE *f) const
{
	int numVertices = size();
	fwrite( &numVertices, sizeof(int), 1, f );

	for (int i = 0; i < (int)size(); i++)
	{
		const CVertex &vertex = get(i);

		vertex.write(f);
	}
}




CVertex& CVertexContainer::get(int i)
{
	return (*this)[i];
}

const CVertex& CVertexContainer::get(int i) const
{
	return (*this)[i];
}



void CVertexContainer::copyFrom(const CVertexContainer *c)
{
	resize( c->size() );
	
	for (int i = 0; i < (int)c->size(); i++)
	{
		get(i).copyFrom( &( c->get(i) ) );
	}
}



void CVertexContainer::transform(const Matrix4& m, bool selectedOnly)
{
	for (int i = 0; i < (int)size(); i++)
	{
		CVertex &vertex = get(i);

		if ( vertex.isMarked() || !selectedOnly )
		{
			vertex.transform(m);
		}
	}
}

void CVertexContainer::transform(const Transformation *x, bool selectedOnly)
{
	for (int i = 0; i < (int)size(); i++)
	{
		CVertex &vertex = get(i);

		if ( vertex.isMarked() || !selectedOnly )
		{
			vertex.transform(x);
		}
	}
}


void CVertexContainer::select(SelectPredicate predicate,
										const SelectRegion* region)
{
	for (int i = 0; i < (int)size(); i++)
	{
		CVertex &vertex = get(i);

		vertex.select(predicate, region);
	}
}



CVertexContainer::VertexBackup CVertexContainer::backup(bool selectedOnly) const
{
	VertexBackup b;

	
	for (int i = 0; i < (int)size(); i++)
	{
		const CVertex &v = get(i);

		if ( v.isMarked() || !selectedOnly )
		{
			VertexBackup::Entry e;

			e.index = i;
			e.data = v.backup();

			b.bakData.push_back(e);
		}
	}

	return b;
}

void CVertexContainer::restore(const VertexBackup& b)
{
	for (int i = 0; i < (int)b.bakData.size(); i++)
	{
		const VertexBackup::Entry &e = b.bakData[i];

		int vtxIndex = e.index;
		CVertex &vtx = get(vtxIndex);

		vtx.restore( e.data );
	}
}



BBox CVertexContainer::getBBox(bool selectedOnly) const
{
	bool found = false;
	BBox r;
	
	for (int i = 0; i < (int)size(); i++)
	{
		const CVertex &v = get(i);

		if ( v.isMarked() || !selectedOnly )
		{
			Point3 p = v.getWorld();
			
			if (!found)
			{
				r = BBox(p);
				found = true;
			}
			else
			{
				r.addPoint(p);
			}
		}
	}

	return r;
}

BBox CVertexContainer::getBBox(const Matrix4& m, bool selectedOnly) const
{
	bool found = false;
	BBox r;
	
	for (int i = 0; i < (int)size(); i++)
	{
		const CVertex &v = get(i);

		if ( v.isMarked() || !selectedOnly )
		{
			Point3 p = v.getWorld() * m;
			
			if (!found)
			{
				r = BBox(p);
				found = true;
			}
			else
			{
				r.addPoint(p);
			}
		}
	}

	return r;
}

BBox CVertexContainer::getBBox(const Transformation *x, bool selectedOnly) const
{
	bool found = false;
	BBox r;
	
	for (int i = 0; i < (int)size(); i++)
	{
		const CVertex &v = get(i);

		if ( v.isMarked() || !selectedOnly )
		{
			Point3 p = x->transformPoint( v.getWorld() );
			
			if (!found)
			{
				r = BBox(p);
				found = true;
			}
			else
			{
				r.addPoint(p);
			}
		}
	}

	return r;
}




#endif
