/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __CADAREA_CPP
#define __CADAREA_CPP


#include "CadArea.h"
#include <iostream>



/*
*******************************************************************************
											CAEventHandler
*******************************************************************************
*/

CAEventHandler::CAEventHandler(CadArea *c)
										 : DAEventHandler( (DrawingArea*)c )
{
	cadArea = c;
}

void CAEventHandler::evXformChange()
{
}

void CAEventHandler::redirectEvent(Event *ev)
{
	DAEventHandler::redirectEvent(ev);

	if (ev->getName() == "xform_change")
	{
		evXformChange();
	}
}



/*
*******************************************************************************
												CadArea
*******************************************************************************
*/


CadArea::CadArea() : DrawingArea()
{
	ws_xlat = ws_direction = ws_zoom = Matrix4(Vector3(1.0, 0.0, 0.0),
															 Vector3(0.0, 1.0, 0.0),
															 Vector3(0.0, 0.0, 1.0));
	sw_zoom = sw_direction = sw_xlat = Matrix4(Vector3(1.0, 0.0, 0.0),
															 Vector3(0.0, 1.0, 0.0),
															 Vector3(0.0, 0.0, 1.0));
	ws_projection = Matrix4(Vector3(1.0, 0.0, 0.0),
									Vector3(0.0, 1.0, 0.0),
									Vector3(0.0, 0.0, -1.0));
	sw_projection = ~ws_projection;
	concatenate();

	frame = 1;
}


void CadArea::nextFrame()
{
	frame++;
	if (frame == 0)
	{
		frame = 1;
	}
}

//respond to configure events
void CadArea::ev_config(const ISize2 &s)
{
	//centre the screen
	Vector3 translation = Vector3( (double)(s.w >> 1),
											 (double)(s.h >> 1),
											 0.0 );
	setProjectionMatrix( Matrix4(Vector3(1.0, 0.0, 0.0),
										  Vector3(0.0, 1.0, 0.0),
										  Vector3(0.0, 0.0, 1.0)) *
								Matrix4_translate(translation) );

	//matrices have been changed; they need concatenating
	concatenate();
}

void CadArea::ev_xformChange()
{
}




Frame CadArea::getFrame() const
{
	return frame;
}


//multiply the four component matrices to get the final transformations
void CadArea::concatenate()
{
	array<Point3> corners;

	world_screen = ws_xlat * ws_direction * ws_zoom * ws_projection;
	screen_world = sw_projection * sw_zoom * sw_direction * sw_xlat;
	world_screen_transformation = MatrixTransformation(world_screen);
	screen_world_transformation = MatrixTransformation(screen_world);
	
	corners.push_back(
		Point3(0.0, 0.0, 0.0) * screen_world );
	corners.push_back(
		Point3(0.0, (double)getSize().h, 0.0) * screen_world );
	corners.push_back(
		Point3((double)getSize().w, (double)getSize().h, 0.0) * screen_world );
	corners.push_back(
		Point3((double)getSize().w, 0.0, 0.0) * screen_world );

	frustrum = ViewFrustrum( Vector3(0.0, 0.0, -1.0) * screen_world,
									 corners );

	nextFrame();

	ev_xformChange();
	Event ev("xform_change");
	event(&ev);	
}

//set and get the various matrices (world to screen only)
void CadArea::setXlatMatrix(const Matrix4& wldscr)
{
	ws_xlat = wldscr;
	sw_xlat = ~wldscr;
}

void CadArea::setDirectionMatrix(const Matrix4& wldscr)
{
	ws_direction = wldscr;
	sw_direction = ~wldscr;
}

void CadArea::setZoomMatrix(const Matrix4& wldscr)
{
	ws_zoom = wldscr;
	sw_zoom = ~wldscr;
}

void CadArea::setProjectionMatrix(const Matrix4& wldscr)
{
	ws_projection = wldscr;
	sw_projection = ~wldscr;
}

Matrix4 CadArea::getXlatMatrix() const
{
	return ws_xlat;
}

Matrix4 CadArea::getDirectionMatrix() const
{
	return ws_direction;
}

Matrix4 CadArea::getZoomMatrix() const
{
	return ws_zoom;
}

Matrix4 CadArea::getProjectionMatrix() const
{
	return ws_projection;
}

Matrix4 CadArea::getWorldScreenMatrix() const
{
	return world_screen;
}

Matrix4 CadArea::getScreenWorldMatrix() const
{
	return screen_world;
}

const ViewFrustrum& CadArea::getViewFrustrum() const
{
	return frustrum;
}


void CadArea::setDirectionMatrix_front()
{
	setDirectionMatrix( Matrix4(
		Vector3(1.0, 0.0, 0.0),
		Vector3(0.0, 1.0, 0.0),
		Vector3(0.0, 0.0, 1.0) ) );
}

void CadArea::setDirectionMatrix_back()
{
	setDirectionMatrix( Matrix4(
		Vector3(-1.0, 0.0, 0.0),
		Vector3(0.0, 1.0, 0.0),
		Vector3(0.0, 0.0, -1.0) ) );
}

void CadArea::setDirectionMatrix_left()
{
	setDirectionMatrix( Matrix4(
		Vector3(0.0, 0.0, -1.0),
		Vector3(0.0, 1.0, 0.0),
		Vector3(1.0, 0.0, 0.0) ) );
}

void CadArea::setDirectionMatrix_right()
{
	setDirectionMatrix( Matrix4(
		Vector3(0.0, 0.0, 1.0),
		Vector3(0.0, 1.0, 0.0),
		Vector3(-1.0, 0.0, 0.0) ) );
}

void CadArea::setDirectionMatrix_top()
{
	setDirectionMatrix( Matrix4(
		Vector3(1.0, 0.0, 0.0),
		Vector3(0.0, 0.0, 1.0),
		Vector3(0.0, -1.0, 0.0) ) );
}

void CadArea::setDirectionMatrix_bottom()
{
	setDirectionMatrix( Matrix4(
		Vector3(1.0, 0.0, 0.0),
		Vector3(0.0, 0.0, -1.0),
		Vector3(0.0, 1.0, 0.0) ) );
}


//convert world coordinates to screen coordinates
Vector3 CadArea::transformWorldScreen(const Vector3& v) const
{
	return v * world_screen;
}

//convert world coordinated to screen coordinates (into integers)
IVector2 CadArea::transformWorldScreenInt2(const Vector3& v) const
{
	return IVector2( transformWorldScreen(v) );
}

//convert screen coordinates to world coordinates
Vector3 CadArea::transformScreenWorld(const Vector3& v) const
{
	return v * screen_world;
}

//convert world coordinates to screen coordinates, no direction
Vector3 CadArea::transformWorldScreen_nodirection(const Vector3& v) const
{
	//return p * (ws_xlat * ws_zoom * ws_projection);
	return v * (sw_direction * world_screen);
}

//convert world coordinates to screen coordinates, no direction
IVector2 CadArea::transformWorldScreen_nodirectionInt2(const Vector3& v) const
{
	//return p * (ws_xlat * ws_zoom * ws_projection);
	return IVector2( v * (sw_direction * world_screen) );
}

//convert screen coordinates to world coordinates, no direction
Vector3 CadArea::transformScreenWorld_nodirection(const Vector3& v) const
{
	//return p * (sw_projection * sw_zoom * sw_xlat);
	return v * (screen_world * ws_direction);
}

//convert world coordinates to screen coordinates
Point3 CadArea::transformWorldScreen(const Point3& p) const
{
	return p * world_screen;
}

//convert world coordinated to screen coordinates (into integers)
IPoint2 CadArea::transformWorldScreenInt2(const Point3& p) const
{
	return IPoint2( transformWorldScreen(p) );
}



//convert screen coordinates to world coordinates
Point3 CadArea::transformScreenWorld(const Point3& p) const
{
	return p * screen_world;
}

//convert world coordinates to screen coordinates, no direction
Point3 CadArea::transformWorldScreen_nodirection(const Point3& p) const
{
	return p * (sw_direction * world_screen);
}

//convert world coordinates to screen coordinates, no direction
IPoint2 CadArea::transformWorldScreen_nodirectionInt2(const Point3& p) const
{
	return IPoint2( p * (sw_direction * world_screen) );
}

//convert screen coordinates to world coordinates, no direction
Point3 CadArea::transformScreenWorld_nodirection(const Point3& p) const
{
	return p * (screen_world * ws_direction);
}



//get world axes in screen space
Axes3 CadArea::getWorldAxesInScreenSpace() const
{
	return Axes3() * world_screen;
}

//get world axes in screen space
Axes3 CadArea::getScreenAxesInWorldSpace() const
{
	return Axes3() * screen_world;
}



//get world to screen transformation
const Transformation* CadArea::getWorldScreenTransformation() const
{
	return (Transformation*)&world_screen_transformation;
}

//get screen to world transformation
const Transformation* CadArea::getScreenWorldTransformation() const
{
	return (Transformation*)&screen_world_transformation;
}


const ViewFrustrum& CadArea::getFrustrum() const
{
	return frustrum;
}


//screen axis inversion
bool CadArea::isScreenXInverted() const
{
	return scrXInv;
}

bool CadArea::isScreenYInverted() const
{
	return scrYInv;
}


//zoom function
void CadArea::zoom(double z)
{
	setZoomMatrix( getZoomMatrix() * Matrix4_scale( Vector3(z, z, z) ) );
	concatenate();
}

//set zoom level
void CadArea::setZoom(double z)
{
	setZoomMatrix( Matrix4_scale( Vector3(z, z, z) ) );
	concatenate();
}

//set translation
void CadArea::setXlat(const Vector3& xl)
{
	setXlatMatrix( Matrix4_translate(xl) );
	concatenate();
}

//drag function
void CadArea::drag(const Vector3& d)
{
	setXlatMatrix( getXlatMatrix() * Matrix4_translate( d ) );
	concatenate();
}

//front
void CadArea::front()
{
	setDirectionMatrix_front();
	scrXInv = false;
	scrYInv = false;
	concatenate();
}

//back
void CadArea::back()
{
	setDirectionMatrix_back();
	scrXInv = true;
	scrYInv = false;
	concatenate();
}

//left
void CadArea::left()
{
	setDirectionMatrix_left();
	scrXInv = false;
	scrYInv = false;
	concatenate();
}

//right
void CadArea::right()
{
	setDirectionMatrix_right();
	scrXInv = true;
	scrYInv = false;
	concatenate();
}

//top
void CadArea::top()
{
	setDirectionMatrix_top();
	scrXInv = false;
	scrYInv = true;
	concatenate();
}

//bottom
void CadArea::bottom()
{
	setDirectionMatrix_bottom();
	scrXInv = false;
	scrYInv = false;
	concatenate();
}



#endif
