/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
/*
	by :					Geoffrey French
	application :		gSculpt
	last revision :	17/Sep/2000
	title :				Cad area classes
	purpose :			Derive subclasses from DAEventHandler, and DrawingArea
							to provide extra functionality for the purpose of
							writing CAD software
*/
#ifndef __CADAREA_H
#define __CADAREA_H



#include "Geometry.h"
#include "DrawingArea.h"
#include "Vector3.h"
#include "Point3.h"
#include "Matrix4.h"
#include "Transformation.h"
#include "MatrixTransformation.h"
#include "ViewFrustrum.h"
#include "Frame.h"
#include "Axes3.h"




/*
*******************************************************************************
										CadArea (prototype)

	Declared here because a pointer to a CadArea class is required in the
CAEventHandler class, which must be declared before the CadArea class
*******************************************************************************
*/

class CadArea;




/*
*******************************************************************************
											CAEventHandler

	Purely for convenience; subclasses of CAEventHandler will require functions
present in CadArea, but not in DrawingArea. CAEventHandler contains a pointer
to a CadArea, so that the DrawingArea pointer in DAEventHandler does not
need to be cast to CadArea* each time it is used.
*******************************************************************************
*/

class CAEventHandler : public DAEventHandler
{
protected:
	CadArea *cadArea;

public:
	CAEventHandler(CadArea *c);


protected:
	virtual void evXformChange();

	virtual void redirectEvent(Event *ev);	
};



/*
*******************************************************************************
												CadArea

	Combines the features of DrawingArea, and CadAreaXforms; has the ability
to transform between world space and screen space, and draw.
	Has the added feature that each time the widget is resized, the contents
centre properly
*******************************************************************************
*/

class CadArea : public DrawingArea
{
private:
	//transformations
	//from world space to screen space, in this order
	Matrix4 ws_xlat, ws_direction, ws_zoom, ws_projection, world_screen;
	//form screen space to world space, in this order
	Matrix4 sw_projection, sw_zoom, sw_direction, sw_xlat, screen_world;

	MatrixTransformation world_screen_transformation;
	MatrixTransformation screen_world_transformation;

	ViewFrustrum frustrum;

	Frame frame;

	//axis inversion flags
	bool scrXInv, scrYInv;


private:
	void nextFrame();

protected:
	//respond to configure events by overloading the ev_config function
	//defined in DrawingArea
	virtual void ev_config(const ISize2& s);
	virtual void ev_xformChange();

public:
	CadArea();

	Frame getFrame() const;

	//concatenate matrices
	void concatenate();

	//set and get the various matrices (world to screen only)
	void setXlatMatrix(const Matrix4& wldscr);
	void setDirectionMatrix(const Matrix4& wldscr);
	void setZoomMatrix(const Matrix4& wldscr);
	void setProjectionMatrix(const Matrix4& wldscr);
	Matrix4 getXlatMatrix() const;
	Matrix4 getDirectionMatrix() const;
	Matrix4 getZoomMatrix() const;
	Matrix4 getProjectionMatrix() const;
	Matrix4 getWorldScreenMatrix() const;
	Matrix4 getScreenWorldMatrix() const;
	const ViewFrustrum& getViewFrustrum() const;

	void setDirectionMatrix_front();
	void setDirectionMatrix_back();
	void setDirectionMatrix_left();
	void setDirectionMatrix_right();
	void setDirectionMatrix_top();
	void setDirectionMatrix_bottom();

	//convert coordinates between world and screen
	Vector3 transformWorldScreen(const Vector3& v) const;
	IVector2 transformWorldScreenInt2(const Vector3& v) const;
	Vector3 transformScreenWorld(const Vector3& v) const;
	Vector3 transformWorldScreen_nodirection(const Vector3& v) const;
	IVector2 transformWorldScreen_nodirectionInt2(const Vector3& v) const;
	Vector3 transformScreenWorld_nodirection(const Vector3& v) const;
	Point3 transformWorldScreen(const Point3& p) const;
	IPoint2 transformWorldScreenInt2(const Point3& p) const;
	Point3 transformScreenWorld(const Point3& p) const;
	Point3 transformWorldScreen_nodirection(const Point3& p) const;
	IPoint2 transformWorldScreen_nodirectionInt2(const Point3& p) const;
	Point3 transformScreenWorld_nodirection(const Point3& p) const;

	//axes
	Axes3 getWorldAxesInScreenSpace() const;
	Axes3 getScreenAxesInWorldSpace() const;
	

	//get screen axis status (inverted)
	//true if +ve in screen X translates to -ve in world
	bool isScreenXInverted() const;
	//true if +ve in screen Y translates to -ve in world
	bool isScreenYInverted() const;

	//get matrix transformation objects
	const Transformation* getWorldScreenTransformation() const;
	const Transformation* getScreenWorldTransformation() const;

	//get frustrum
	const ViewFrustrum& getFrustrum() const;

	//viewport manipulation functions
	void zoom(double z);
	void setZoom(double z);
	void drag(const Vector3& d);
	void setXlat(const Vector3& xl);
	void front();
	void back();
	void left();
	void right();
	void top();
	void bottom();
};


#endif
