/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
/*
	by:					Geoffrey French
	application:		gSculpt
	last revision:		09/Jul/2000
	title:				Drawing Area class
	puspose:				Provide a class used for manipulating a gtk drawing area
							widget. Provides a convenient method of drawing to a
							drawing area and setting callbacks for the various
							events. Uses a backing pixmap to do double buffering.
							ALL drawing is done to the backing pixmap, which is then
							drawn onto the widget
*/
#ifndef __DRAWINGAREA_H
#define __DRAWINGAREA_H

#include <list>

#include <gtk/gtkwidget.h>

#include "foreach.h"

#include "Event.h"
#include "Geometry.h"
#include "PointerStatus.h"



/******************************************************************************
 ******************************************************************************
											FIXME
										***********

	Keyboard events are NOT received!!!!!!! Need to grab keyboard using
	gdk_keyboard_grab() and gdk_keyboard_ungrab()
 ******************************************************************************
 ******************************************************************************/




/*
*******************************************************************************
												DAColour

	Provides a class to manipulate colours used with the DrawingArea class
*******************************************************************************
*/

class DAColour
{
private:
	GdkColor gdkColour;									//info stored in a GdkColor object
	bool allocatedFlag;									//allocated yet?

public:
	DAColour();												//default constructor
	DAColour(int iRed, int iGreen, int iBlue);	//construct from rgb value:
	DAColour(const GdkColor& c);				//(private) construct from a GdkColor


	int getRed() const;										//get the r, g, and b components
	int getGreen() const;
	int getBlue() const;

	GdkColor getGdkColour() const;

	bool allocate();										//allocate the colour
	bool isAllocated() const;								//allocated?

//	friend class DrawingArea;
};




/*
*******************************************************************************
*/

//prototype the DrawingArea class
class DrawingArea;



/*
*******************************************************************************
										EventMouseButton
*******************************************************************************
*/

class EventMouseButton : public Event
{
private:
	int button;

public:
	EventMouseButton();
	EventMouseButton(const string& nm, int bt);

	int getButton() const;
};


/*
*******************************************************************************
										EventMouseMotion
*******************************************************************************
*/

class EventMouseMotion : public Event
{
private:
	IPoint2 position;

public:
	EventMouseMotion();
	EventMouseMotion(const string& nm, const IPoint2& pos);

	IPoint2 getPosition() const;
};


/*
*******************************************************************************
											EventKey
*******************************************************************************
*/

class EventKey : public Event
{
private:
	int keyVal;

public:
	EventKey();
	EventKey(const string& nm, int keyValue);

	int getKeyValue() const;
};


/*
*******************************************************************************
											EventConfig
*******************************************************************************
*/

class EventConfig : public Event
{
private:
	ISize2 size;

public:
	EventConfig();
	EventConfig(const string& nm, const ISize2& sz);

	ISize2 getSize() const;
};



/*
*******************************************************************************
											DAEventHandler
*******************************************************************************
*/

class DAEventHandler : public EventHandler
{
protected:
	DrawingArea *drawingArea;		//this is often needed for use by the event handlers
	
public:
	DAEventHandler(DrawingArea *drawArea);

protected:
	//respond to:
	//update requests
	virtual void paint();
	//button events
	virtual void evButtonDown(int button);
	virtual void evButtonUp(int button);
	//pointer movements
	virtual void evMotion(const IPoint2& p);
	virtual void evEnter(const IPoint2& p);
	virtual void evLeave(const IPoint2& p);
	//key events
	virtual void evKeyPress(int keyValue);
	virtual void evKeyRelease(int keyValue);
	//resize events
	virtual void evConfig(const ISize2& sz);

	virtual void redirectEvent(Event *ev);

	friend class DrawingArea;
};



/*
*******************************************************************************
											DrawingArea

	The DrawingArea class itself.

	If subclasses of DrawingArea need to respond to events, overload the ev_*()
functions. use getWidget() to get the GTK widget. Events are generally
processed by the event handler, which would be a subclass of DAEventHandler.
Set it using setEventHandler(). To update the widget contents, call update().
Draw using the draw*() functions.
*******************************************************************************
*/

class DrawingArea
{
private:
	//the gtk widget
	GtkWidget *gtkDrawingAreaWidget;
	//GC for drawing to the backing pixmap
	GdkGC *drawingAreaGC;
	//the backing pixmap
	GdkPixmap *backingPixmap;
	//the snapshot pixmap
	GdkPixmap *snapPixmap;
	//background and foreground colours
	GdkColor background, foreground;
	//size of widget
	ISize2 drawingAreaSize;
	//current pointer status
	PointerStatus pointerStatus;

protected:
	//event handlers
	list<DAEventHandler*> eventHandlers;
	bool eventHandlersActiveFlag;


public:
	//constructor
	DrawingArea();
	virtual ~DrawingArea();


protected:
	//class's private event response functions
	static gint configure_event(GtkWidget *widget, GdkEventConfigure *event, DrawingArea *klass);
	static gint expose_event(GtkWidget *widget, GdkEventExpose *event, DrawingArea *klass);
	static gint button_press_event(GtkWidget *widget, GdkEventButton *event, DrawingArea *klass);
	static gint button_release_event(GtkWidget *widget, GdkEventButton *event, DrawingArea *klass);
	static gint motion_notify_event(GtkWidget *widget, GdkEventMotion *event, DrawingArea *klass);
	static gint enter_notify_event(GtkWidget *widget, GdkEventCrossing *event, DrawingArea *klass);
	static gint leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, DrawingArea *klass);
	static gint key_press_event(GtkWidget *widget, GdkEventKey *event, DrawingArea *klass);
	static gint key_release_event(GtkWidget *widget, GdkEventKey *event, DrawingArea *klass);

	virtual void ev_paint();
	virtual void ev_buttonDown(int button);
	virtual void ev_buttonUp(int button);
	virtual void ev_motion(const IPoint2& p);
	virtual void ev_enter(const IPoint2& p);
	virtual void ev_leave(const IPoint2& p);
	virtual void ev_keyPress(int keyValue);
	virtual void ev_keyRelease(int keyValue);
	virtual void ev_config(const ISize2& s);

	void event(Event *ev);



public:
	//get the GTK widget
	GtkWidget* getWidget();
	//get the size
	ISize2 getSize() const;
	//get the pointer status
	const PointerStatus& getPointerStatus() const;

	//event handlers
	void addEventHandler(DAEventHandler *h);
	void removeEventHandler(DAEventHandler *h);
	bool replaceEventHandler(DAEventHandler *old, DAEventHandler *h);
	void activateEventHandlers();
	void deactivateEventHandlers();
	bool isActiveEventHandlers() const;
	

	//redraw the drawing area
	void redraw();

	//set foreground and background colours
	void setForeground(DAColour& fore);
	void setBackground(DAColour& back);
	//clear the drawing area to the background colour
	void drawClear();
	//draw primitives
	void drawDot(const IPoint2& p);
	void drawLine(const ILine2& l);
	void drawRectangle(const IRect2& r, bool filled = false);
	void drawArc(const IPoint2& centre, const ISize2& s, int a1, int a2,
					 bool filled = false);
	void drawString(const IPoint2& p, const char *str);

	//get foreground and background colours
	DAColour getForeground() const;
	DAColour getBackground() const;

	//take a snapshot of the drawing area
	void takeSnapshot();
	//draw the snapshot previously takes onto the drawing area
	void drawSnap(const IRect2& r);
	void drawSnap();
};



#endif
