/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLXFBEND_CPP
#define __TOOLXFBEND_CPP

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>

#include <strstream>

#include "ToolXfBend.h"

#include "JournalModel.h"
#include "Matrix4.h"
#include "radians.h"
#include "BendTransformation.h"





void* __create_ProcXfBend()
{
	return (void*)( new ProcXfBend() );
}

ObjectType type_ProcXfBend( "ProcXfBend", "ProcXformAxis",
									 __create_ProcXfBend );






ProcXfBend::ProcXfBend() : ProcXformAxis()
{
	direction = 0.0;
	bendAngle = 0.0;
}

ProcXfBend::ProcXfBend(Axis a, double dir, double bend)
							  : ProcXformAxis(a)
{
	direction = dir;
	bendAngle = bend;
}

ProcXfBend::~ProcXfBend()
{
}



void ProcXfBend::read(FILE *f)
{
	ProcXformAxis::read(f);

	fread(&direction, sizeof(double), 1, f);
	fread(&bendAngle, sizeof(double), 1, f);
}

void ProcXfBend::write(FILE *f)
{
	ProcXformAxis::write(f);

	fwrite(&direction, sizeof(double), 1, f);
	fwrite(&bendAngle, sizeof(double), 1, f);
}





string ProcXfBend::getDescription() const
{
	return "Bend";
}

const ObjectType* ProcXfBend::getType() const
{
	return &type_ProcXfBend;
}

const ObjectType* ProcXfBend::getTypeStatic()
{
	return &type_ProcXfBend;
}





JournalProcess* ProcXfBend::clone() const
{
	return new ProcXfBend(axis, direction, bendAngle);
}



void ProcXfBend::setDirection(double dir)
{
	direction = dir;
	modified();
}

void ProcXfBend::setBendAngle(double bend)
{
	bendAngle = bend;
	modified();
}

double ProcXfBend::getDirection() const
{
	return direction;
}

double ProcXfBend::getBendAngle() const
{
	return bendAngle;
}




void ProcXfBend::transformModel(JournalModel *model, const Point3& centre,
										  const Matrix4& axesMatrix,
										  const Matrix4& axesMatrixInverse,
										  const BBox& boundingBoxInTransformationSpace) const
{
	double bendInRadians = DEGTORAD( -bendAngle );
	double dirInRadians = DEGTORAD( direction );
	
	if (bendInRadians != 0.0)
	{
		Matrix4 bendAxesMatrixInverse = Matrix4_rotateZ( -dirInRadians ) *
												  axesMatrix;
		Matrix4 bendAxesMatrix = ~bendAxesMatrixInverse;
		
		BBox box = model->getModel()->getBBox(bendAxesMatrix);
		Point3 bendBoxUpper = box.getUpper();
		Point3 bendBoxCentre = box.getCentre();
		
		//calculate bend radius
		double bendRadius = box.getDepth() / bendInRadians;

		//make sure origin is outside of bounding box
		if ((bendRadius > 0.0) && (bendRadius < box.getWidth() * 0.5))
		{
			bendRadius = box.getWidth() * 0.5;
		}
		if ((bendRadius < 0.0) && (bendRadius > -box.getWidth() * 0.5))
		{
			bendRadius = -box.getWidth() * 0.5;
		}

		//calculate bend origin
		Point3 bendOrigin = Point3( bendBoxCentre.x,
											 bendBoxCentre.y,
											 bendBoxUpper.z ) -
										Vector3( bendRadius, 0.0, 0.0 );

		Matrix4 worldToBendMatrix = bendAxesMatrix * 
						Matrix4_translate( -bendOrigin.toVector3() );
		Matrix4 bendToWorldMatrix = ~worldToBendMatrix;

		MatrixTransformation worldToBend( worldToBendMatrix );
		BendTransformation bend( bendInRadians / box.getDepth() );
		MatrixTransformation bendToWorld( bendToWorldMatrix );
		bend.setChild( &bendToWorld );
		worldToBend.setChild( &bend );
		
		model->getModel()->transform( &worldToBend );
	}
}






ToolXfBend::ToolXfBend(CadView *v, ObjPainterManager *pman)
							  : ToolXformAxis( v, pman, "Bend", "Bend", 2 )
{
	directionLabel = gtk_label_new("Direction:");
	gtk_widget_ref(directionLabel);
	directionEntry = new NumericEntry( 0.0, 0.0, 360.0, 0.1, 1.0, 0.2, 4,
												  GTK_SIGNAL_FUNC(directionChanged), this );
	directionEntry->wrap();


	snapLabel = gtk_label_new("Snap:");
	gtk_widget_ref(snapLabel);
	snapEntry = new NumericEntry( 0.0, -1.0e8, 1.0e8, 0.01, 0.5, 0.2, 4,
											NULL, NULL );

	bendAngleLabel = gtk_label_new("Bend:");
	gtk_widget_ref(bendAngleLabel);
	bendAngleEntry = new NumericEntry( 0.0, -1.0e8, 1.0e8, 0.1, 1.0, 0.2, 4,
												  GTK_SIGNAL_FUNC(bendAngleChanged), this );


	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);


	gtk_signal_connect(GTK_OBJECT(doneButton), "clicked",
							 GTK_SIGNAL_FUNC(doneClicked), this);
	
		
	gtk_widget_show(directionLabel);
	gtk_widget_show( directionEntry->getWidget() );
	gtk_widget_show(snapLabel);
	gtk_widget_show( snapEntry->getWidget() );
	gtk_widget_show(bendAngleLabel);
	gtk_widget_show( bendAngleEntry->getWidget() );
	gtk_widget_show(doneButton);
	

	layout.lineSpace(1);
	layout.add(directionLabel);
	layout.add( directionEntry->getWidget() );
	layout.lineSpace(1);
	layout.add(snapLabel);
	layout.add( snapEntry->getWidget() );
	layout.add(bendAngleLabel);
	layout.add( bendAngleEntry->getWidget() );
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolXfBend::~ToolXfBend()
{
	delete directionEntry;
	delete snapEntry;
	delete bendAngleEntry;

	gtk_widget_unref(directionLabel);
	gtk_widget_unref(snapLabel);
	gtk_widget_unref(bendAngleLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolXfBend::getProcessType() const
{
	return ProcXfBend::getTypeStatic();
}




void ToolXfBend::initialise()
{
	ToolXformAxis::initialise();

	if ( hasProcess() )
	{
		ProcXfBend *proc = (ProcXfBend*)journalProcess;

		directionEntry->setValue( proc->getDirection() );
		bendAngleEntry->setValue( proc->getBendAngle() );
	}
}


void ToolXfBend::start()
{
	ToolXformAxis::start();

	if ( hasProcess() )
	{
		ProcXfBend *proc = (ProcXfBend*)journalProcess;
		initialBendAngleValue = proc->getBendAngle();

		snapper = ValueSnap( snapEntry->getValue() );
	}
}

void ToolXfBend::finish(bool terminated)
{
	ToolXformAxis::finish(terminated);

	if ( hasProcess() )
	{
		ProcXfBend *proc = (ProcXfBend*)journalProcess;
		if (terminated)
		{
			proc->setBendAngle( initialBendAngleValue );
			bendAngleEntry->setValue( initialBendAngleValue );
			modified();
		}
	}
}

void ToolXfBend::function()
{
	ToolXformAxis::function();

	if ( hasProcess() )
	{
		ProcXfBend *proc = (ProcXfBend*)journalProcess;
		double bendAngle = ( currentCursorPosition.getScreenFloat().x -
									startCursorPosition.getScreenFloat().x ) * 0.25;
		bendAngle += initialBendAngleValue;

		bendAngle = snapper.snap(bendAngle);
		bendAngleEntry->setValue(bendAngle);
		proc->setBendAngle( bendAngle );

		modified();
	}
}




void ToolXfBend::directionChanged(NumericEntry *widget, ToolXfBend *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfBend *proc = (ProcXfBend*)klass->journalProcess;
		proc->setDirection( klass->directionEntry->getValue() );
		klass->modified();
	}		
}

void ToolXfBend::bendAngleChanged(NumericEntry *widget, ToolXfBend *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfBend *proc = (ProcXfBend*)klass->journalProcess;
		proc->setBendAngle( klass->bendAngleEntry->getValue() );
		klass->modified();
	}		
}

void ToolXfBend::doneClicked(GtkWidget *widget, ToolXfBend *klass)
{
	klass->done();
}


#endif

