/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLXFBEND3D_CPP
#define __TOOLXFBEND3D_CPP

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>

#include <strstream>

#include "ToolXfBend3D.h"

#include "JournalModel.h"
#include "Matrix4.h"
#include "radians.h"
#include "Bend3DTransformation.h"





void* __create_ProcXfBend3D()
{
	return (void*)( new ProcXfBend3D() );
}

ObjectType type_ProcXfBend3D( "ProcXfBend3D", "ProcXformAxis",
									 __create_ProcXfBend3D );






ProcXfBend3D::ProcXfBend3D() : ProcXformAxis()
{
	bendAngle = 0.0;
}

ProcXfBend3D::ProcXfBend3D(Axis a, double bend)
									: ProcXformAxis(a)
{
	bendAngle = bend;
}

ProcXfBend3D::~ProcXfBend3D()
{
}



void ProcXfBend3D::read(FILE *f)
{
	ProcXformAxis::read(f);

	fread(&bendAngle, sizeof(double), 1, f);
}

void ProcXfBend3D::write(FILE *f)
{
	ProcXformAxis::write(f);

	fwrite(&bendAngle, sizeof(double), 1, f);
}





string ProcXfBend3D::getDescription() const
{
	return "Bend 3D";
}

const ObjectType* ProcXfBend3D::getType() const
{
	return &type_ProcXfBend3D;
}

const ObjectType* ProcXfBend3D::getTypeStatic()
{
	return &type_ProcXfBend3D;
}





JournalProcess* ProcXfBend3D::clone() const
{
	return new ProcXfBend3D(axis, bendAngle);
}



void ProcXfBend3D::setBend3DAngle(double bend)
{
	bendAngle = bend;
	modified();
}

double ProcXfBend3D::getBend3DAngle() const
{
	return bendAngle;
}




void ProcXfBend3D::transformModel(JournalModel *model, const Point3& centre,
									const Matrix4& axesMatrix,
									const Matrix4& axesMatrixInverse,
									const BBox& boundingBoxInTransformationSpace) const
{
	double bendInRadians = DEGTORAD( bendAngle );
	
	if (bendInRadians != 0.0)
	{
		const BBox &box = boundingBoxInTransformationSpace;
		
		//calculate bend radius
		double largerBoxDimension =
			box.getWidth() > box.getHeight() ? box.getWidth() : box.getHeight();

		double bendRadius = largerBoxDimension / bendInRadians;

		//make sure origin is outside of bounding box
		if ((bendRadius > 0.0) && (bendRadius < box.getDepth() * 0.5))
		{
			bendRadius = box.getDepth() * 0.5;
		}
		if ((bendRadius < 0.0) && (bendRadius > -box.getDepth() * 0.5))
		{
			bendRadius = -box.getDepth() * 0.5;
		}

		//calculate bend origin
		Point3 bendOrigin = centre - Vector3( 0.0, 0.0, bendRadius );

		Matrix4 worldToBend3DMatrix = axesMatrix * 
						Matrix4_translate( -bendOrigin.toVector3() );
		Matrix4 bendToWorldMatrix = ~worldToBend3DMatrix;

		MatrixTransformation worldToBend3D(worldToBend3DMatrix);
		Bend3DTransformation bend(bendInRadians / largerBoxDimension);
		MatrixTransformation bendToWorld(bendToWorldMatrix);
		bend.setChild(&bendToWorld);
		worldToBend3D.setChild(&bend);
		
		model->getModel()->transform(&worldToBend3D);
	}
}






ToolXfBend3D::ToolXfBend3D(CadView *v, ObjPainterManager *pman)
							  : ToolXformAxis( v, pman, "Bend3D", "Bend3D", 2 )
{
	snapLabel = gtk_label_new("Snap:");
	gtk_widget_ref(snapLabel);
	gtk_widget_show(snapLabel);
	snapEntry = new NumericEntry( 0.0, -1.0e8, 1.0e8, 0.01, 0.5, 0.2, 4,
											NULL, NULL );
	gtk_widget_show( snapEntry->getWidget() );




	bendAngleLabel = gtk_label_new("Bend3D:");
	gtk_widget_ref(bendAngleLabel);
	gtk_widget_show(bendAngleLabel);
	bendAngleEntry = new NumericEntry( 0.0, -1.0e8, 1.0e8, 0.1, 1.0, 0.2, 4,
												  GTK_SIGNAL_FUNC(bendAngleChanged), this );
	gtk_widget_show( bendAngleEntry->getWidget() );


	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);


	gtk_signal_connect(GTK_OBJECT(doneButton), "clicked",
							 GTK_SIGNAL_FUNC(doneClicked), this);
	
		
	
	layout.lineSpace(1);
	layout.add(snapLabel);
	layout.add( snapEntry->getWidget() );
	layout.add(bendAngleLabel);
	layout.add( bendAngleEntry->getWidget() );
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolXfBend3D::~ToolXfBend3D()
{
	delete snapEntry;
	delete bendAngleEntry;

	gtk_widget_unref(snapLabel);
	gtk_widget_unref(bendAngleLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolXfBend3D::getProcessType() const
{
	return ProcXfBend3D::getTypeStatic();
}




void ToolXfBend3D::initialise()
{
	ToolXformAxis::initialise();

	if ( hasProcess() )
	{
		ProcXfBend3D *proc = (ProcXfBend3D*)journalProcess;

		bendAngleEntry->setValue( proc->getBend3DAngle() );
	}
}


void ToolXfBend3D::start()
{
	ToolXformAxis::start();

	if ( hasProcess() )
	{
		ProcXfBend3D *proc = (ProcXfBend3D*)journalProcess;
		initialBend3DAngleValue = proc->getBend3DAngle();

		snapper = ValueSnap( snapEntry->getValue() );
	}
}

void ToolXfBend3D::finish(bool terminated)
{
	ToolXformAxis::finish(terminated);

	if ( hasProcess() )
	{
		ProcXfBend3D *proc = (ProcXfBend3D*)journalProcess;
		if (terminated)
		{
			proc->setBend3DAngle( initialBend3DAngleValue );
			bendAngleEntry->setValue( initialBend3DAngleValue );
			modified();
		}
	}
}

void ToolXfBend3D::function()
{
	ToolXformAxis::function();

	if ( hasProcess() )
	{
		ProcXfBend3D *proc = (ProcXfBend3D*)journalProcess;
		double bendAngle = ( currentCursorPosition.getScreenFloat().x -
									startCursorPosition.getScreenFloat().x ) * 0.25;
		bendAngle += initialBend3DAngleValue;

		bendAngle = snapper.snap(bendAngle);
		bendAngleEntry->setValue(bendAngle);
		proc->setBend3DAngle( bendAngle );

		modified();
	}
}




void ToolXfBend3D::bendAngleChanged(NumericEntry *widget, ToolXfBend3D *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfBend3D *proc = (ProcXfBend3D*)klass->journalProcess;
		proc->setBend3DAngle( klass->bendAngleEntry->getValue() );
		klass->modified();
	}		
}

void ToolXfBend3D::doneClicked(GtkWidget *widget, ToolXfBend3D *klass)
{
	klass->done();
}


#endif

