/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLXFSCALE3D_CPP
#define __TOOLXFSCALE3D_CPP

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>

#include <strstream>

#include "ToolXfScale3D.h"

#include "JournalModel.h"
#include "Matrix4.h"





void* __create_ProcXfScale3D()
{
	return (void*)( new ProcXfScale3D() );
}

ObjectType type_ProcXfScale3D( "ProcXfScale3D", "ProcXformCentre",
										 __create_ProcXfScale3D );






ProcXfScale3D::ProcXfScale3D() : ProcXformCentre()
{
	scaleFactor = 1.0;
}

ProcXfScale3D::ProcXfScale3D(Centre cen, double r)
									  : ProcXformCentre(cen)
{
	scaleFactor = r;
}

ProcXfScale3D::~ProcXfScale3D()
{
}



void ProcXfScale3D::read(FILE *f)
{
	ProcXformCentre::read(f);

	fread(&scaleFactor, sizeof(double), 1, f);
}

void ProcXfScale3D::write(FILE *f)
{
	ProcXformCentre::write(f);

	fwrite(&scaleFactor, sizeof(double), 1, f);
}





string ProcXfScale3D::getDescription() const
{
	return "Scale 3D";
}

const ObjectType* ProcXfScale3D::getType() const
{
	return &type_ProcXfScale3D;
}

const ObjectType* ProcXfScale3D::getTypeStatic()
{
	return &type_ProcXfScale3D;
}





JournalProcess* ProcXfScale3D::clone() const
{
	return new ProcXfScale3D(whichCentre, scaleFactor);
}



void ProcXfScale3D::setScaleFactor(double s)
{
	scaleFactor = s;
	modified();
}

double ProcXfScale3D::getScaleFactor() const
{
	return scaleFactor;
}




void ProcXfScale3D::transformModel(JournalModel *model, const Point3& centre) const
{
	Vector3 cenXlat = centre.toVector3();

	Matrix4 m = Matrix4_translate( -cenXlat ) *
					Matrix4_scale( Vector3(scaleFactor, scaleFactor, scaleFactor) ) *
					Matrix4_translate( cenXlat );

	model->getModel()->transform(m);
}






ToolXfScale3D::ToolXfScale3D(CadView *v, ObjPainterManager *pman)
									: ToolXformCentre( v, pman, "Scale3D",
															 "Scale3D", 2 )
{
	snapLabel = gtk_label_new("Snap:");
	gtk_widget_ref(snapLabel);
	snapEntry = new NumericEntry( 0.0, 0.0, 1.0e8, 0.01, 0.1, 0.2, 4,
											NULL, NULL );

	scaleEntry = new NumericEntry( 0.0, -1.0e8, 1.0e8, 0.0005, 0.005, 0.2, 4,
											 GTK_SIGNAL_FUNC(scaleChanged), this );
	scaleEntry->wrap();
	scaleLabel = gtk_label_new("Scale:");
	gtk_widget_ref(scaleLabel);


	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);


	gtk_signal_connect(GTK_OBJECT(doneButton), "clicked",
							 GTK_SIGNAL_FUNC(doneClicked), this);
	
		
	gtk_widget_show(snapLabel);
	gtk_widget_show( snapEntry->getWidget() );
	gtk_widget_show(scaleLabel);
	gtk_widget_show( scaleEntry->getWidget() );
	gtk_widget_show(doneButton);
	
	layout.lineSpace(1);
	layout.add(snapLabel);
	layout.add( snapEntry->getWidget() );
	layout.add(scaleLabel);
	layout.add( scaleEntry->getWidget() );
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolXfScale3D::~ToolXfScale3D()
{
	delete snapEntry;
	delete scaleEntry;

	gtk_widget_unref(snapLabel);
	gtk_widget_unref(scaleLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolXfScale3D::getProcessType() const
{
	return ProcXfScale3D::getTypeStatic();
}




void ToolXfScale3D::initialise()
{
	ToolXformCentre::initialise();

	if ( hasProcess() )
	{
		ProcXfScale3D *proc = (ProcXfScale3D*)journalProcess;
		scaleEntry->setValue( proc->getScaleFactor() );
	}
}


void ToolXfScale3D::start()
{
	ToolXformCentre::start();

	if ( hasProcess() )
	{
		ProcXfScale3D *proc = (ProcXfScale3D*)journalProcess;
		initialScaleFactor = proc->getScaleFactor();

		snapper = ValueSnap( snapEntry->getValue() );
	}
}

void ToolXfScale3D::finish(bool terminated)
{
	ToolXformCentre::finish(terminated);

	if ( hasProcess() )
	{
		ProcXfScale3D *proc = (ProcXfScale3D*)journalProcess;
		
		if (terminated)
		{
			proc->setScaleFactor( initialScaleFactor );
			scaleEntry->setValue( initialScaleFactor );
			modified();
		}
	}
}

void ToolXfScale3D::function()
{
	ToolXformCentre::function();

	if ( hasProcess() )
	{
		ProcXfScale3D *proc = (ProcXfScale3D*)journalProcess;
		double scaleFactor = ( currentCursorPosition.getScreenFloat().x -
									  startCursorPosition.getScreenFloat().x ) * 0.005;
		scaleFactor += initialScaleFactor;

		scaleFactor = snapper.snap(scaleFactor);
		scaleEntry->setValue(scaleFactor);
		proc->setScaleFactor( scaleFactor );

		modified();
	}
}




void ToolXfScale3D::scaleChanged(NumericEntry *widget, ToolXfScale3D *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfScale3D *proc = (ProcXfScale3D*)klass->journalProcess;
		proc->setScaleFactor( klass->scaleEntry->getValue() );
		klass->modified();
	}		
}

void ToolXfScale3D::doneClicked(GtkWidget *widget, ToolXfScale3D *klass)
{
	klass->done();
}


#endif

