/************************
 This code is released under the terms of the GNU GPL, v2 (or any later version).
 The GNU GPL can be found in the file 'COPYING'.  Copyright Geoffrey French 1999-2001.
 ************************/
#ifndef __TOOLXFTRIST_CPP
#define __TOOLXFTRIST_CPP

#include <gtk/gtklabel.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkradiobutton.h>

#include <strstream>

#include "ToolXfTwist.h"

#include "JournalModel.h"
#include "Matrix4.h"
#include "radians.h"
#include "MatrixTransformation.h"
#include "AbsTransformation.h"





void* __create_ProcXfTwist()
{
	return (void*)( new ProcXfTwist() );
}

ObjectType type_ProcXfTwist( "ProcXfTwist", "ProcXformCentreAxis",
									  __create_ProcXfTwist );






ProcXfTwist::ProcXfTwist() : ProcXformCentreAxis()
{
	twistAngle = 0.0;
	twistType = TwistTransformation::LINEAR;
}

ProcXfTwist::ProcXfTwist(Centre cen, Axis a, double t,
								 TwistTransformation::TwistType type)
								 : ProcXformCentreAxis(cen, a)
{
	twistAngle = t;
	twistType =  type;
}

ProcXfTwist::~ProcXfTwist()
{
}



void ProcXfTwist::read(FILE *f)
{
	ProcXformCentreAxis::read(f);

	fread(&twistAngle, sizeof(double), 1, f);
	fread(&twistType, sizeof(TwistTransformation::TwistType), 1, f);
}

void ProcXfTwist::write(FILE *f)
{
	ProcXformCentreAxis::write(f);

	fwrite(&twistAngle, sizeof(double), 1, f);
	fwrite(&twistType, sizeof(TwistTransformation::TwistType), 1, f);
}





string ProcXfTwist::getDescription() const
{
	return "Twist";
}

const ObjectType* ProcXfTwist::getType() const
{
	return &type_ProcXfTwist;
}

const ObjectType* ProcXfTwist::getTypeStatic()
{
	return &type_ProcXfTwist;
}





JournalProcess* ProcXfTwist::clone() const
{
	return new ProcXfTwist(whichCentre, axis, twistAngle, twistType);
}



void ProcXfTwist::setTwist(double t)
{
	twistAngle = t;
	modified();
}

void ProcXfTwist::setTwistType(TwistTransformation::TwistType type)
{
	twistType = type;
	modified();
}

double ProcXfTwist::getTwist() const
{
	return twistAngle;
}

TwistTransformation::TwistType ProcXfTwist::getTwistType() const
{
	return twistType;
}




void ProcXfTwist::transformModel(JournalModel *model, const Point3& centre,
									const Matrix4& axesMatrix,
									const Matrix4& axesMatrixInverse,
									const BBox& boundingBoxInTransformationSpace) const
{
	double twistInRadians = DEGTORAD(-twistAngle);
	Vector3 cenXlat = centre.toVector3();

	Matrix4 worldToTwistMatrix = axesMatrix *
										  Matrix4_translate( -cenXlat );
	Matrix4 twistToWorldMatrix = Matrix4_translate( cenXlat ) *
										  axesMatrixInverse;

	MatrixTransformation *worldToTwist =
					new MatrixTransformation(worldToTwistMatrix);
	MatrixTransformation *twistToWorld =
					new MatrixTransformation(twistToWorldMatrix);

	TwistTransformation *twist;



	if (twistType == TwistTransformation::LINEAR)
	{
		twist = new TwistTransformation(
			boundingBoxInTransformationSpace.getLower().z, 0.0,
			boundingBoxInTransformationSpace.getUpper().z, twistInRadians,
			twistType );


		Transformation *xform = worldToTwist->clone();


		twist->setChild(twistToWorld);
		xform->setChild(twist);


		model->getModel()->transform(xform);		

		delete twist;
		delete xform;
	}
	else if (twistType == TwistTransformation::RADIAL)
	{
		AbsTransformation *absolute = new AbsTransformation();
		Transformation *absBoxTransformation = worldToTwist->clone();
		absBoxTransformation->setChild(absolute);
		BBox absBoundingBox = model->getModel()->getBBox(absBoxTransformation);

		twist = new TwistTransformation( absBoundingBox.getLower().z, 0.0,
													absBoundingBox.getUpper().z, twistInRadians,
													twistType );

		
		Transformation *xform = worldToTwist->clone();


		twist->setChild(twistToWorld);
		xform->setChild(twist);


		model->getModel()->transform(xform);		

		delete twist;
		delete xform;
		delete absolute;	
	}

	delete worldToTwist;
	delete twistToWorld;
}






ToolXfTwist::ToolXfTwist(CadView *v, ObjPainterManager *pman)
								 : ToolXformCentreAxis( v, pman, "Twist", "Twist", 2 )
{
	typeLabel = gtk_label_new("Type:");
	gtk_widget_ref(typeLabel);
	gtk_widget_show(typeLabel);


	GSList *group = (GSList*)NULL;
	radioLinear = gtk_radio_button_new_with_label(group, "Linear");
	group = gtk_radio_button_group( GTK_RADIO_BUTTON(radioLinear) );
	radioRadial = gtk_radio_button_new_with_label(group, "Radial");
	gtk_widget_ref(radioLinear);
	gtk_widget_ref(radioRadial);
	gtk_widget_show(radioLinear);
	gtk_widget_show(radioRadial);

	
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(radioLinear), true );

	gtk_signal_connect( GTK_OBJECT(radioLinear), "clicked",
							  GTK_SIGNAL_FUNC(linearClicked), this );
	gtk_signal_connect( GTK_OBJECT(radioRadial), "clicked",
							  GTK_SIGNAL_FUNC(radialClicked), this );


	
	snapLabel= gtk_label_new("Snap:");
	gtk_widget_ref(snapLabel);
	gtk_widget_show(snapLabel);
	snapEntry = new NumericEntry( 0.0, 0.0, 1000.0, 0.01, 0.5, 0.2, 4,
											NULL, NULL );
	gtk_widget_show( snapEntry->getWidget() );



	twistLabel = gtk_label_new("Twist:");
	gtk_widget_ref(twistLabel);
	gtk_widget_show(twistLabel);
	twistEntry = new NumericEntry( 0.0, -3.5e8, 3.6e8, 0.1, 1.0, 0.2, 4,
											 GTK_SIGNAL_FUNC(twistChanged), this );
	gtk_widget_show( twistEntry->getWidget() );


	doneButton = gtk_button_new_with_label("Done");
	gtk_widget_ref(doneButton);
	gtk_widget_show(doneButton);


	gtk_signal_connect(GTK_OBJECT(doneButton), "clicked",
							 GTK_SIGNAL_FUNC(doneClicked), this);
	
		
	
	layout.lineSpace(1);
	layout.add(typeLabel);
	layout.add(radioLinear);
	layout.add(radioRadial);
	layout.lineSpace(1);
	layout.add(snapLabel);
	layout.add( snapEntry->getWidget() );
	layout.add(twistLabel);
	layout.add( twistEntry->getWidget() );
	layout.lineSpace(1);
	layout.addAsLine(doneButton);
	layout.newLine();
}

ToolXfTwist::~ToolXfTwist()
{
	delete snapEntry;
	delete twistEntry;

	
	gtk_widget_unref(typeLabel);
	gtk_widget_unref(radioLinear);
	gtk_widget_unref(radioRadial);
	gtk_widget_unref(snapLabel);
	gtk_widget_unref(twistLabel);
	gtk_widget_unref(doneButton);
}



const ObjectType* ToolXfTwist::getProcessType() const
{
	return ProcXfTwist::getTypeStatic();
}




void ToolXfTwist::initialise()
{
	ToolXformCentreAxis::initialise();

	if ( hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)journalProcess;

		switch( proc->getTwistType() )
		{
		case TwistTransformation::LINEAR:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(radioLinear), true );
			break;
		case TwistTransformation::RADIAL:
			gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(radioRadial), true );
			break;
		default:
			break;
		}

		twistEntry->setValue( proc->getTwist() );
	}
}


void ToolXfTwist::start()
{
	ToolXformCentreAxis::start();

	if ( hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)journalProcess;
		initialTwistValue = proc->getTwist();

		snapper = ValueSnap( snapEntry->getValue() );
	}
}

void ToolXfTwist::finish(bool terminated)
{
	ToolXformCentreAxis::finish(terminated);

	if ( hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)journalProcess;
		if (terminated)
		{
			proc->setTwist( initialTwistValue );
			twistEntry->setValue( initialTwistValue );
			modified();
		}
	}
}

void ToolXfTwist::function()
{
	ToolXformCentreAxis::function();

	if ( hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)journalProcess;

		double twist = ( currentCursorPosition.getScreenFloat().x -
								  startCursorPosition.getScreenFloat().x ) * 0.25;
		twist += initialTwistValue;

		twist = snapper.snap(twist);
		twistEntry->setValue(twist);
		proc->setTwist( twist );

		modified();
	}
}




void ToolXfTwist::linearClicked(GtkWidget *widget, ToolXfTwist *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)klass->journalProcess;
		proc->setTwistType( TwistTransformation::LINEAR );
		klass->modified();
	}		
}

void ToolXfTwist::radialClicked(GtkWidget *widget, ToolXfTwist *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)klass->journalProcess;
		proc->setTwistType( TwistTransformation::RADIAL );
		klass->modified();
	}		
}

void ToolXfTwist::twistChanged(NumericEntry *widget, ToolXfTwist *klass)
{
	if ( klass->hasProcess() )
	{
		ProcXfTwist *proc = (ProcXfTwist*)klass->journalProcess;
		proc->setTwist( klass->twistEntry->getValue() );
		klass->modified();
	}		
}

void ToolXfTwist::doneClicked(GtkWidget *widget, ToolXfTwist *klass)
{
	klass->done();
}













#endif

